/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package ch.ethz.exorciser.cyk;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Vincent Tscherter
 */
public class CYKParseTable {

   private List[] lists;

   private Object[] word;

   public CYKParseTable (Object[] word) {
      this.word = word;
      lists = new List[word.length * (word.length + 1) / 2];
      for (int x = 0; x < word.length; x++) {
         for (int y = 0; y < word.length - x; y++) {
            lists[index(x, y)] = new ArrayList();
         }
      }
   }

   public Rule getRule (int x, int y, int z) {
      return ((TableEntry) lists[index(x, y)].get(z)).rule;
   }

   public int getDY (int x, int y, int z) {
      return ((TableEntry) lists[index(x, y)].get(z)).dy;
   }

   public int getZ0 (int x, int y, int z) {
      return ((TableEntry) lists[index(x, y)].get(z)).z0;
   }

   public int getZ1 (int x, int y, int z) {
      return ((TableEntry) lists[index(x, y)].get(z)).z1;
   }

   public boolean isValid (int x, int y, int z) {
      return ((TableEntry) lists[index(x, y)].get(z)).valid;
   }

   public int height (int x, int y) {
      return lists[index(x, y)].size();
   }

   public void add (Rule rule, int x, int y) {
      TableEntry e = new TableEntry(rule);
      if (!lists[index(x, y)].contains(e)) {
         lists[index(x, y)].add(e);
      }
   }

   public void remove (int x, int y, int z) {
      lists[index(x, y)].remove(z);
   }

   public void removeAll () {
      for (int x = 0; x < word.length; x++) {
         for (int y = 0; y < word.length - x; y++) {
            lists[index(x, y)].clear();
         }
      }
   }

   public boolean contains (Rule rule, int x, int y) {
      TableEntry e = new TableEntry(rule);
      return lists[index(x, y)].contains(e);
   }

   public boolean containsPattern (Object pattern, int x, int y) {
      List l = lists[index(x, y)];
      for (int i = 0; i < l.size(); i++) {
         Rule rule = ((TableEntry) l.get(i)).rule;
         if (rule.getPattern().length == 1
               && rule.getPattern()[0].equals(pattern)) { return true; }
      }
      return false;
   }

   private int indexOfValidPattern (Object pattern, int x, int y) {
      List l = lists[index(x, y)];
      for (int z = 0; z < l.size(); z++) {
         TableEntry e = (TableEntry) l.get(z);
         if (e.valid && e.rule.getPattern().length == 1
               && e.rule.getPattern()[0].equals(pattern)) { return z; }
      }
      return -1;
   }

   private int index (int x, int y) {
      return (x + y) * (x + y + 1) / 2 + y;
   }

   public int size () {
      return word.length;
   }

   public String toString () {
      String desc = ""; //$NON-NLS-1$
      for (int y = 0; y < word.length; y++) {
         for (int x = 0; x < word.length - y; x++) {
            for (int z = 0; z < lists[index(x, y)].size(); z++) {
               desc += lists[index(x, y)].get(z) + ":"; //$NON-NLS-1$
            }
            desc += ";"; //$NON-NLS-1$
         }
         desc += "\n"; //$NON-NLS-1$
      }
      return desc;
   }

   public void validate () {
      for (int x = 0; x < word.length; x++) {
         for (int z = 0; z < height(x, 0); z++) {
            TableEntry e = (TableEntry) lists[index(x, 0)].get(z);
            e.valid = e.rule.getPattern().length == 1
                  && e.rule.getReplacement().length == 1
                  && e.rule.getReplacement()[0].equals(word[x]);
         }
      }
      for (int y = 1; y < word.length; y++) {
         for (int x = 0; x < word.length - y; x++) {
            for (int z = 0; z < lists[index(x, y)].size(); z++) {
               TableEntry e = (TableEntry) lists[index(x, y)].get(z);
               if (e.rule.getPattern().length == 1
                     && e.rule.getReplacement().length == 2) {
                  e.z0 = -1;
                  e.z1 = -1;
                  e.dy = 0;
                  while (e.dy < y && (e.z0 == -1 || e.z1 == -1)) {
                     e.z0 = indexOfValidPattern(e.rule.getReplacement()[0], x,
                           e.dy);
                     e.z1 = indexOfValidPattern(e.rule.getReplacement()[1], x
                           + e.dy + 1, y - e.dy - 1);
                     e.dy++;
                  }
                  e.valid = e.z0 != -1 && e.z1 != -1;
                  e.dy--;
               } else {
                  e.valid = false;
               }
            }
         }
      }
   }

   private class TableEntry {

      Rule rule;

      boolean valid;

      int dy, z0, z1;

      TableEntry (Rule rule) {
         this.rule = rule;
      }


      public boolean equals (Object obj) {
         return obj instanceof TableEntry
               && ((TableEntry) obj).rule.equals(rule);
      }
   }
}