/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * A SortedSet with the Character constants EPSILON and EMPTY and a method containsEpsilon():boolean. 
 * Delegates all SortedSet methods to a TreeSet.
 * With one constructor, an element to be added can be passed (usually EPSILON).
 * 
 * @author		Reto Lamprecht
 */
public class Alphabet implements SortedSet {

	private TreeSet alphImpl;

	public static Character EPSILON = new Character('\u03b5');
	public static Character EMPTY = new Character('\u00f8');

	public boolean containsEpsilon() {
		return alphImpl.contains(EPSILON);
	}

	public Alphabet() {
		alphImpl = new TreeSet();
	}

	public Alphabet(Object element) {
		this();
		if (null != element) {
			alphImpl.add(element);
		}
	}

	protected Alphabet(Alphabet alphabet) {
		this.alphImpl = (TreeSet) alphabet.alphImpl.clone();
	}

	public Object clone() {
		return new Alphabet(this);
	}

	/*
	 * @see SortedSet#comparator()
	 */
	public Comparator comparator() {
		return alphImpl.comparator();
	}

	/*
	 * @see SortedSet#first()
	 */
	public Object first() {
		return alphImpl.first();
	}

	/*
	 * @see SortedSet#headSet(Object)
	 */
	public SortedSet headSet(Object arg0) {
		return alphImpl.headSet(arg0);
	}

	/*
	 * @see SortedSet#last()
	 */
	public Object last() {
		return alphImpl.last();
	}

	/*
	 * @see SortedSet#subSet(Object, Object)
	 */
	public SortedSet subSet(Object arg0, Object arg1) {
		return alphImpl.subSet(arg0, arg1);
	}

	/*
	 * @see SortedSet#tailSet(Object)
	 */
	public SortedSet tailSet(Object arg0) {
		return alphImpl.tailSet(arg0);
	}

	/*
	 * @see Collection#add(Object)
	 */
	public boolean add(Object arg0) {
		return alphImpl.add(arg0);
	}

	/*
	 * @see Collection#addAll(Collection)
	 */
	public boolean addAll(Collection arg0) {
		return alphImpl.addAll(arg0);
	}

	/*
	 * @see Collection#clear()
	 */
	public void clear() {
		alphImpl.clear();
	}

	/*
	 * @see Collection#contains(Object)
	 */
	public boolean contains(Object arg0) {
		return alphImpl.contains(arg0);
	}

	/*
	 * @see Collection#containsAll(Collection)
	 */
	public boolean containsAll(Collection arg0) {
		return alphImpl.containsAll(arg0);
	}

	/*
	 * @see Collection#isEmpty()
	 */
	public boolean isEmpty() {
		return alphImpl.isEmpty();
	}

	/*
	 * @see Collection#iterator()
	 */
	public Iterator iterator() {
		return alphImpl.iterator();
	}

	/*
	 * @see Collection#remove(Object)
	 */
	public boolean remove(Object arg0) {
		return alphImpl.remove(arg0);
	}

	/*
	 * @see Collection#removeAll(Collection)
	 */
	public boolean removeAll(Collection arg0) {
		return alphImpl.removeAll(arg0);
	}

	/*
	 * @see Collection#retainAll(Collection)
	 */
	public boolean retainAll(Collection arg0) {
		return alphImpl.retainAll(arg0);
	}

	/*
	 * @see Collection#size()
	 */
	public int size() {
		return alphImpl.size();
	}

	/*
	 * @see Collection#toArray()
	 */
	public Object[] toArray() {
		return alphImpl.toArray();
	}

	/*
	 * @see Collection#toArray(Object[])
	 */
	public Object[] toArray(Object[] arg0) {
		return alphImpl.toArray(arg0);
	}

	public static Alphabet create01Alphabet() {
		Alphabet a = new Alphabet();
		a.add(new Character('0'));
		a.add(new Character('1'));
		return a;
	}

	public static Alphabet create01eAlphabet() {
		Alphabet a = new Alphabet();
		a.add(new Character('0'));
		a.add(new Character('1'));
		a.add(Alphabet.EPSILON);
		return a;
	}

	public static Alphabet createABAlphabet() {
		Alphabet a = new Alphabet();
		a.add(new Character('a'));
		a.add(new Character('b'));
		return a;
	}

	public static Alphabet createABeAlphabet() {
		Alphabet a = new Alphabet();
		a.add(new Character('a'));
		a.add(new Character('b'));
		a.add(Alphabet.EPSILON);
		return a;
	}

	public boolean isDualAlphabeh() {
		return (containsEpsilon() && size() == 3)
			|| (!containsEpsilon() && size() == 2);
	}
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object o) {
		if (o instanceof Alphabet) {
			Alphabet alpha = (Alphabet) o;
			Iterator i = alpha.iterator();
			while (i.hasNext()) {
				if (!contains(i.next())) {
					return false;
				}
			}
			i = iterator();
			while (i.hasNext()) {
				if (!alpha.contains(i.next())) {
					return false;
				}
			}
			return true;
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String s = "{"; //$NON-NLS-1$
		Iterator i = iterator();
		if (i.hasNext()) {
			s = s+i.next();
		}
		while (i.hasNext()) {
			s = s+", "+i.next(); //$NON-NLS-1$
		}
		s = s+"}"; //$NON-NLS-1$
		return s;
	}

}