/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.util.ArrayList;
import java.util.List;

/**
 * @author		Reto Lamprecht
 */
public class CircleLayouter {

	public static boolean layout(FSMView fsmView) {

		double radius = fsmView.getStateDiameter() / 2;

		double cx = fsmView.getWidth() / 2;
		double cy = fsmView.getHeight() / 2;
		// radius that would fit into window
		double rWin = radius + 0.6 * (Math.min(cx, cy) - 2 * radius);
		// minimal radius for current number of states.
		double rStates = 0.5 / Math.PI * (fsmView.getNofStates() * fsmView.minStateDist * 1.5);
		double r = Math.max(rWin, rStates);

		int n = fsmView.getNofStates();

		// keep initial configuration
		Object[] tmpStates = fsmView.getStates();

		// init perturbation array
		int[] perturb = new int[n];
		int[] bestperturb = new int[n];
		for (int i = 0; i < n; i++) {
			perturb[i] = i;
			bestperturb[i] = i;
		}

		// get first 100 permutations into Vector perm.
		List perm = new ArrayList();
		first100Permutations(perm, 0, new int[n], perturb);

		// look for arrangement with least intersections
		int isects;
		int minisects = n * n * n * n;

		// turn off grid alignment but keep settings
		boolean alignbak = fsmView.isGridAligned();
		fsmView.alignWithGrid(false);

		for (int k = 0; k < perm.size(); k++) {
			perturb = (int[]) perm.get(k);
			placeStates(fsmView, tmpStates, perturb, cx, cy, r);
			// count intersections
			isects = 0;
			Transition current1, current2;
			for (int i = 0; i < fsmView.getNofTransitions() - 1; i++) {
				current1 = fsmView.getTransition(i);
				for (int j = i + 1; j < fsmView.getNofTransitions(); j++) {
					current2 = fsmView.getTransition(j);
					fsmView.calcTransitionCoords(current1);
					fsmView.calcTransitionCoords(current2);
					if (intersect(current1, current2)) {
						isects++;
					}
				}
			}
			// save intersections with copy of corresponding perturb.
			if (isects < minisects) {
				minisects = isects;
				System.arraycopy(perturb, 0, bestperturb, 0, perturb.length);
			}
			if (minisects == 0) {
				break;
			}

		}

		// place states on model with bestperturb:
		// get startstate index to bring startstate to the left.
		int startindex = 0;
		for (int i = 0; i < n; i++) {
			if (((State) tmpStates[bestperturb[i]]).isStart()) {
				startindex = i;
				break;
			}
		}
		fsmView.states.clear();
		State current;
		for (int i = 0; i < n; i++) {
			current = ((State) tmpStates[bestperturb[(i + startindex) % n]]);
			fsmView.states.add(current);
			current.model.setPosition(
				(cx + r * Math.cos(Math.PI + i * 2 * Math.PI / n)),
				(cy + r * Math.sin(Math.PI + i * 2 * Math.PI / n)));
		}

		fsmView.alignWithGrid(alignbak);
		fsmView.fit();

		return minisects <= 2;
	}

	private static void placeStates(FSMView fsmView, Object[] tmpStates, int[] perturb, double cx, double cy, double r) {
		int n = tmpStates.length;
		fsmView.states.clear();
		State current;
		for (int i = 0; i < n; i++) {
			current = ((State) tmpStates[perturb[i]]);
			fsmView.states.add(current);
			fsmView.moveState(
				current,
				(int) (cx + r * Math.cos(i * 2 * Math.PI / n)),
				(int) (cy + r * Math.sin(i * 2 * Math.PI / n)));
		}
	}

	private static boolean intersect(Transition t1, Transition t2) {

		// convert line segments to y=mx+c and y=nx+d line format
		double m = (t1.toY - t1.fromY) / (t1.toX - t1.fromX);
		double c = t1.fromY - m * t1.fromX;
		double n = (t2.toY - t2.fromY) / (t2.toX - t2.fromX);
		double d = t2.fromY - n * t2.fromX;

		// determine point of intersection of lines
		double x, y;
		if (Double.isInfinite(m) && Double.isInfinite(n)) {
			return (false); // parallel lines - no intersection
		} else if (Double.isInfinite(m)) { // first line is vertical
			x = t1.fromX;
			y = n * x + d;
		} else if (Double.isInfinite(n)) { // second line is vertical
			x = t2.fromX;
			y = m * x + c;
		} else {
			x = (d - c) / (m - n);
			y = m * x + c;
		}

		// does point of intersection lie on both line segments?
		if ((x >= Math.min(t1.fromX, t1.toX))
			&& (x <= Math.max(t1.fromX, t1.toX))
			&& (y >= Math.min(t1.fromY, t1.toY))
			&& (y <= Math.max(t1.fromY, t1.toY))
			&& (x >= Math.min(t2.fromX, t2.toX))
			&& (x <= Math.max(t2.fromX, t2.toX))
			&& (y >= Math.min(t2.fromY, t2.toY))
			&& (y <= Math.max(t2.fromY, t2.toY)))
			return (true);
		return (false);
	}

	private static List first100Permutations(List collection, int level, int[] trail, int[] rest) {
		if (rest.length == 1) {
			trail[level] = rest[0];
			collection.add(trail);
		} else {
			int ireached;
			for (int i = 0;(i < rest.length && collection.size() < 100); i++) {
				int[] newtrail = new int[trail.length];
				System.arraycopy(trail, 0, newtrail, 0, trail.length);
				int[] newrest = new int[rest.length - 1];
				ireached = 0;
				for (int k = 0; k < rest.length; k++) {
					if (i == k) {
						ireached = 1;
					} else {
						newrest[k - ireached] = rest[k];
					}
				}
				newtrail[level] = rest[i];
				first100Permutations(collection, level + 1, newtrail, newrest);
			}
		}
		return collection;

	}
}