/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;

/**
 * @author		Reto Lamprecht
 */
public abstract class ComponentRendererAutoCalcTrans
	implements ComponentRenderer {

	/** The maximum distance of converse transitions from a straight connecting line between from- and to-state. */
	int CONV_TRANS_GAP = 25;

	/** The size of self-transitions. */
	int SELFSIZE = 50;
	/** The angle of self-transitions. */
	double SELFANGLE = Math.toRadians(25);

	/** The transition head's length. */
	double HEADLENGTH = 12.0;


	/** The FSMView Object that this renderer should belong to. */
	FSMView fsmView;

	public String getKey() {
		return ""; //$NON-NLS-1$
	}

	/** sets this Renderer's FAscape Object. */
	public void setFSMView(FSMView fsmView) {
		this.fsmView = fsmView;
	}

	/** Draws the background. */
	public abstract void drawBackground(int width, int height, Graphics2D g);

	/** Returns the text color of this renderer. */
	public abstract Color getTextColor();

	/** Returns the width of the state strokes. Used to calculate attaching points for transitions. */
	protected double getStateStrokeWidth() {
		return 0.0;
	}

	/** Returns the length of the transition's head. Used to calculate attaching points for transitions. */
	protected double getTransitionHeadLength() {
		return HEADLENGTH;
	}

	/** Returns the distance of the state's edge from it's center for a given angle.
	 */
	public abstract double getStateDiameter(double angle);

	/** Returns the bounding box of the specified state including his selftransition.
	 */
	public abstract Rectangle2D getStateBounds(State s);

	/** Draws the specified state in the specified Graphics window.
	 */
	public abstract void drawState(State s, Graphics2D g);

	/** Draws special marks separately.
	 */
	public abstract void markState(State s, Graphics2D g);

	/** Draws the specified transition in the specified Graphics window.
	 */
	public abstract void drawTransition(Transition t, Graphics2D g);

	/** Calculates the specified transition's coordinates depending on it's type and from- and to-states.
	 */
	public void calcTransitionCoords(Transition t) {

		// java.lang.Math atan2(x, y) and sin(a), cos(a) definition:
		//
		// - atan2(1.0, 0.0) = Math.PI/2
		// - atan2(0.0, 1.0) = 0.0
		// i.e. counterclockwise with 0 = down, [PI..2PI] = [-PI..0]
		//
		// - atan2(sin(1.234), cos(1.234)): 1.234
		// - sin(atan2(1.0, 0.0)): 1.0
		// - cos(atan2(0.0, 1.0)): 1.0
		// i.e. sin = x, cos = y in atan2(x, y)
		//

		// use radius of dynamic type
		double stateRadius = this.getStateDiameter(0.0) / 2;
		double stroke = this.getStateStrokeWidth() / 2;
		double headLength = this.getTransitionHeadLength();

		if (null == t.to) {
			t.type = Transition.LINE;

			t.angle = Math.atan2(t.toX - t.from.centerX, t.toY - t.from.centerY);
			t.angleHead = t.angle;
			t.fromX = t.from.centerX + (stateRadius + stroke) * Math.sin(t.angle);
			t.fromY = t.from.centerY + (stateRadius + stroke) * Math.cos(t.angle);
			t.toX = t.toX - (int) ((headLength / 2 + stroke) * Math.sin(t.angle));
			t.toY = t.toY - (int) ((headLength / 2 + stroke) * Math.cos(t.angle));

		} else if (t.from == t.to) {
			t.type = Transition.CUBIC;

			t.angle = getSelfTransitionAngle(t);
			t.angleHead = t.angle + SELFANGLE;
			t.fromX =
				t.from.centerX - (stateRadius + stroke) * Math.sin(t.angle - SELFANGLE);
			t.fromY =
				t.from.centerY - (stateRadius + stroke) * Math.cos(t.angle - SELFANGLE);
			t.toX =
				t.to.centerX
					- (stateRadius + headLength / 2 + stroke) * Math.sin(t.angle + SELFANGLE);
			t.toY =
				t.to.centerY
					- (stateRadius + headLength / 2 + stroke) * Math.cos(t.angle + SELFANGLE);
			t.ctrlFromX = t.fromX - SELFSIZE * Math.sin(t.angle - SELFANGLE);
			t.ctrlFromY = t.fromY - SELFSIZE * Math.cos(t.angle - SELFANGLE);
			t.ctrlToX = t.toX - SELFSIZE * Math.sin(t.angle + SELFANGLE);
			t.ctrlToY = t.toY - SELFSIZE * Math.cos(t.angle + SELFANGLE);

		} else {
			if (converseTransition(t)) {
				t.type = Transition.QUAD;

				t.angle =
					Math.atan2(t.to.centerX - t.from.centerX, t.to.centerY - t.from.centerY);
				double midX = t.from.centerX + (t.to.centerX - t.from.centerX) / 2;
				double midY = t.from.centerY + (t.to.centerY - t.from.centerY) / 2;
				t.ctrlFromX = midX + CONV_TRANS_GAP * Math.cos(t.angle);
				t.ctrlFromY = midY - CONV_TRANS_GAP * Math.sin(t.angle);
				double angleFrom =
					Math.atan2(t.ctrlFromX - t.from.centerX, t.ctrlFromY - t.from.centerY);
				double angleTo =
					Math.atan2(t.to.centerX - t.ctrlFromX, t.to.centerY - t.ctrlFromY);
				t.fromX = t.from.centerX + (stateRadius + stroke) * Math.sin(angleFrom);
				t.fromY = t.from.centerY + (stateRadius + stroke) * Math.cos(angleFrom);
				t.toX =
					t.to.centerX - (stateRadius + headLength / 2 + stroke) * Math.sin(angleTo);
				t.toY =
					t.to.centerY - (stateRadius + headLength / 2 + stroke) * Math.cos(angleTo);
				t.angleHead = angleTo;

			} else {
				t.type = Transition.LINE;

				t.angle =
					Math.atan2(t.to.centerX - t.from.centerX, t.to.centerY - t.from.centerY);
				t.angleHead = t.angle;
				t.fromX = t.from.centerX + (stateRadius + stroke) * Math.sin(t.angle);
				t.fromY = t.from.centerY + (stateRadius + stroke) * Math.cos(t.angle);
				t.toX =
					t.to.centerX - (stateRadius + headLength / 2 + stroke) * Math.sin(t.angle);
				t.toY =
					t.to.centerY - (stateRadius + headLength / 2 + stroke) * Math.cos(t.angle);
			}
		}
	}

	/** Returns the bounding box of the specified transitionlabel.
	 */
	public Rectangle2D getTransitionLabelBounds(TransitionLabel tl) {
		updateTransitionLabelBounds(tl);
		return tl.bounds;
	}

	/** Updates the bounding box of the specified transitionlabel.
	 */
	protected abstract void updateTransitionLabelBounds(TransitionLabel tl);

	/** Draws the specified transitionlabel in the specified Graphics window.
	 */
	public abstract void drawTransitionLabel(TransitionLabel tl, Graphics2D g);

//	private double fromLineOffset = 0.0;

	protected abstract double getAdjustX(TransitionLabel tl, double adjustAngle);
	//{
//		return 0.0;
//	}

	protected abstract double getAdjustY(TransitionLabel tl, double adjustAngle); //{
		//return 0.0;
	//}

	protected static String RenderingHint_WithBGR = "w"; //$NON-NLS-1$

	/** Calculates the specified transitionlabel's coordinates depending on it's transition's type and from- and to-states.
	 */

	public void calcTransitionLabelCoords(TransitionLabel tl) {

		// 1: transitionlabels of QUAD and LINE transitions avoid 
		//    (a) states, 
		//    (b) transition lines and 
		//    (c) other transitionlabels.
		//    the method labelTransFreePoint(tl) returns true, if (b) and (c) are no problem
		//    the method labelStateFreePoint(tl) returns true, if (a) are no problem.
		//
		// 2: transitionlabels retreat
		//    (a) by moving onto their transition line (change adjustX/Y by stepAdjustX/Y) and 
		//    (b) by moving away from the middle of the transition (change offs by stepOffs).

		updateTransitionLabelBounds(tl);
		// middle of connecting line between from- and to-state:
		double midX = tl.fat.fromX + (tl.fat.toX - tl.fat.fromX) / 2;
		double midY = tl.fat.fromY + (tl.fat.toY - tl.fat.fromY) / 2;

		// CUBIC //////////////////////////
		if (tl.fat.type == Transition.CUBIC) {
			// middle between the two control points of the cubic curve:
			double ctrlMidX = tl.fat.ctrlFromX + (tl.fat.ctrlToX - tl.fat.ctrlFromX) * 0.5;
			double ctrlMidY = tl.fat.ctrlFromY + (tl.fat.ctrlToY - tl.fat.ctrlFromY) * 0.5;
			// place the label on the line first
			tl.x = ctrlMidX - (ctrlMidX - midX) * 0.25;
			tl.y = ctrlMidY - (ctrlMidY - midY) * 0.25;
			tl.angle = Math.atan2(ctrlMidX - midX, ctrlMidY - midY);
			// move it next to the line
			tl.x = tl.x + getAdjustX(tl, tl.angle);
			tl.y = tl.y + getAdjustY(tl, tl.angle);
		}

		// QUAD & LINE ///////////////////////////////	
		else {

			// QUAD //////////////////////////////////
			// define origin (oX,oY) and coordinate vectors (base1X, base1Y), (base2X, base2Y) in which the quad curve forms a y=x*x curve.
			// this makes it possible to place the label on an arbitrary point on the quad curve by using the formula:
			// tl.x = oX + base1X * offs*offs + base2X * prefix*offs;
			// tl.y = oY + base1Y * offs*offs + base2Y * prefix*offs;
			// where offs goes from 0 (middle) to 1 or -1 (edges)
			double oX = tl.fat.ctrlFromX - (tl.fat.ctrlFromX - midX) * 0.5;
			double oY = tl.fat.ctrlFromY - (tl.fat.ctrlFromY - midY) * 0.5;
			double base1X = midX - oX;
			double base1Y = midY - oY;
			double base2X = tl.fat.fromX - midX;
			double base2Y = tl.fat.fromY - midY;

			// an angle orthogonal to the transitions angle, used to adjust the label next to the line
			double adjustAngle = 0.0;
			// QUAD ///////////////////////////
			if (tl.fat.type == Transition.QUAD) {
				adjustAngle = tl.fat.angle + Math.toRadians(90);
			}
			// LINE ////////////////////////////////
			else if (tl.fat.type == Transition.LINE) {
				// the transition label should be above the transition line:
				// for angles in [0..PI] place label to the left 
				// for angles in [PI..2PI] place label to the right.
				if (tl.fat.angle > Math.toRadians(0) && tl.fat.angle < Math.toRadians(180))
					adjustAngle = tl.fat.angle + Math.toRadians(90);
				else
					adjustAngle = tl.fat.angle - Math.toRadians(90);
			}

			// adjustX/Y define how far the label is moved away from the transition line in direction of adjustAngle. 
			double adjustX = getAdjustX(tl, adjustAngle);
			double adjustY = getAdjustY(tl, adjustAngle);

			// place the label on the line first
			// QUAD ///////////////////////////
			if (tl.fat.type == Transition.QUAD) {
				tl.x = oX;
				tl.y = oY;
			}
			// LINE ////////////////////////////////
			else if (tl.fat.type == Transition.LINE) {
				tl.x = tl.fat.fromX + (tl.fat.toX - tl.fat.fromX) * 0.5;
				tl.y = tl.fat.fromY + (tl.fat.toY - tl.fat.fromY) * 0.5;
			}

			// move it next to the line
			tl.x = tl.x + adjustX;
			tl.y = tl.y + adjustY;

			if (fsmView.REPAINT_SOPHISTICATED) {

				// retreat strategy (a) (moving onto transition line):
				int nOfAdjustSteps = 10;
				double stepAdjustX = adjustX / nOfAdjustSteps;
				double stepAdjustY = adjustY / nOfAdjustSteps;

				// retreat strategy (b) (moving away from the middle of the transition):
				double stepOffs =
					1
						/ Math.max(
							Math.abs(tl.fat.toX - tl.fat.fromX),
							Math.abs(tl.fat.toY - tl.fat.fromY));
				double offs = 0.0;
				double prefix = 1; // indicates the direction from the middle

				// big loop: transition label retreats from middle of transition
				while (offs < 0.5
					&& (!fsmView.labelTransFreePoint(tl) || !fsmView.labelStateFreePoint(tl))) {

					adjustX = getAdjustX(tl, adjustAngle);
					adjustY = getAdjustY(tl, adjustAngle);

					int tryCount = 0;

					// inner loop: transition label retreats onto transition line for each distance from middle.
					while (tryCount < nOfAdjustSteps
						&& (!fsmView.labelTransFreePoint(tl) || !fsmView.labelStateFreePoint(tl))) {

						// place it on the line first
						// QUAD ///////////////////////////
						if (tl.fat.type == Transition.QUAD) {
							tl.x = oX + base1X * offs * offs + base2X * prefix * offs;
							tl.y = oY + base1Y * offs * offs + base2Y * prefix * offs;
						}
						// LINE ////////////////////////////////
						else if (tl.fat.type == Transition.LINE) {
							tl.x = tl.fat.fromX + (tl.fat.toX - tl.fat.fromX) * (prefix * offs + 0.5);
							tl.y = tl.fat.fromY + (tl.fat.toY - tl.fat.fromY) * (prefix * offs + 0.5);
						}

						// move label next to the line
						tl.x = tl.x + adjustX;
						tl.y = tl.y + adjustY;

						adjustX -= stepAdjustX;
						adjustY -= stepAdjustY;

						tryCount++;
					}

					// paint label with background if it's too close to line:
					tl.removeRenderingHint(RenderingHint_WithBGR);
					if (Math.abs(adjustX)
						< Math.abs((tl.bounds.width * 0.5 + 0.0) * Math.sin(adjustAngle))
						|| Math.abs(adjustY)
							< Math.abs((tl.bounds.height * 0.5 + 0.0) * Math.cos(adjustAngle))) {
						tl.addRenderingHint(RenderingHint_WithBGR);
					}

					offs += stepOffs;
					prefix = -prefix;
				}
			}
		}
	}

	/** Alters the specified state's state depending on mouse position.
	 *  Returns true if any changes have to be repainted.
	 */
	public abstract boolean onMouseOver(State s, double x, double y);

	/** Alters the specified transitions's state depending on mouse position.
	 */
	public abstract void onMouseOver(Transition t, double x, double y);

	/** Alters the specified transitionlabel's state depending on mouse position.
	 */
	public abstract void onMouseOver(TransitionLabel tl, double x, double y);

	/** Returns true if (x, y) lies inside the specified state.
	 */
	public abstract boolean contains(State s, double x, double y);

	/** Returns true if (x, y) lies inside the specified transition.
	 */
	public abstract boolean contains(Transition t, double x, double y);

	/** Returns true if (x, y) lies inside the specified transitionlabel.
	 */
	public abstract boolean contains(TransitionLabel tl, double x, double y);

	/** Checks if there is a transition coming from this transition's to state going to this transition's from state.
	 *
	 * @return  true if a converse transition exists.
	 */
	protected boolean converseTransition(Transition t) {
		if (null != t.to && t.to != t.from) {
			for (int i = 0; i < t.from.inTransitions.size(); i++) {
				if (t.to == ((Transition) t.from.inTransitions.get(i)).from) {
					return true;
				}
			}
		}
		return false;
	}

	/** Calculates the best position for a selftransition (only one per state) .
	 *
	 * @return angle of self-transition in radians.
	 */
	protected double getSelfTransitionAngle(Transition t) {
		List angles = new ArrayList();
		// startstate has a pseudo incoming transition on the left.
		if (t.from.isStart()) {
			angles.add(new Double(1.5 * Math.PI));
		}
		addOutTransitionAngles(t, angles);
		addInTransitionAngles(t, angles);
		return getMiddleOfLargestGap(angles);
	}

	/** Adds angles of all leaving transitions to the specified Vector.
	 *  Values are translated into range [0..2PI] and put in ascending order.
	 */
	protected void addOutTransitionAngles(Transition t, List angles) {

		List tOut = t.from.outTransitions;

		Double sentinel = new Double(9.9);
		angles.add(sentinel);

		Transition current;
		for (int i = 0; i < tOut.size(); i++) {
			current = (Transition) tOut.get(i);
			if (current.to != t.from) { // current is this selftransition
				calcTransitionCoords(current);
				if (current.angle < 0)
					current.angle += 2 * Math.PI;
				for (int j = 0; j < angles.size(); j++) {
					if (current.angle < ((Double) angles.get(j)).doubleValue()) {
						angles.add(j, new Double(current.angle));
						break;
					}
				}
			}
		}
		angles.remove(sentinel);
	}

	/** Adds angles of all incoming transitions to the specified Vector
	 *  Values are translated into range [0..2PI] and put in ascending order.
	 */
	protected void addInTransitionAngles(Transition t, List angles) {

		List tIn = t.from.inTransitions;

		Double sentinel = new Double(9.9);
		angles.add(sentinel);

		Double tmp;
		Transition current;
		for (int i = 0; i < tIn.size(); i++) {
			current = (Transition) tIn.get(i);
			if (current.from != t.from) { // current is this selftransition
				calcTransitionCoords(current);
				tmp = new Double((current.angle + Math.PI) % (2 * Math.PI));
				// reverse angle, it's an incoming transition
				if (tmp.doubleValue() < 0)
					tmp = new Double(tmp.doubleValue() + 2 * Math.PI);
				for (int j = 0; j < angles.size(); j++) {
					if (tmp.doubleValue() < ((Double) angles.get(j)).doubleValue()) {
						angles.add(j, tmp);
						break;
					}
				}
			}
		}
		angles.remove(sentinel);
	}

	/** Returns the angle of the middle of the largest gap in the angle values in the specified Vector.
	 *  To use this method properly, angle values have to be translated into range [0..2PI].
	 */
	protected double getMiddleOfLargestGap(List angles) {
		if (angles.size() == 0)
			return 0.0;
		if (angles.size() == 1)
			return ((Double) angles.get(0)).doubleValue();
		// find largest gap between transitions:
		// determine gap between consecutive angles and compare it with the maximum gap so far.
		// if current gap is larger, la gets the value of the smaller angle that limits the gap.
		double maxRange = 0, currRange, bottomOfRange = 0.0;
		for (int j = 1; j < angles.size(); j++) {
			currRange =
				((Double) angles.get(j)).doubleValue()
					- ((Double) angles.get(j - 1)).doubleValue();
			if (currRange > maxRange) {
				maxRange = currRange;
				bottomOfRange = ((Double) angles.get(j - 1)).doubleValue();
			}
		}
		// handle gap between last and first angle
		currRange =
			((Double) angles.get(0)).doubleValue()
				+ 2 * Math.PI
				- ((Double) angles.get(angles.size() - 1)).doubleValue();
		if (currRange > maxRange) {
			maxRange = currRange;
			bottomOfRange = ((Double) angles.get(angles.size() - 1)).doubleValue();
		}

		return (Math.PI + bottomOfRange + maxRange / 2) % (2 * Math.PI);
	}

	/** Converts a float-array to a GeneralPath. Used to draw the startstate nick.
	 *
	 * @param   res     The float array containing coordinates. even elements are x-coords, odd elements y -coords.
	 * @retruns The specified ccordinates as GeneralPath.
	 */
	protected GeneralPath makePath(float[] coords) {
		GeneralPath path = new GeneralPath();
		path.reset();
		path.moveTo(coords[0], coords[1]);
		for (int i = 2; i < coords.length - 1; i += 2) {
			path.lineTo(coords[i], coords[i + 1]);
		}
		path.closePath();
		return path;
	}

}