/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;


public class FSMEditorOperations {

	FSMModelInterface model;
	// needed for view actions (mark) and to get corresponding visible elements to model elements.
	FSMView view;

	Stack undoList = new Stack();
	Stack redoList = new Stack();

	public FSMEditorOperations(FSMModelInterface model, FSMView view) {
		this.model = model;
		this.view = view;
	}

	public void setModel(FSMModelInterface model) {
		//undo behaviour?
		this.model = model;
	}

	///////////////////
	// Model Changes //
	///////////////////

	public void addState(FSMStateInterface s) {
		EditorAction ea =
			new EditorAction(
				model,
				FSMEvent.createDeleteStateEvent(s),
				FSMEvent.createAddStateEvent(s));
		model.addState(s);
		addUndoEditorAction(ea);
	}

	public void addStartState(FSMStateInterface s) {
		EditorAction ea =
			new EditorAction(
				model,
				FSMEvent.createDeleteStateEvent(s),
				FSMEvent.createAddStateEvent(s));
		model.addState(s);
		ea.addRedoEvent(FSMEvent.createSetStartStateEvent(s));
		ea.addUndoEvent(FSMEvent.createSetStartStateEvent(null));
		model.setStartState(s);
		addUndoEditorAction(ea);
	}

	public void deleteState(FSMStateInterface s) {
		EditorAction ea = new EditorAction(model);
		// save inTransitions
		//Vector deletedTransitions = new Vector();
		Iterator iter = model.getAllStates().iterator();
		while (iter.hasNext()) { // for each state in model
			FSMStateInterface tmp = (FSMStateInterface) iter.next();
			Set inpSet = tmp.getInputs();
			if (null != inpSet) { // if any outTransitions
				Iterator alphIter = inpSet.iterator();
				while (alphIter.hasNext()) { // for each input of an outTransition
					Character inp = (Character) alphIter.next();
					Set nextStates = tmp.getNextStateSet(inp);
					Iterator stateIter = nextStates.iterator();
					while (stateIter.hasNext()) { // save the transition...
						FSMStateInterface nextState = (FSMStateInterface) stateIter.next();
						if (nextState == s || tmp == s) {
							// if it leads to the deleted state or emanates from the deleted state
							ea.addUndoEvent(FSMEvent.createAddTransitionEvent(tmp, nextState, inp));
							//deletedTransitions.add(new TransitionObj(tmp, st, inp));
						}
					}
				}
			}
		}
		ea.addUndoEvent(FSMEvent.createAddStateEvent(s));
		ea.addRedoEvent(FSMEvent.createDeleteStateEvent(s));
		addUndoEditorAction(ea);
		model.removeState(s);
	}

	public void setStateAccepting(FSMStateInterface s, boolean accepting) {
		EditorAction ea =
			new EditorAction(
				model,
				FSMEvent.createSetAcceptingEvent(s, !accepting),
				FSMEvent.createSetAcceptingEvent(s, accepting));
		addUndoEditorAction(ea);
		s.setAccepting(accepting);
	}

	public void setStateLabel(FSMStateInterface s, String label) {
		EditorAction ea =
			new EditorAction(
				model,
				FSMEvent.createSetStateLabelEvent(s, s.getLabel()),
				FSMEvent.createSetStateLabelEvent(s, label));
		addUndoEditorAction(ea);
		s.setLabel(label);
	}

	public void setStartState(FSMStateInterface s) {
		EditorAction ea =
			new EditorAction(
				model,
				FSMEvent.createSetStartStateEvent(model.getStartState()),
				FSMEvent.createSetStartStateEvent(s));
		addUndoEditorAction(ea);
		model.setStartState(s);
	}

	private void deleteTransition(
		EditorAction ea,
		FSMStateInterface from,
		FSMStateInterface to,
		Object input) {
		ea.addRedoEvent(FSMEvent.createDeleteTransitionEvent(from, to, input));
		ea.addUndoEvent(FSMEvent.createAddTransitionEvent(from, to, input));
		from.removeTransition(input, to);
	}

	private void addTransition(
		EditorAction ea,
		FSMStateInterface from,
		FSMStateInterface to,
		Object input) {
		ea.addRedoEvent(FSMEvent.createAddTransitionEvent(from, to, input));
		ea.addUndoEvent(FSMEvent.createDeleteTransitionEvent(from, to, input));
		from.addTransition(input, to);
	}

	public void addTransition(
		FSMStateInterface from,
		FSMStateInterface to,
		Object input) {
		EditorAction ea = new EditorAction(model);
		addTransition(ea, from, to, input);
		addUndoEditorAction(ea);
	}

	public void deleteTransition(
		FSMStateInterface from,
		FSMStateInterface to,
		Object input) {
		EditorAction ea = new EditorAction(model);
		deleteTransition(ea, from, to, input);
		addUndoEditorAction(ea);
	}

	public void deleteTransitions(FSMStateInterface from, Object input) {
		EditorAction ea = new EditorAction(model);
		Iterator iter = from.getNextStateSet(input).iterator();
		while (iter.hasNext()) {
			deleteTransition(ea, from, (FSMStateInterface) iter.next(), input);
		}
		addUndoEditorAction(ea);
	}

	/** Changes the target state of the specified Transition.
	 */
	public void changeTransition(Transition fat, State sOld, State sNew) {
		EditorAction ea = new EditorAction(model);
		if (sNew != null) {
			Transition already = view.getTransition(fat.from.model, sNew.model);
			if (null != already) {
				for (int i = 0; i < already.label.number(); i++) {
					ea.addUndoEvent(
						FSMEvent.createAddTransitionEvent(
							already.from.model,
							already.to.model,
							already.label.getInput(i)));
				}
			}
		}
		for (int i = 0; i < fat.label.number(); i++) {
			deleteTransition(ea, fat.from.model, sOld.model, fat.label.getInput(i));
		}
		if (null != sNew) {
			for (int i = 0; i < fat.label.number(); i++) {
				// avoid self-epsilon transitions
				if (!(model.getAlphabet() != null
					&& fat.from == sNew
					&& Alphabet.EPSILON.equals(fat.label.getInput(i)))) {
					addTransition(ea, fat.from.model, sNew.model, fat.label.getInput(i));
				}
			}
		}

		addUndoEditorAction(ea);
	}

	/** Replaces inputs of the specified transition with the specified inputs.
	*/
	public void setTransitionInput(Transition t, Object[] c) {
		if (t != null && c != null) {
			EditorAction ea = new EditorAction(model);
			// add inputs to newly drawn transition
			if (t.label.number() < 1) {
				addTransitionInput(t, c);
			}
			// replace labels of existing transition
			else {
				Object mark = t.mark;
				int n = t.label.number();
				for (int i = 0; i < n; i++) {
					deleteTransition(ea, t.from.model, t.to.model, t.label.getInput(0));
				}
				for (int i = 0; i < c.length; i++) {
					addTransition(ea, t.from.model, t.to.model, c[i]);
				}
				addUndoEditorAction(ea);
				Transition after = view.getTransition(t.from, t.to);
				if (after != null)
					after.mark = mark;
			}
		}
	}

	/** Adds the specified inputs to the specified transition. */
	public void addTransitionInput(Transition t, Object[] c) {
		if (t != null && c != null) {
			EditorAction ea = new EditorAction(model);
			for (int i = 0; i < c.length; i++) {
				addTransition(ea, t.from.model, t.to.model, c[i]);
			}
			addUndoEditorAction(ea);
		}
	}

	/** Adds the specified inputs to the specified transition. */
	public void addTransitions(List trans) {
		if (trans != null) {
			EditorAction ea = new EditorAction(model);
			for (int i = 0; i < trans.size(); i += 3) {
				addTransition(
					ea,
					(FSMStateInterface) trans.get(i),
					(FSMStateInterface) trans.get(i + 1),
					(Character) trans.get(i + 2));
			}
			addUndoEditorAction(ea);
		}

	}

	public void setFAName(String name) {
		EditorAction ea =
			new EditorAction(
				model,
				FSMEvent.createSetFSMNameEvent(model.getName()),
				FSMEvent.createSetFSMNameEvent(name));
		addUndoEditorAction(ea);
		model.setName(name);
	}

	public void clear() {
		EditorAction ea =
			new EditorAction(
				model,
				FSMEvent.createResetModelEvent(model.getAllStates()),
				FSMEvent.createClearAllEvent());
		model.clear();
		addUndoEditorAction(ea);
	}

	public void markState(FSMStateInterface state, Object mark) {
		Hashtable ht = new Hashtable();
		ht.put(state, mark);
		view.notify(FSMEvent.createMarkEvent(ht));
	}

	public void markTransition(
		FSMStateInterface from,
		FSMStateInterface to,
		Object mark) {
		Hashtable ht = new Hashtable();
		Object[] trans = { from, to };
		ht.put(trans, mark);
		view.notify(FSMEvent.createMarkEvent(ht));
	}

	public void undo() {
		if (!undoList.empty()) {
			redoList.push(((EditorAction) undoList.peek()).getRedoEditorAction());
			((EditorAction) undoList.pop()).undo();
		} 
	}

	public void redo() {
		if (!redoList.empty()) {
			undoList.push(((EditorAction) redoList.peek()).getRedoEditorAction());
			((EditorAction) redoList.pop()).undo();
		} 
	}

	public void addUndoEditorAction(EditorAction ea) {
		undoList.push(ea);
		redoList.clear();
	}

	public boolean undoListEmpty() {
		return undoList.empty();
	}

	public boolean redoListEmpty() {
		return redoList.empty();
	}

}
