/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.awt.Color;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.JPanel;

/**
 * A FSMView Object manages the visible representatives of states and
 * transitions.
 * 
 * It implements the method notify(FSMEvent) from the interface FSMModelListener
 * and performs model changes on the view.
 * 
 * It contains methods to optimize diagram layout such as - freePoint(): -
 * labelfreePoint(): - fitIntoGrid():
 * 
 * 
 * @author Reto Lamprecht
 */
public class FSMView extends JPanel implements FSMModelListener {

	/**
	 * PAULI_PRINCIPLE = true means: states can't overlap.
	 */
	private final boolean PAULI_PRINCIPLE = true;

	/**
	 * LIVE_UPDATE = false suppresses repaint upon changes in the method
	 * updateView(). Accessible via setLiveUpdate(boolean).
	 */
	private boolean LIVE_UPDATE = true;

	/**
	 * If you want to fit a diagram but the size of the frame is not yet
	 * determined, this flag is set to true in the fitSilently method. In the
	 * method drawAll fitSilently() is called again then.
	 */
	private boolean FIT_OWED = false;

	/**
	 * Flag to turn alignment with grid on and off.
	 */
	private boolean FIT_INTO_GRID = false;

	/**
	 * True if the topology of the diagram changed. By either moving states or
	 * changing transitions. Some layout calculations only need to be done if
	 * this flag is true. It is set to false after these calculations are
	 * executed in the method FSMView.drawAll().
	 */
	public boolean TOPOLOGY_CHANGED = true;

	/**
	 * Will be set to true in the first call of drawAll. Because repaint
	 * requests are to slow when adopting a new model, this flag turns
	 * TOPOLOGY_CHANGED to true as long as the new diagram has not been drawn
	 * the first time.
	 */
	public boolean DRAWN_FIRST_TIME = false;

	/**
	 * REPAINT_SOPHISTICATED = true turns sophisticated placement of transition
	 * labels on. This flag is set to true in the notify method.
	 */
	public boolean REPAINT_SOPHISTICATED = true;

	/**
	 * REPAINT_SOPHISTICATED_ALWAYS = true turns REPAINT_SOPHISTICATED on for
	 * every repaint, also temporary (mouse draging). This flag is set to false,
	 * if the FA gets too large to avoid delays.
	 */
	public boolean REPAINT_SOPHISTICATED_ALWAYS = true;

	/**
	 * If the FA has more than repaintSophisticatedTransLimit transitions,
	 * sophisticated repainting is only done REPAINT_SOPHISTICATED is set to
	 * true;
	 */
	int repaintSophisticatedTransLabelLimit = 12;

	/**
	 * The ch.ethz.ifa.model.FA this FSMView represents.
	 */
	private FSMModelInterface model;

	// ////////////////////////
	// The visible elements //

	/**
	 * All States.
	 */
	public List states = new ArrayList();

	/**
	 * Mapping of ch.ethz.ifa.model.States to States.
	 */
	private Hashtable stateMap = new Hashtable();

	/**
	 * All Transitions.
	 */
	private List transitions = new ArrayList();

	/** The currently selected State. */
	private State selectedState = null;

	/** The currently selected Transitions. */
	private Transition selectedTransition = null;

	// /** The currently highlighted State. */
	// private State highlightedState = null;

	// /** The currently highlighted Transitions. */
	// private Transition highlightedTransition = null;

	// ////////////////////////

	// //////////////////
	// Renderer stuff //

	/**
	 * The Object to paint states, transitions and transitionlabels in a certain
	 * style.
	 */
	private ComponentRenderer componentRenderer;

	/**
	 * A Hashtable that maps Strings to Renderers. Renderers can be chosen by
	 * calling chooseRenderer.
	 */
	private Hashtable renderers = new Hashtable();

	/**
	 * Adds a Renderer to the views library of renderers. The renderer can be
	 * chosen by calling chooseRenderer with its key-String.
	 */
	public void registerRenderer(String key, ComponentRenderer renderer) {
		renderers.put(key, renderer);
	}

	/** Installs the specified ComponentRenderer. */
	public void chooseRenderer(String key) {
		this.componentRenderer = (ComponentRenderer)renderers.get(key);
		radDist = fillDistArray((int) componentRenderer.getStateDiameter(0.0)
				/ 2 + labelToStateDist, radDist);
		for (int i = 0; i < states.size(); i++) {
			((State) states.get(i)).setFont(null);
		}
		TOPOLOGY_CHANGED = true;
	}

	/** Returns the key-String of the installed Renderer. */
	public String getChosenRendererKey() {
		return componentRenderer.getKey();
	}

	public ComponentRenderer getRenderer(String key) {
		if (key != null) {
			return (ComponentRenderer) renderers.get(key);
		}
		return null;
	}

	// //////////////////

	// /////////////////////////
	// FSMViewListener Stuff //

	private List fsmViewListeners = new ArrayList();

	/**
	 * adds a FAscapeListener Object to receive state- and transition selection
	 * events.
	 */
	public void addFSMViewListener(FSMViewListener fl) {
		fsmViewListeners.add(fl);
	}

	/**
	 * adds a FAscapeListener Object to receive state- and transition selection
	 * events.
	 */
	public void removeFSMViewListener(FSMViewListener fl) {
		fsmViewListeners.remove(fl);
	}

	private void notifyStateSelection(State s) {
		for (int i = 0; i < fsmViewListeners.size(); i++) {
			((FSMViewListener) fsmViewListeners.get(i)).stateSelected(s);
		}
	}

	// /////////////////////////

	// ////////////////////
	// Layout paramters //

	/** Bounds of the diagram. */
	private Rectangle faBounds = new Rectangle(50, 50, 50, 50);

	/** Distance in pixels between FA's bounds and its name */
	int nameDist = 20;

	/** Margin when FA is fit into the window */
	int fitMargin = 10;

	/** minimal distance between states. 2 * suggested Stateradius + 20 pixels */
	public int minStateDist = 2 * 25 + 20;

	/** Sets the minimum distance between states to the specified value. */
	public void setMinStateDist(int d) {
		minStateDist = d;
	}

	/** Returns the minimum distance between states. */
	public int getMinStateDist() {
		return minStateDist;
	}

	/** minimal distance between stateborders and transition labels. */
	int labelToStateDist = 8;

	/** Width of the grid. */
	int gridWidth = minStateDist + 1;

	/**
	 * Sets the width of the grid that states align to to the specified value.
	 * Alignment of states with a grid can be turned on and off by
	 * alignWithGrid(boolean b).
	 */
	public void setGridWidth(int w) {
		gridWidth = w;
	}

	/** Returns the width of the grid that states can be forced to align to. */
	public int getGridWidth() {
		return gridWidth;
	}

	/**
	 * Turns alignment with grid on and off. Width of the grid can be set by
	 * setGridWidth(int w).
	 */
	public void alignWithGrid(boolean b) {
		FIT_INTO_GRID = b;
		if (b)
			rearrangeStates();
	}

	/** Returns true if alignment with grid is turned on. */
	public boolean isGridAligned() {
		return FIT_INTO_GRID;
	}

	// ////////////////////

	// /////////////////////////
	// Translation and scale //

	protected ZoomNDragablePane zdPane = new ZoomNDragablePane();

	/**
	 * Sets the scale to the specified value. scale > 1 magnifies, scale < 1
	 * minimises.
	 */
	public void setScale(double scale) {
		zdPane.setScale(scale);
	}

	/** Sets translation to the specified value. */
	public void setTranslation(double x, double y) {
		zdPane.setTranslation(x, y);
	}

	public void setScaleRef(int x, int y) {
		zdPane.setScaleRef(x, y);
	}

	public void setTranslationRef(int x, int y) {
		zdPane.setTranslationRef(x, y);
	}

	public void scale(int y) {
		zdPane.scale(y);
	}

	public void translate(int x, int y) {
		zdPane.translate(x, y);
	}

	public Point2D getModelCoords(Point2D p, int x, int y) {
		zdPane.getModelCoords(p, x, y);
		return p;
	}

	public Point2D getViewCoords(Point2D p, double x, double y) {
		zdPane.getViewCoords(p, x, y);
		return p;
	}

	// /////////////////////////

	/**
	 * Constructs a FAView representing the specified BasicFA.
	 */
	public FSMView(FSMModelInterface fsm) {
		registerRenderer(Component3DRenderer.key, new Component3DRenderer(this));
		registerRenderer(ComponentContrastRenderer.key,
				new ComponentContrastRenderer(this));
		registerRenderer(ComponentDiodeRenderer.key,
				new ComponentDiodeRenderer(this));
		chooseRenderer(ComponentContrastRenderer.key);
		zdPane.setScaleRange(0.2, 6);
		fillCosSinArrays();
		stateDist = fillDistArray(minStateDist, stateDist);
		radDist = fillDistArray((int) componentRenderer.getStateDiameter(0.0)
				/ 2 + labelToStateDist, radDist);
		setModel(fsm);
	}

	/**
	 * Sets the model that this FSMView represents.
	 * 
	 * @param model
	 *            the specified FA.
	 */
	public void setModel(FSMModelInterface model) {
		if (this.model != null) {
			this.model.removeFSMModelListener(this);
		}
		this.model = model;
		if (this.model != null) {
			model.addFSMModelListener(this);
		}
		adoptFA(model);
	}

	/**
	 * Returns the model that this FSMView represents.
	 */
	public FSMModelInterface getModel() {
		return model;
	}

	/**
	 * Traverses the specified FA and adds the corresponding States and
	 * Transitions to this FSMView.
	 */
	private void adoptFA(FSMModelInterface model_) {
		if (model_ == null)
			return;
		clear();
		LIVE_UPDATE = false;
		DRAWN_FIRST_TIME = false;

		// add states:
		boolean MISSING_STATE_POS = false;
		// if some state has NaN coordinates, the diagram is autolayouted.
		Set sts = model_.getAllStates();
		if (sts == null)
			return;
		Iterator modelStates = sts.iterator();
		while (modelStates.hasNext()) {
			FSMStateInterface s = (FSMStateInterface) modelStates.next();
			if (("" + s.getX()).equals("NaN") || ("" + s.getY()).equals("NaN")) { //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				MISSING_STATE_POS = true;
				s.setPosition(0.0, 0.0);
			}
			addState(s);
		}
		// add transitions
		State currState;
		for (int i = 0; i < states.size(); i++) {
			currState = (State) states.get(i);
			Set inputs = currState.model.getInputs();
			if (null != inputs) {
				Iterator inputIter = inputs.iterator();
				Object currInput;
				while (inputIter.hasNext()) {
					currInput = inputIter.next();
					Iterator nextStateIter = currState.model.getNextStateSet(
							currInput).iterator();
					while (nextStateIter.hasNext()) {
						attachTransition(addTransition(currState, currInput),
								(State) stateMap.get(nextStateIter.next()));
					}
				}
			}
		}
		if (MISSING_STATE_POS) {
			boolean ok = true;
			ok = CircleLayouter.layout(this);
			// circle layouter returns false, if the best arrangement found has
			// more than 1 intersection.
			if (!ok) {
				OrganicLayouter.layout(this);
			}
		} else {
			rearrangeStates();
		}
		selectTransition(null);
		fit();
		setLiveUpdate(true);
	}

	/**
	 * Deletes all States and Transitions.
	 */
	public void clear() {
		stateMap.clear();
		states.clear();
		transitions.clear();
		updateView(true);
	}

	public void notify(FSMEvent e) {
		REPAINT_SOPHISTICATED = true;
		switch (e.getId()) {
		case FSMEvent.STATE_ADDED:
			addState(e.getState());
			break;
		case FSMEvent.STATE_DELETED:
			deleteState(e.getState());
			break;
		case FSMEvent.STATE_POS_CHANGED:
			setStatePosition(e.getState(), e.getState().getX(), e.getState()
					.getY());
			break;
		case FSMEvent.STATE_LABEL_CHANGED:
			setStateLabel(e.getState(), e.getState().getLabel());
			break;
		case FSMEvent.STATE_MODE_CHANGED:
			setStateAccepting(e.getState(), e.getState().isAccepting());
			break;
		case FSMEvent.START_STATE_CHANGED:
			setStartState(model.getStartState());
			break;
		case FSMEvent.CLEAR_ALL:
			clear();
			break;
		case FSMEvent.TRANSITION_ADDED:
			addTransition(e.getInput(), e.getFromState(), e.getToState());
			if (transitions.size() > repaintSophisticatedTransLabelLimit) {
				REPAINT_SOPHISTICATED_ALWAYS = false;
			}
			break;
		case FSMEvent.TRANSITION_DELETED:
			deleteTransition(e.getInput(), e.getFromState(), e.getToState());
			if (transitions.size() < repaintSophisticatedTransLabelLimit) {
				REPAINT_SOPHISTICATED_ALWAYS = true;
			}
			break;
		case FSMEvent.FSM_NAME_CHANGED:
			updateView(false);
			break;
		case FSMEvent.MARK:
			Hashtable marks = e.getMarks();
			Iterator objs = marks.keySet().iterator();
			Object current;
			while (objs.hasNext()) {
				current = objs.next();
				if (current instanceof FSMStateInterface) {
					markState((FSMStateInterface) current, marks.get(current));
				} else if (current instanceof Object[]) {
					Object[] t = (Object[]) current;
					markTransition((FSMStateInterface) t[0],
							(FSMStateInterface) t[1], marks.get(current));
				}
			}
			break;
		case FSMEvent.UNMARK_ALL:
			State s;
			for (int i = 0; i < states.size(); i++) {
				s = (State) states.get(i);
				s.setMark(ComponentRenderer.MARK_NOT_MARKED);
				for (int j = 0; j < s.outTransitions.size(); j++) {
					((Transition) s.outTransitions.get(j))
							.setMark(ComponentRenderer.MARK_NOT_MARKED);
				}
			}
			updateView(false);
		default:
			updateView(false);
		}
	}

	/**
	 * Adds a FSMStateInterface.
	 */
	public void addState(FSMStateInterface s) {
		if (null == s)
			return;
		Rectangle faBounds_ = getFABounds();
		State viewS = new State(s);
		stateMap.put(s, viewS);
		states.add(viewS);
		Point p;
		if (("" + s.getX()).equals("NaN") || ("" + s.getY()).equals("NaN")) { //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
			p = getNearestFreePoint(viewS, faBounds_.x + faBounds_.width / 2,
					faBounds_.y + faBounds_.height / 2);
		} else {
			p = getNearestFreePoint(viewS, (int) s.getX(), (int) s.getY());
		}
		viewS.setCenter(p.x, p.y);
		// s.setPositionSilently((double)p.x, (double)p.y);

		updateView(true);
	}

	/**
	 * Deletes the specified FSMStateInterface.
	 */
	public void deleteState(FSMStateInterface s) {
		if (null == s)
			return;
		State viewS = (State) stateMap.remove(s);
		states.remove(viewS);
		Transition currTrans;
		for (int i = 0; i < transitions.size(); i++) {
			currTrans = (Transition) transitions.get(i);
			if (currTrans.from == viewS || currTrans.to == viewS) {
				removeTransition(currTrans);
				i--;
			}
		}
		updateView(true);
	}

	/**
	 * Sets the specified FSMStateInterface's position. Setters have to make
	 * sure on their own that no overlapping of states occurs.
	 */
	public void setStatePosition(FSMStateInterface s, double x, double y) {
		if (null == s)
			return;
		State viewS = (State) stateMap.get(s);
		if (viewS != null) {
			viewS.setCenter(x, y);
		}
		updateView(true);
	}

	/**
	 * Sets the specified FSMStateInterface's position in view coordinates.
	 * Overlapping is prevented.
	 */
	public void setStateViewPosition(FSMStateInterface s, int x, int y) {
		if (null == s)
			return;
		State viewS = (State) stateMap.get(s);
		if (viewS != null) {
			Point2D p = new Point2D.Double();
			getModelCoords(p, x, y);
			p = getNearestFreePoint(viewS, (int) p.getX(), (int) p.getY());
			viewS.setCenter(p.getX(), p.getY());
			s.setPosition(p.getX(), p.getY());
		}
		updateView(true);
	}

	/**
	 * Sets the specified FSMStateInterface's label.
	 */
	public void setStateLabel(FSMStateInterface s, Object label) {
		if (null == s)
			return;
		((State) stateMap.get(s)).setLabel(label);
		updateView(false);
	}

	/**
	 * The specified FSMStateInterface is turned to an accepting state if
	 * accepting is true and to a non-accepting state if accepting is false.
	 */
	public void setStateAccepting(FSMStateInterface s, boolean accepting) {
		if (null == s)
			return;
		((State) stateMap.get(s)).setAccepting(accepting);
		updateView(false);
	}

	/**
	 * The specified FSMStateInterface is turned to the startstate if start is
	 * true and to a non-startstate if start is false.
	 */
	public void setStartState(FSMStateInterface s) {
		if (null == s)
			return;
		for (int i = 0; i < states.size(); i++) {
			((State) states.get(i)).setStart(false);
		}
		((State) stateMap.get(s)).setStart(true);
		updateView(false);
	}

	/**
	 * Adds a transition.
	 * 
	 * @param input
	 *            The transition's label.
	 * @param from
	 *            The transition's from-state.
	 * @param to
	 *            The transition's to-state.
	 */
	public void addTransition(Object input, FSMStateInterface from,
			FSMStateInterface to) {
		if (null == from || null == to)
			return;
		attachTransition(addTransition((State) stateMap.get(from), input),
				(State) stateMap.get(to));
		updateView(true);
	}

	/**
	 * Removes a transition.
	 * 
	 * @param input
	 *            The transition's label.
	 * @param from
	 *            The transition's from-state.
	 * @param to
	 *            The transition's to-state.
	 */
	public void deleteTransition(Object input, FSMStateInterface from,
			FSMStateInterface to) {
		if (null == from || null == to)
			return;
		List ot = ((State) stateMap.get(from)).outTransitions;
		for (int i = 0; i < ot.size(); i++) {
			Transition currTrans = (Transition) ot.get(i);
			if (currTrans.to.model == to) {
				currTrans.label.removeInput(input);
				if (null == currTrans.label.getLabelString()) {
					transitions.remove(currTrans);
					currTrans.from.removeOutTransition(currTrans);
					currTrans.to.removeInTransition(currTrans);
				}
			}
		}
		updateView(true);
	}

	/** Marks the specified FSMStateInterface */
	public void markState(FSMStateInterface s, Object mark) {
		if (null == s)
			return;
		State viewS;
		viewS = (State) stateMap.get(s);
		if (null != viewS) {
			viewS.setMark(mark);
		}
		updateView(false);
	}

	/** Marks the specified transition. */
	public void markTransition(FSMStateInterface from, FSMStateInterface to,
			Object mark) {
		if (from != null && to != null) {
			Transition fat = getTransition((State) stateMap.get(from),
					(State) stateMap.get(to));
			if (null != fat) {
				fat.setMark(mark);
				updateView(false);
			}
		}
	}

	/**
	 * Determines the smallest label of the form 'qn' that is not used as
	 * statelabel already.
	 */
	public String smallestFreeQLabel() {
		// determine existing labels
		int[] exist = new int[states.size() + 1];
		int c;
		for (int i = 0; i < states.size(); i++) {
			c = states.size();
			try {
				c = Integer.parseInt(((State) states.get(i)).getLabel()
						.substring(1));
			} catch (Exception e) {
			}
			if (c < states.size())
				exist[c] = 1;
		}
		for (int i = 0; i < states.size(); i++) {
			if (exist[i] == 0)
				return "q" + i; //$NON-NLS-1$
		}
		return "q" + states.size(); //$NON-NLS-1$
	}

	/**
	 * Returns a label that is different to all other statelabels by adding
	 * '#''s to the specified label if needed.
	 * 
	 * @param label
	 *            The suggested statelabel.
	 * @return The unique statelabel.
	 */
	public String uniqueStateLabel(String label) {
		State current = null;
		for (int i = 0; i < states.size(); i++) {
			current = (State) states.get(i);
			if (current.getLabel().equals(label)) {
				return uniqueStateLabel("#" + label); //$NON-NLS-1$
			}
		}
		return label;
	}

	public void rearrangeStates() {
		Object[] tmpStates = states.toArray();
		states.clear();
		FSMStateInterface current;
		Point p;
		for (int i = 0; i < tmpStates.length; i++) {
			states.add(tmpStates[i]);
			current = ((State) tmpStates[i]).model;
			p = getNearestFreePoint((State) stateMap.get(current),
					(int) current.getX(), (int) current.getY());
			if (null != p) {
				if ((int) current.getX() != p.x || (int) current.getY() != p.y) {
					current.setPosition(p.x, p.y);
				}
			}
		}
		fit();
		updateView(true);
	}

	public boolean onMouseOver(State state, int x, int y) {
		return componentRenderer.onMouseOver(state, x, y);
	}

	public void grabState(State s, double x, double y) {
		for (int i = 0; i < states.size(); i++) {
			((State) states.get(i)).release();
		}
		s.grab(x, y);
	}

	public void selectState(State s, boolean selected) {
		selectedState = s;
		if (selected) {
			for (int i = 0; i < states.size(); i++) {
				((State) states.get(i)).setSelected(false);
			}
			notifyStateSelection(s);
		}
		if (s != null) {
			selectedState.setSelected(selected);
		}
	}

	public State getSelectedState() {
		return selectedState;
	}

	public Transition getSelectedTransition() {
		return selectedTransition;
	}

	/**
	 * Dehighlights all States' centers and borders.
	 */
	public void dehighlightStates() {
		State s;
		for (int i = 0; i < states.size(); i++) {
			s = (State) states.get(i);
			s.setBorderHighlighted(false);
			s.setCenterHighlighted(false);
			s.SHOW_WHOLE_LABEL = false;
		}
		updateView(false);
	}

	/**
	 * Dehighlights all Transitions.
	 */
	public void dehighlightTransitions() {
		Transition t;
		for (int i = 0; i < transitions.size(); i++) {
			t = (Transition) transitions.get(i);
			t.setHighlighted(false);
		}
		updateView(false);
	}

	/**
	 * Gets the first State that contains the point (x, y).
	 * 
	 * @param x
	 *            The x coordinate of the specified point.
	 * @param y
	 *            The y coordinate of the specified point.
	 * @return the State found.
	 */
	public State getStateAt(double x, double y) {
		State current = null, s = null;
		for (int i = 0; i < states.size(); i++) {
			current = (State) states.get(i);
			if (componentRenderer.contains(current, x, y))
				s = current;
		}
		return s;
	}

	public State getState(FSMStateInterface state) {
		if (state == null)
			return null;
		return (State) stateMap.get(state);
	}

	public State getState(int i) {
		return (State) states.get(i);
	}

	public int getNofStates() {
		return stateMap.size();
	}

	protected Object[] getStates() {
		return states.toArray();
	}

	public int getNofTransitions() {
		return transitions.size();
	}

	public Transition getTransition(int i) {
		return (Transition) transitions.get(i);
	}

	protected void calcTransitionCoords(Transition t) {
		componentRenderer.calcTransitionCoords(t);
	}

	/**
	 * Moves the center of the grabbed State to (x, y) - (mouseRefX, mouseRefY).
	 * I.e. the arguments are treated as mousepointer-coordinates, not as
	 * state-center-positions.
	 * 
	 * @param x
	 *            The x coordinate of the State's new center.
	 * @param y
	 *            The y coordinate of the State's new center.
	 */
	public void moveState(State movingState, double x, double y) {
		if (null != movingState) {
			// calc center position of state
			double cx = x - movingState.mouseRefX;
			double cy = y - movingState.mouseRefY;
			if (PAULI_PRINCIPLE) {
				Point p = getNearestFreePoint(movingState, (int) cx, (int) cy);
				if (null != p) {
					movingState.setCenter(p.x, p.y);
				}
			} else {
				movingState.setCenter(cx, cy);
			}
			updateView(true);
		}
	}

	/**
	 * adds a Transition originating from the specified State.
	 * 
	 * @param from
	 *            The from-state.
	 * @param label
	 *            The label.
	 * @return the Transition that was added.
	 */
	public Transition addTransition(State from, Object input) {
		Transition fat = null;
		if (null != from) {
			fat = new Transition(from, input);
			transitions.add(fat);
			from.addOutTransition(fat);
			selectTransition(fat);
			updateView(true);
		}
		return fat;
	}

	/**
	 * adds a Transition originating from the specified State. The input is the
	 * smallest element of the FA's Alphabet that is not used already. If all
	 * elements are used, no Transition is added.
	 * 
	 * @param from
	 *            The from-state.
	 * @return the Transition that was added or null if all elements of the FA's
	 *         Alphabet are used.
	 */
	public Transition addTransition(State from) {
		return addTransition(from, null);
	}

	/**
	 * Removes a Transition
	 * 
	 * @param fat
	 *            the specified Transition.
	 */
	public void removeTransition(Transition fat) {
		if (null != fat) {
			fat.from.removeOutTransition(fat);
			if (null != fat.to) {
				fat.to.removeInTransition(fat);
			}
			transitions.remove(fat);
			updateView(true);
		}
	}

	/**
	 * Selects the specified Transition and deselects every other Transition.
	 * 
	 * @param fat
	 *            The Transitions to be selected or null to deselect all
	 *            Transitions.
	 */
	public void selectTransition(Transition fat) {
		selectedTransition = fat;
		for (int i = 0; i < transitions.size(); i++) {
			((Transition) transitions.get(i)).setSelected(false);
		}
		if (null != fat) {
			fat.setSelected(true);
		}
		updateView(false);
	}

	/**
	 * Returns the transition originating from State from and pointing to State
	 * to or null, if there is no transition.
	 * 
	 * @param from
	 *            The from-state.
	 * @param to
	 *            The to-state.
	 * @returns The found Transition from from to to.
	 */
	public Transition getTransition(FSMStateInterface from, FSMStateInterface to) {
		if (from != null && to != null) {
			return getTransition((State) stateMap.get(from), (State) stateMap
					.get(to));
		}
		return null;
	}

	/**
	 * Returns the transition originating from State from and pointing to State
	 * to or null, if there is no transition.
	 * 
	 * @param from
	 *            The from-state.
	 * @param to
	 *            The to-state.
	 * @returns The found Transition from from to to.
	 */
	public Transition getTransition(State from, State to) {
		if (from != null && to != null) {
			for (int i = 0; i < from.outTransitions.size(); i++) {
				if (((Transition) from.outTransitions.get(i)).to == to) {
					return (Transition) from.outTransitions.get(i);
				}
			}
		}
		return null;
	}

	/**
	 * Gets the Transition whose head is hit by point (x, y).
	 * 
	 * @param x
	 *            The x coordinate of the specified point.
	 * @param y
	 *            The y coordinate of the specified point.
	 * @return The Transition found or null if no Transition's head is at (x,
	 *         y).
	 */
	public Transition getTransitionHeadAt(double x, double y) {
		Transition current = null, sel = null;
		for (int i = 0; i < transitions.size(); i++) {
			current = (Transition) transitions.get(i);
			if (componentRenderer.contains(current, x, y))
				sel = current;
		}
		return sel;
	}

	/**
	 * Gets the Transition whose Label is hit by point (x, y).
	 * 
	 * @param x
	 *            The x coordinate of the specified point.
	 * @param y
	 *            The y coordinate of the specified point.
	 * @return The Transition found or null if no Transition's Label is at (x,
	 *         y).
	 */
	public Transition getTransitionLabelAt(double x, double y) {
		Transition current = null, sel = null;
		for (int i = 0; i < transitions.size(); i++) {
			current = (Transition) transitions.get(i);
			if (componentRenderer.contains(current.label, x, y))
				sel = current;
		}
		return sel;
	}

	/**
	 * Gets the Transition whose Label or Head is hit by point (x, y).
	 * 
	 * @param x
	 *            The x coordinate of the specified point.
	 * @param y
	 *            The y coordinate of the specified point.
	 * @return The Transition found or null if no Transition's Label or Head is
	 *         at (x, y).
	 */
	public Transition getTransitionAt(int x, int y) {
		Transition sel = getTransitionLabelAt(x, y);
		if (null == sel) {
			return getTransitionHeadAt(x, y);
		}
		return sel;
	}

	/**
	 * Moves a Transition's head to the position (x, y). If the Transition was
	 * attached to a State, this State is removed.
	 * 
	 * @param x
	 *            the x coordinate of the head's center
	 * @param y
	 *            the y coordinate of the head's center
	 * @param fat
	 *            the <code>Transition</code>
	 */
	public void moveTransition(double x, double y, Transition fat) {
		if (null != fat) {
			if (null != fat.to) {
				fat.to.removeInTransition(fat);
				fat.setToState(null);
			}
			fat.setTo(x, y);
			updateView(true);
		}
	}

	/**
	 * Attaches a Transition to a State that becomes this Transition's to-state.
	 * If a Transition with identical from- and to-states already exists, no new
	 * Transition is added but the specified Transition's Label is added to the
	 * existing one's Label.
	 * 
	 * @param fat
	 *            The Transition to attach.
	 * @param to
	 *            The State that fat will point to.
	 */
	public void attachTransition(Transition fat, State to) {
		if (null == to) {
			removeTransition(fat);
		} else if (null != fat) {
			to.addInTransition(fat);
			// check if an identical transition already exists. if so add
			// existing labels to new transition.
			// (its essential to keep the new transition and delete the existing
			// one,
			// because the existing one is often the temporary transition
			// that the controller adds as apreview and that is deleted after
			// this method call.)
			Transition current;
			for (int i = 0; i < fat.from.outTransitions.size(); i++) {
				current = (Transition) fat.from.outTransitions.get(i);
				if (to == current.to) {
					current.from.removeOutTransition(current);
					to.removeInTransition(current);
					transitions.remove(current);
					fat.label.addFATLabel(current.label);
					break;
				}
			}
			fat.setToState(to);
			selectTransition(null);
		}
		updateView(true);
	}

	/**
	 * Puts the specified Transition on top of all other transitions.
	 */
	public void toTop(Transition fat) {
		if (null != fat) {
			transitions.remove(fat);
			transitions.add(fat);
		}
	}

	public State getStartState() {
		if (null != model.getStartState()) {
			return (State) stateMap.get(model.getStartState());
		}
		return null;
	}

	public double getStateDiameter() {
		return componentRenderer.getStateDiameter(0.0);
	}

	/**
	 * Returns the bounds of the FA currently displayed.
	 */
	Rectangle getFABounds() {
		if (states != null && states.size() > 0) {
			State current = (State) states.get(0);
			Rectangle2D currBounds = componentRenderer.getStateBounds(current);
			double minX = currBounds.getX();
			double maxX = minX + currBounds.getWidth();
			double minY = currBounds.getY();
			double maxY = minY + currBounds.getHeight();
			for (int i = 1; i < states.size(); i++) {
				current = (State) states.get(i);
				currBounds = componentRenderer.getStateBounds(current);
				minX = Math.min(minX, currBounds.getX());
				maxX = Math
						.max(maxX, currBounds.getX() + currBounds.getWidth());
				minY = Math.min(minY, currBounds.getY());
				maxY = Math.max(maxY, currBounds.getY()
						+ currBounds.getHeight());
			}
			faBounds.setRect((int) minX, (int) minY, (int) (maxX - minX),
					(int) (maxY - minY));
			return faBounds;
		}
		faBounds
				.setRect(
						(int) (zdPane.getTranslationX() * zdPane.getScale() + getWidth() / 2),
						(int) (zdPane.getTranslationY() * zdPane.getScale() + getHeight() / 2),
						0, 0);
		return faBounds;
	}

	/**
	 * Scales the FA to fit into the window and centers it.
	 */
	public void fit() {
		fitSilently();
		updateView(true); // usually after layouting
	}

	public void fitSilently() {
		Rectangle faBounds_ = getFABounds();
		if (faBounds_.width > 0) {
			double minX = faBounds_.x - fitMargin;
			double maxX = faBounds_.x + faBounds_.width + fitMargin;
			double minY = faBounds_.y - fitMargin;
			double maxY = faBounds_.y + faBounds_.height + fitMargin + nameDist;
			zdPane.setScale(Math.min(getWidth() / (maxX - minX), getHeight()
					/ (maxY - minY)));
			zdPane.setScale(Math.min(1.0, zdPane.getScale()));
			zdPane.setTranslation(-minX * zdPane.getScale()
					+ (getWidth() - (maxX - minX) * zdPane.getScale()) / 2,
					-minY * zdPane.getScale()
							+ (getHeight() - (maxY - minY) * zdPane.getScale())
							/ 2);
			if (zdPane.getScale() == 0)
				FIT_OWED = true;
			else
				FIT_OWED = false;
		}
	}

	private Color bgrColor = Color.white;

	public Color getMyBackground() {
		return this.bgrColor;
	}

	public void setMyBackground(Color bgrColor) {
		this.bgrColor = bgrColor;
	}

	/**
	 * Sets LIVE_UPDATE.
	 */
	public void setLiveUpdate(boolean liveUpdate) {
		LIVE_UPDATE = liveUpdate;
		updateView(true);
	}

	/**
	 * Initiates repaint if LIVE_UPDATE is true.
	 */
	public void updateView(boolean topologyChanged) {
		if (topologyChanged)
			TOPOLOGY_CHANGED = true;
		if (LIVE_UPDATE) {
			repaint();
		}
	}

	public void paint(Graphics g) {
		paintComponent(g);
	}

	/**
	 * Paints this component.
	 */
	public void paintComponent(Graphics g) {
		drawAll(g);
	}

	/**
	 * Draws all States first, then all Transitions, then all TransitionLabels
	 * and then marks activeState.
	 */
	private void drawAll(Graphics g) {
		super.paintComponent(g);
		componentRenderer.drawBackground(getWidth(), getHeight(),
				(Graphics2D) g);

		if (FIT_OWED) {
			fitSilently();
		}

		Graphics2D g2 = (Graphics2D) g;
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);

		zdPane.transform(g2);

		if (!DRAWN_FIRST_TIME) {
			TOPOLOGY_CHANGED = true;
			DRAWN_FIRST_TIME = true;
		}

		if (REPAINT_SOPHISTICATED_ALWAYS)
			REPAINT_SOPHISTICATED = true;

		// paint States
		for (int i = 0; i < states.size(); i++) {
			componentRenderer.drawState((State) states.get(i), g2);
		}

		// paint Transitions
		for (int i = 0; i < transitions.size(); i++) {
			componentRenderer.drawTransition((Transition) transitions.get(i),
					g2);
		}

		// paint CUBIC TransitionLabels
		Transition current;
		for (int i = 0; i < transitions.size(); i++) {
			current = (Transition) transitions.get(i);
			if (current.type == Transition.CUBIC) {
				componentRenderer.drawTransitionLabel(((Transition) transitions
						.get(i)).label, g2);
			}

		}
		fillStateArray(null);
		// paint LINE & QUAD TransitionLabels (they can adapt their position to
		// avoid overlapping
		for (int i = 0; i < transitions.size(); i++) {
			current = (Transition) transitions.get(i);
			if (current.type == Transition.LINE
					|| current.type == Transition.QUAD) {
				componentRenderer.drawTransitionLabel(((Transition) transitions
						.get(i)).label, g2);
			}
		}
		// draw marks on top
		for (int i = 0; i < states.size(); i++) {
			componentRenderer.markState((State) states.get(i), g2);
		}

		// draw FA name
		if (model != null) {
			FontMetrics fm = g2.getFontMetrics();
			int faNameW = fm.stringWidth(model.getName());
			Rectangle faBounds_ = getFABounds();
			g2.setColor(componentRenderer.getTextColor());
			g2.drawString(model.getName(), faBounds_.x + faBounds_.width / 2
					- faNameW / 2, faBounds_.y + faBounds_.height + nameDist);
		}

		TOPOLOGY_CHANGED = false;
		REPAINT_SOPHISTICATED = false;

	}

	/**
	 * Lets give the garbage collector a break! getNearestFreePoint() and
	 * gridFit() use this Point to return their results.
	 */
	private Point deliveryPoint = new Point();

	/**
	 * Returns the point with minimal distance from the specified point that has
	 * distance > stateradius from every non-grabbed state. This point is found
	 * by testing freePoint(x, y) for every point on a circle around the
	 * specified Point with this circle's radius growing from 1 to the most
	 * distant corner. Works only for round states.
	 * 
	 * @param x
	 *            The x coordinate of the specified point.
	 * @param y
	 *            The y coordinate of the specified point.
	 * @return The nearest free Point.
	 */
	public final Point getNearestFreePoint(State movingState, int x, int y) {

		fillStateArray(movingState);

		if (FIT_INTO_GRID) {
			Point p = gridFit(gridWidth, x, y);
			if (freePoint(p.x, p.y))
				return p;
			Rectangle bounds = getFABounds();
			int lim = (Math.max(bounds.width, bounds.height) / gridWidth) + 2;
			double phi;
			for (int r = 1; r < lim; r++) {
				int trys = (int) (Math.PI * 4 * Math.sqrt(2) * r);
				for (int i = 0; i < trys; i++) {
					phi = i * Math.PI * 2 / (trys);
					p = gridFit(gridWidth, (int) (x + cos[(int) (phi * 100)]
							* r * gridWidth), (int) (y + sin[(int) (phi * 100)]
							* r * gridWidth));
					if (freePoint(p.x, p.y)) {
						return p;
					}
				}
			}
			deliveryPoint.setLocation(x, y);
			return deliveryPoint;

		}
		if (stateDist.length != minStateDist * minStateDist) {
			stateDist = fillDistArray(minStateDist, stateDist);
		}
		int tx = x;
		int ty = y;
		int w = getWidth();
		int h = getHeight();

		if (freePoint(x, y)) {
			return new Point(x, y);
		}
		// max. distance from (x, y) to corners:
		int lim = (int) (Math.sqrt(Math.max(w - x, x)
				* Math.max(w - x, x) + Math.max(h - y, y)
				* Math.max(h - y, y)));
		double phi;
		for (int r = 1; r < lim; r++) {
			for (int i = 0; i < (int) (2 * r * Math.PI); i++) { // ~number
				// of
				// pixels
				// on
				// circumference
				phi = (double) i / r; // phi goes from 0 to 2PI.
				tx = (int) (x + cos[(int) (phi * 100)] * r);
				ty = (int) (y + sin[(int) (phi * 100)] * r);
				if (freePoint(tx, ty)) {
					return new Point(tx, ty);
				}
			}
		}
		deliveryPoint.setLocation(x, y);
		return deliveryPoint;
	}

	/**
	 * Returns the nearest gridPoint from (x, y) on the grid with the specified
	 * width.
	 */
	public final Point gridFit(int gridWidth_, int x, int y) {
		double origX = 0.0;
		double origY = 0.0;
		double adjust;
		double prefixX, prefixY;
		if (x < origX)
			prefixX = -1.0;
		else
			prefixX = 1.0;
		if (y < origY)
			prefixY = -1.0;
		else
			prefixY = 1.0;
		adjust = Math.abs((x - origX) % gridWidth_);
		if (adjust > 0.0) {
			if (adjust < gridWidth_ / 2) {
				x -= prefixX * adjust;
			} else {
				x += prefixX * (gridWidth_ - adjust);
			}
		}
		adjust = Math.abs((y - origY) % gridWidth_);
		if (adjust > 0.0) {
			if (adjust < gridWidth_ / 2) {
				y -= prefixY * adjust;
			} else {
				y += prefixY * (gridWidth_ - adjust);
			}
		}
		deliveryPoint.setLocation(x, y);
		return deliveryPoint;
	}

	/**
	 * Returns true if (x, y) has distance > minStateDist from every non-grabbed
	 * state. To avoid states to overlap, they have to be placed at positions
	 * (x, y) where freePoint(x, y) is true.
	 */
	private final boolean freePoint(int x, int y) {
		double dist = minStateDist;
		double dx = 0, dy = 0;
		for (int i = 0; i < statesX.length; i++) {
			if (i != movingStateIndex) {
				if (x > statesX[i])
					dx = x - statesX[i];
				else
					dx = statesX[i] - x;
				if (y > statesY[i])
					dy = y - statesY[i];
				else
					dy = statesY[i] - y;
				if (dx < dist && dy < dist
						&& !stateDist[(int) (dx * dist + dy)])
					return false;
			}
		}
		return true;
	}

	/**
	 * Returns true if the specified transitionlabel doesn't overlap any state.
	 */
	final boolean labelStateFreePoint(TransitionLabel tl) {
		double dist = componentRenderer.getStateDiameter(0.0) / 2
				+ labelToStateDist;
		double dx = 0, dy = 0;
		// stateArray gets filled in drawAll because ist used for fitting states
		// there first
		for (int i = 0; i < statesX.length; i++) {
			if (tl.x > statesX[i])
				dx = (int) (tl.x - statesX[i]);
			else
				dx = (int) (statesX[i] - tl.x);
			if (tl.y > statesY[i])
				dy = (int) (tl.y - statesY[i]);
			else
				dy = (int) (statesY[i] - tl.y);
			if (dx < dist && dy < dist && !radDist[(int) (dx * dist + dy)]) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns true if the specified transitionlabel doesn't overlap any other
	 * label nor any transition that is not between the same states as this
	 * label's transition.
	 */
	final boolean labelTransFreePoint(TransitionLabel tl) {
		Transition current;
		for (int i = 0; i < transitions.size(); i++) {
			current = (Transition) transitions.get(i);
			// dont check identical labels nor labels of transitions with same
			// from- and to-states.
			if (current.label != tl
					&& !(tl.fat.from == current.from && tl.fat.to == current.to)) {
				// check label
				if (componentRenderer.getTransitionLabelBounds(tl).intersects(
						componentRenderer
								.getTransitionLabelBounds(current.label))) {
					return false;
				}
				// check transition line if transitions are not attached to the
				// same state
				else if (tl.fat.from != current.from
						&& tl.fat.from != current.to
						&& tl.fat.to != current.to
						&& tl.fat.to != current.from
						&& current.getCurve().intersects(
								componentRenderer.getTransitionLabelBounds(tl))) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Arrays to store the states' coordinates to speed up the method
	 * freePoint(x, y).
	 */
	private int[] statesX, statesY;

	/** The index of statesX and statesY where the grabbed State is stored. */
	private int movingStateIndex;

	/**
	 * This method fills the arrays statesX and statesY with the positions of
	 * all states. This is done for performance reasons: It avoids iterating
	 * through a Vector and access Object members in freePoint(x, y). The method
	 * has to be called whenever state positions changed or a different state is
	 * moved.
	 */
	private final void fillStateArray(State movingState) {
		State current;
		statesX = new int[states.size()];
		statesY = new int[states.size()];
		movingStateIndex = -1;
		for (int i = 0; i < states.size(); i++) {
			current = (State) states.get(i);
			if (current == movingState)
				movingStateIndex = i;
			statesX[i] = (int) current.getCenterX();
			statesY[i] = (int) current.getCenterY();
		}
	}

	/**
	 * This array will be filled with booleans such that the "false" elements
	 * form a quarter of a circle with r = minStateDist.
	 */
	private boolean stateDist[] = null;

	/** Same as stateDist, but the circle has r = statediameter. */
	private boolean radDist[] = null;

	/**
	 * This method fills the specified array (either radDist or minStateDist)
	 * with booleans Array elements - arranged as a square with sidelength =
	 * dist correspond to pixels. Elements that are set to "false" form a
	 * quarter of a circle. "true" means: the pixel (x, y) is outside the circle
	 * with r = dist and origin = (0,0).
	 * 
	 * e.g.: [00000 00000 00001 00011 01111]
	 */
	private final boolean[] fillDistArray(int dist, boolean[] distArray) {
		int w = dist;
		int h = dist;
		int distSquare = dist * dist;
		distArray = new boolean[w * h];
		for (int j = 0; j < h; j++) {
			for (int i = 0; i < w; i++) {
				distArray[j * w + i] = ((i * i + j * j) > distSquare) ? true
						: false;
			}
		}
		return distArray;
	}

	/** Precalculated sin and cos values to speed up the method freePoint(x, y). */
	private double[] cos, sin;

	/**
	 * Precalculates sin and cos values for angles in [0..2PI] to speed up
	 * freePoint(x, y). Each Array contains 200PI values, so cos(phi) can be
	 * obtained by cos[(int)(phi*100)]
	 */
	private final void fillCosSinArrays() {
		cos = new double[(int) (2 * Math.PI * 100)];
		sin = new double[(int) (2 * Math.PI * 100)];
		for (int i = 0; i < sin.length; i++) {
			cos[i] = Math.cos((double) i / 100);
			sin[i] = Math.sin((double) i / 100);
		}
	}

}