/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Rectangle;
import java.awt.Shape;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import javax.swing.JComponent;


/** An State Object represents a ch.ethz.ifa.model.State,
 *  which is referenced by it's member <code>model</code>.
 *
 * @author		Reto Lamprecht
 */
public class State extends JComponent {

	/** Center of State is highlighted. */
	boolean HIGHLIGHT_CENTER = false;

	/** Border of State is highlighted. */
	boolean HIGHLIGHT_BORDER = false;

	/** This State is selected. */
	boolean SELECTED = false;

	/** This state's marks. */
	public Object mark = ComponentRenderer.MARK_NOT_MARKED;

	/** Indicates if this state is an accepting state.
	 */
	boolean ACCEPTING = false;

	/** Indicates if this state is the start state.
	 */
	boolean START = false;

	/** The state's label. */
	String label;
	String visibleLabel;

	// todo: accessors?? comments!
	Shape labelShape;
	float labelOffsX;
	float labelOffsY;

	/** This State's corresponding ch.ethz.ifa.model.State.
	 */
	public FSMStateInterface model;

	/** font sizes for the different Renderers. */
	Hashtable fontSizes = new Hashtable();

	/** All Transitions originating from this state.
	 */
	public List outTransitions = new ArrayList();
	/** All Transitions pointing to this state.
	 */
	public List inTransitions = new ArrayList();

	/** The position of this states center. */
	double centerX, centerY;
	/** This state's bounding box. */
	double x, y, width, height;

	/** The position relative to this state's center where this state was grabbed by the mouse.
	 *  Needed in FSMView.moveState(...).
	 */
	double mouseRefX, mouseRefY;

	boolean SHOW_WHOLE_LABEL = false;

	/** Constructs a State that represents the specified ch.ethz.ifa.model.State.
	 */
	public State(FSMStateInterface s) {
		this.model = s;
		this.label = s.getLabel().toString();
		this.setAccepting(s.isAccepting());
		this.setStart(s.isStartState());
		//this.setToolTipText(this.label);
		this.centerX = s.getX();
		this.centerY = s.getY();
	}

	boolean hasTransitionTo(State s) {
		for (int i = 0; i < outTransitions.size(); i++) {
			if (outTransitions.get(i) == s)
				return true;
		}
		return false;
	}

	/** Highlights or dehighlights this state's border. */
	public void setBorderHighlighted(boolean h) {
		HIGHLIGHT_BORDER = h;
	}

	/** Returns true if this state's border is highlighted. */
	public boolean isBorderHighlighted() {
		return HIGHLIGHT_BORDER;
	}

	/** Highlights or dehighlights this state's center. */
	public void setCenterHighlighted(boolean h) {
		HIGHLIGHT_CENTER = h;
	}

	/** Returns true if this state's center is highlighted. */
	public boolean isCenterHighlighted() {
		return HIGHLIGHT_CENTER;
	}

	/** Selects this state. */
	public void setSelected(boolean s) {
		SELECTED = s;
	}

	/** Returns true if this state is selected. */
	public boolean isSelected() {
		return SELECTED;
	}

	/** Sets this state's mark to the specified Integer. */
	public void setMark(Object mark) {
		this.mark = mark;
	}

	/** Returns this state's mark. */
	public Object getMark() {
		return this.mark;
	}

	/** Turns this state to an accepting state. */
	public void setAccepting(boolean accept) {
		ACCEPTING = accept;
	}

	/** Returns true if this state is an accepting state. */
	public boolean isAccepting() {
		return ACCEPTING;
	}

	/** Turns this state to a startstate if start is true. */
	public void setStart(boolean start) {
		START = start;
	}

	/** Returns true if this state is a startstate. */
	public boolean isStart() {
		return START;
	}

	/** Sets the state's label to the specified String. */
	public void setLabel(Object label) {
		this.label = label.toString();
		//this.setToolTipText(label);
		fontSizes.clear();
	}

	/** Returns the state's label. */
	public String getLabel() {
		return this.label;
	}

	public void setVisibleLabel(String label) {
		visibleLabel = label;
	}

	public String getVisibleLabel() {
		return visibleLabel;
	}

	public void setFontSize(ComponentRenderer renderer, float size) {
		fontSizes.put(renderer, new Float(size));
	}

	public Float getFontSize(ComponentRenderer renderer) {
		return (Float) fontSizes.get(renderer);
	}

	private Font font;
	public void setFont(Font font) {
		this.font = font;
	}

	public Font getFont() {
		return this.font;
	}

	public Rectangle getWholeLabelBounds(FontMetrics fm) {
		return new Rectangle(
			(int) (centerX - fm.stringWidth(getLabel()) / 2) - 3,
			(int) (centerY - fm.getHeight() / 2),
			fm.stringWidth(getLabel()) + 6,
			fm.getHeight());
	}

	/** Grabs this state.
	 *  MouseRefX and MouseRefY are set to the distances
	 *  between the specified coordinates and the center of this state.
	 *
	 * @param   x    The x-coordinate of the point where the state was grabbed.
	 * @param   y    The y-coordinate of the point where the state was grabbed.
	 */
	public void grab(double x_, double y_) {
		mouseRefX = x_ - centerX;
		mouseRefY = y_ - centerY;
	}

	/** Releases this state. Inverse Operation of grab(). MouseRefX and MouseRefY are set to 0. */
	public void release() {
		mouseRefX = 0.0;
		mouseRefY = 0.0;
	}

	public double getMouseRefX() {
		return mouseRefX;
	}

	public double getMouseRefY() {
		return mouseRefY;
	}

	/** Adds a Transition originating from this state. */
	public void addOutTransition(Transition fat) {
		outTransitions.add(fat);
	}

	/** Deletes a Transition originating from this state. */
	public void removeOutTransition(Transition fat) {
		outTransitions.remove(fat);
	}

	/** Adds a Transition pointing to this state. */
	public void addInTransition(Transition fat) {
		inTransitions.add(fat);
	}

	/** Deletes a Transition pointing to this state. */
	public void removeInTransition(Transition fat) {
		inTransitions.remove(fat);
	}

	/** Sets the state's center to the specified position.
	 *
	 * @param   x    The x-coordinate of the state's new center.
	 * @param   y    The y-coordinate of the state's new center.
	 */
	public void setCenter(double x, double y) {
		centerX = x;
		centerY = y;
	}

	/** Returns the x coordinate of this state's center. */
	public double getCenterX() {
		return centerX;
	}

	/** Returns the y coordinate of this state's center. */
	public double getCenterY() {
		return centerY;
	}

	/** Returns this state's selftransition or null if it doesn't have one.
	 */
	public Transition getSelfTransition() {
		Transition currTrans;
		for (int i = 0; i < outTransitions.size(); i++) {
			currTrans = (Transition) outTransitions.get(i);
			if (currTrans.to == this)
				return currTrans;
		}
		return null;
	}

}