/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.awt.Shape;
import java.awt.geom.CubicCurve2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.QuadCurve2D;

/** A Transition Object represents all the transitions going from one ch.ethz.ifa.model.State to another.
 *  If more than one exist, they are represented as multiple transitionlabels.
 *  The curve of a Transition is either LINE, QUAD or CUBIC.
 *
 * @author		Reto Lamprecht
 */
public class Transition {

	public static int LINE = 1;
	public static int QUAD = 2;
	public static int CUBIC = 3;

	/** The transition's type. */
	int type;

	/** Indicates that this transition is highlighted. */
	boolean HIGHLIGHTED = false;

	/** indicates that this transition is selected. */
	boolean SELECTED = false;

	/** This transitions marks. */
	public Object mark = ComponentRenderer.MARK_NOT_MARKED;

	/** the TransitionLabel of this transition.
	 */
	public TransitionLabel label;

	/** the States that this transition connects.
	 */
	public State from, to;

	/** The coordinates of the point where the transition leaves the from-state. */
	double fromX, fromY;
	/** The coordinates of the point where the transition ends. (In the middle of it's head.) */
	double toX, toY;
	/** Controlpoint for QUAD curves. (convers transitions) */
	double ctrlFromX, ctrlFromY;
	/** Second controlpoint for CUBIC curves. (self-transitions) */
	double ctrlToX, ctrlToY;
	/** The transition's orientation. */
	double angle;
	/** The transition's head's orientation. */
	double angleHead;

	/** The GeneralPath Object representing the transition's head. */
	GeneralPath headGP = new GeneralPath(GeneralPath.WIND_NON_ZERO, 4);

	/** The GeneralPath Object representing the sensitive area of the transition's head. */
	GeneralPath sensHeadGP = new GeneralPath(GeneralPath.WIND_NON_ZERO, 4);

	/** Constructs a new Transition.
	 *
	 * @param   from     the State this transition emanates from.
	 * @param   label    this transition's label.
	 */
	Transition(State from, Object input) {
		this.label = new TransitionLabel(this, input);
		this.from = from;
		toX = from.getCenterX();
		toY = from.getCenterY();
		this.type = LINE;
	}

	public Transition(State from, State to, Object input) {
		this.label = new TransitionLabel(this, input);
		this.from = from;
		this.to = to;
		if (from == to)
			this.type = CUBIC;
		else if (to.hasTransitionTo(from))
			this.type = QUAD;
		else
			this.type = LINE;
	}

	/** Highlights or dehighlights this transition. */
	public void setHighlighted(boolean h) {
		this.HIGHLIGHTED = h;
	}

	/** Returns true if this transition is highlighted. */
	public boolean isHighlighted() {
		return HIGHLIGHTED;
	}

	/** Selects or deselects the transition. */
	public void setSelected(boolean s) {
		SELECTED = s;
	}

	/** Returns true if this transition is selected; returns false otherwise. */
	public boolean isSelected() {
		return SELECTED;
	}

	/** Sets this transition's mark to the specified Integer. */
	public void setMark(Object mark) {
		this.mark = mark;
	}

	/** Returns this transition's mark. */
	public Object getMark() {
		return this.mark;
	}

	/** Sets the coordinates of this transition's head.
	 *
	 * @param   x      The x coordinate of the head's center.
	 * @param   y      The y coordinate of the head's center.
	 */
	public void setTo(double x, double y) {
		toX = x;
		toY = y;
	}

	/** Sets the State that this transition is leading to. */
	public void setToState(State fas) {
		to = fas;
	}

	/** Returns the transition line as a Shape Object. */
	public Shape getCurve() {
		if (type == CUBIC) {
			return new CubicCurve2D.Double(
				fromX,
				fromY,
				ctrlFromX,
				ctrlFromY,
				ctrlToX,
				ctrlToY,
				toX,
				toY);
		} else if (type == QUAD) {
			return new QuadCurve2D.Double(fromX, fromY, ctrlFromX, ctrlFromY, toX, toY);
		} else if (type == LINE) {
			return new Line2D.Double(fromX, fromY, toX, toY);
		}
		return new Line2D.Double(0, 0, 0, 0);
	}
}