/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.StringTokenizer;

/** A FAtransitionLabel Object belongs to a Transition and displays the inputs of all transitions
 *  going from one ch.ethz.ifa.model.State to another. The inputs can be accessed separately by getInput(int)
 *  as a Vector by getInputs() or as a String of all inputs, separated by '|', by getLabelString().
 * 
 * @author		Reto Lamprecht
 */
public class TransitionLabel {

	/** The Transition this label belongs to. */
	Transition fat;

	/** The Vector containing the Character Objects representing the inputs. */
	List inputs = new ArrayList();

	/** The String of all inputs, separated by '|'. If the transitionlabel contains no inputs, this String is set to null */
	String labelString;

	/** The transitionlabel's center. */
	double x, y;

	/** The angle of the transition at the label's position.
	 *  Used to translate label correctly, if it must not lie on transition. */
	double angle;

	/** The sensitive area bounding box of this transitionlabel. */
	Rectangle2D.Double sensBounds = new Rectangle2D.Double();

	/** The drawing bounding box of this transitionlabel. */
	Rectangle2D.Double bounds = new Rectangle2D.Double();

	private List renderingHints = new ArrayList();
	public void addRenderingHint(Object hint) {
		renderingHints.add(hint);
	}
	public void removeRenderingHint(Object hint) {
		renderingHints.remove(hint);
	}
	public boolean hasRenderingHint(Object hint) {
		return renderingHints.contains(hint);
	}

	/** Constructs a new transitionlabel belonging to the specified transition with the specified input Character. */
	public TransitionLabel(Transition fat, Object input) {
		this.fat = fat;
		this.addInput(input);
	}

	/** Adds the specified Character to this transitionlabel's inputs if it is not null or '' or already added. */
	public void addInput(Object input) {
		if (null != input && !"".equals(input.toString())) { //$NON-NLS-1$
			for (int i = 0; i < inputs.size(); i++) {
				if (input.equals(getInput(i))) {
					return;
				}
			}
			inputs.add(input);
			updateLabelString();
		}
	}

	/** Removes the specified Character from this transitionlabel's inputs. */
	public void removeInput(Object input) {
		for (int i = 0; i < inputs.size(); i++) {
			if (input.equals(getInput(i))) {
				inputs.remove(getInput(i));
				updateLabelString();
				break;
			}
		}
	}

	/** Adds all inputs of the specified TransitionLabel to this transitionlabel's inputs. */
	public void addFATLabel(TransitionLabel fatLabel) {
		for (int i = 0; i < fatLabel.inputs.size(); i++) {
			addInput(fatLabel.getInput(i));
		}
	}

	/** Removes all inputs of the specified TransitionLabel from this transitionlabel's inputs. */
	public void removeFATLabel(TransitionLabel fatLabel) {
		if (fatLabel != null) {
			for (int i = 0; i < fatLabel.inputs.size(); i++) {
				removeInput(fatLabel.getInput(i));
			}
		}
	}

	/** Replaces this transitionlabel's inputs by the inputs of the specified TransitionLabel. */
	public void setFATLabel(TransitionLabel fatLabel) {
		clear();
		for (int i = 0; i < fatLabel.inputs.size(); i++) {
			addInput(fatLabel.getInput(i));
		}
	}

	/** Replaces this transitionlabel's inputs by the inputs specified in the argument String (separated by "|").
	 *  Only the first char of each input is taken into account.
	 */
	//todo: is this consistent??
	public void setInputs(String labelString) {
		StringTokenizer tmp = new StringTokenizer(labelString, "|"); //$NON-NLS-1$
		clear();
		while (tmp.hasMoreTokens()) {
			addInput(new Character(tmp.nextToken().charAt(0)));
		}
	}

	/** Returns the Vector containing all the input Characters. */
	public List getInputs() {
		return inputs;
	}

	/** Returns the input Character with the specified index. */
	public Object getInput(int i) {
		return inputs.get(i);
	}

	/** Returns the number of inputs in this transitionlabel. */
	public int number() {
		return inputs.size();
	}

	/** Deletes all inputs. */
	public void clear() {
		inputs.clear();
		labelString = null;
	}

	/** Updates the labelString with the Characters of the inputs Vector. */
	void updateLabelString() {
		String[] labels = new String[inputs.size()];
		for (int i = 0; i < inputs.size(); i++) {
			labels[i] = inputs.get(i).toString();
		}
		//Object[] ls = inputs.toArray();
		Arrays.sort(labels);
		labelString = null;
		if (labels.length > 0 && !"".equals(labels[0])) { //$NON-NLS-1$
			labelString = String.valueOf(labels[0]);
			for (int i = 1; i < labels.length; i++) {
				labelString += "|" + labels[i]; //$NON-NLS-1$
			}
		}
	}

	/** Returns the all inputs as a String, separated by "|". */
	public String getLabelString() {
		return labelString;
	}

	/** Returns a copy of this transitionlabel. */
	public TransitionLabel copy() {
		TransitionLabel clone = new TransitionLabel(this.fat, null);
		for (int i = 0; i < inputs.size(); i++) {
			clone.inputs.add(inputs.get(i));
		}
		return clone;
	}

}