/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.fsmgui;

import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;

/**
 * @author		Reto Lamprecht
 */
public class ZoomNDragablePane {

	double scale = 1;

	/** The scale factor at the beginning of a scale drag. */
	double refScale;
	/* the distance the mouse has to be dragged for a 2x magni-/minification. */
	double dScaleDist = 150.0;

	double maxScale = 8.0;
	double minScale = 0.125;
	double dx = 0.0, dy = 0.0;
	double refDx, refDy;

	/** translation and scale are limited by a ViewRectangle and a Model Rectangle... */
	boolean limited = false;
	/** the rectangles that limit the translation. */
	Rectangle viewRect;
	Rectangle objectRect;

	/** coordinates where a zoom or tanslation gesture started. */
	double transRefX, transRefY;
	double scaleRefX, scaleRefY;

	private Point2D modScaleRef = new Point2D.Double();

	private Point2D p = new Point2D.Double();

	public ZoomNDragablePane() {}

	/** first two args specify dimensions of the viewPort, args 3 and 4 the dimensions of the zoom-n-draggable rectangle to be displayed in this viewport.
	 *  If this method is called, zooming is limited to zooming out until the displayed rectangle fits into the view rectangle,
	 *  and dragging is limited by the borders of the displayed rectangle.
	 *
	 */
	public void limited(int viewW, int viewH, int objW, int objH) {
		limited = true;
		this.viewRect = new Rectangle(0, 0, viewW, viewH);
		this.objectRect = new Rectangle(0, 0, objW, objH);
		maxScale = 1.0;
	}

	public boolean isLimited() {
		return viewRect != null && objectRect != null;
	}

	public void reset() {
		scale = 1.0;
		dx = 1.0;
		dy = 1.0;
	}

	public void setScale(double scale) {
		this.scale = scale;
	}

	public double getScale() {
		return scale;
	}

	public void setTranslation(double dx, double dy) {
		this.dx = dx;
		this.dy = dy;
	}

	public double getTranslationX() {
		return dx;
	}

	public double getTranslationY() {
		return dy;
	}

	public void setDoubleScaleDistance(int distance) {
		dScaleDist = distance;
	}

	public void setScaleRange(double min, double max) {
		minScale = min;
		maxScale = max;
	}

	public void setTranslationRef(int x, int y) {
		transRefX = x;
		transRefY = y;
		refDx = dx;
		refDy = dy;
	}

	public void setTranslationRef(double x, double y) {
		transRefX = x;
		transRefY = y;
		refDx = dx;
		refDy = dy;
	}

	public void translate(int x, int y) {
		dx = refDx + x - transRefX;
		dy = refDy + y - transRefY;
		if (limited)
			limitDrag();
	}

	public void setScaleRef(int x, int y) {
		scaleRefX = x;
		scaleRefY = y;
		refScale = scale;
	}

	public void setScaleRef(double x, double y) {
		scaleRefX = x;
		scaleRefY = y;
		refScale = scale;
	}

	public void scale(int y) {
		double newScale;
		if (scaleRefY < y) {
			newScale = refScale * (1 + 2 * (y - scaleRefY) / dScaleDist);
		} else {
			newScale = refScale * (1 - 0.5 * (scaleRefY - y) / dScaleDist);
		}
		if (newScale < minScale)
			newScale = minScale;
		if (newScale > maxScale)
			newScale = maxScale;

		getModelCoords(modScaleRef, scaleRefX, scaleRefY);
		scale = newScale;
		if (limited)
			newScale = limitZoom();

		double diffX = scaleRefX - (modScaleRef.getX() * newScale + dx);
		double diffY = scaleRefY - (modScaleRef.getY() * newScale + dy);
		dx = dx + diffX;
		dy = dy + diffY;

		if (limited)
			limitDrag();
	}

	/** Call this method to center the object rectangle in the view rectangle. */
	public void adjustLimitations() {
		limitZoom();
		limitDrag();
	}

	private double limitZoom() {
		if (scale < 1) {
			double widthQuot =
				Math.max(
					objectRect.width * scale / viewRect.width,
					objectRect.height * scale / viewRect.height);
			if (widthQuot < 1)
				scale = Math.min(scale / widthQuot, 1.0);
		}
		return scale;
	}

	private void limitDrag() {
		getViewCoords(p, objectRect.x, 0.0);
		double diffLeftX = p.getX() - viewRect.x;
		getViewCoords(p, objectRect.width + objectRect.x, 0.0);
		double diffRightX = p.getX() - viewRect.width - viewRect.x;
		getViewCoords(p, 0.0, objectRect.y);
		double diffLeftY = p.getY() - viewRect.y;
		getViewCoords(p, 0.0, objectRect.height + objectRect.y);
		double diffRightY = p.getY() - viewRect.height - viewRect.y;

		if (objectRect.width * scale < viewRect.width) {
			// if objRect smaller than viewRect center objRect.
			dx = (diffLeftX - diffRightX) / 2;
		}
		// left edge of objRect is dragged away from left edge of viewRect
		else if (diffLeftX > 0) {
			dx -= diffLeftX;
		} else {
			if (diffRightX < 0) {
				getViewCoords(p, objectRect.x, 0.0);
				dx -= Math.max(p.getX() - viewRect.x, diffRightX);
			}
		}

		if (objectRect.height * scale < viewRect.height) {
			// if objRect smaller than viewRect center objRect.
			dy = (diffLeftY - diffRightY) / 2;
		}
		// upper edge of objRect is dragged away from upper edge of viewRect
		else if (diffLeftY > 0) {
			dy -= diffLeftY;
		} else {
			if (diffRightY < 0) {
				getViewCoords(p, 0.0, objectRect.y);
				dy -= Math.max(p.getY() - viewRect.y, diffRightY);
			}
		}
	}

	public Point2D getModelCoords(Point2D p_, double x, double y) {
		p_.setLocation((x - dx) / scale, (y - dy) / scale);
		return p_;
	}

	public Point2D getViewCoords(Point2D p_, double x, double y) {
		p_.setLocation(x * scale + dx, y * scale + dy);
		return p_;
	}

	public GeneralPath makeTransformedPath(float[] coords) {
		GeneralPath path = new GeneralPath();
		path.reset();
		path.moveTo((float) (coords[0] * scale + dx), (float) (coords[1] * scale + dy));
		for (int i = 2; i < coords.length - 1; i += 2) {
			path.lineTo(
				(float) (coords[i] * scale + dx),
				(float) (coords[i + 1] * scale + dy));
		}
		path.closePath();
		return path;
	}

	public void transform(Graphics2D g) {
		g.translate((int) dx, (int) dy);
		g.scale(scale, scale);
	}
}