/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import ch.ethz.exorciser.Debug;
import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.fsmgui.FSMEvent;
import ch.ethz.exorciser.fsmgui.FSMModelListener;
import ch.ethz.exorciser.fsmgui.FSMStateInterface;

/**
 * @author Reto Lamprecht
 * @author Christian Hfeli
 * @author Vincent Tscherter
 */
public class BasicFAModel implements BasicFA {

	protected Alphabet alphabet;

	private List listeners;

	protected String name;

	protected State startState;

	protected Hashtable states;

	protected int type;

	/**
	 * Constructs an FABase. The FABase will get the initial name
	 * <code>'sampleFA'</code>.
	 */
	public BasicFAModel() {
		setFA(""); //$NON-NLS-1$
	}

	/**
	 * Constructs an empty FABase with the given inital name.
	 * 
	 * @param name
	 *            The name of the FA
	 */
	public BasicFAModel(String name) {
		setFA(name);
	}

	/**
	 * The method allows to subscribe other component as a FA model listener.
	 * Therefore they need to have implemented the FAListerner inteface. After
	 * subscribtion the model will inform the listener by sending event about
	 * any modification on the model.
	 * 
	 * @param listener
	 *            A FAListener.
	 */
	public void addFSMModelListener(FSMModelListener listener) {
		if (listener != null && !listeners.contains(listener)) {
			listeners.add(listener);
		}
	}

	public void addState(FSMStateInterface state) {
		addState((State) state);
	}

	/**
	 * Adds the given state to the FABase. By default the the state is concerned
	 * as NON_ACCEPTING. Use <code>setAccepting</code> to add an accepting
	 * state to the FA.
	 * 
	 * @param state
	 *            The state.
	 */
	public void addState(State state) {
		if (state != null && !states.containsKey(state)) {
			states.put(state, NON_ACCEPTING);
			state.setFA(this);
			notifyFSMModelListeners(FSMEvent.createAddStateEvent(state));
		}
	}

	/**
	 * This method clears all states and their transitions from the FA.
	 */

	public void clear() {
		startState = null;
		states = new Hashtable();
		notifyFSMModelListeners(FSMEvent.createClearAllEvent());
	}

	/**
	 * Returns a clone of the given BasicFAModel. Even all related states will
	 * be cloned.
	 * 
	 * @return A BasicFAModel clone.
	 */
	public Object clone() {
		Hashtable marks = new Hashtable();
		BasicFAModel theClone = new BasicFAModel();
		theClone.setAlphabet(this.getAlphabet());
		theClone.setName(this.name);
		theClone.setLayoutType(this.getLayoutType());

		Iterator en = states.keySet().iterator();
		while (en.hasNext()) {
			State tmpState = (State) en.next();
			if (!marks.containsKey(tmpState)) {
				stateCloning(theClone, tmpState, marks, false);
			}
		}
		return theClone;
	}

	/**
	 * Returns <code>true</code> if the given state is part of the FABase.
	 * 
	 * @return <code>true</code> if and only if the specified state is part of
	 *         the FABase; <code>false</code> otherwise.
	 */
	public boolean contains(State state) {
		return states.containsKey(state);
	}

	public boolean containsTransitionOnInput(Character c) {
		State current;
		Iterator iter = getStateSet().iterator();
		while (iter.hasNext()) {
			current = (State) iter.next();
			if (current.hasTransitionOnInput(c))
				return true;
		}
		return false;
	}

	/**
	 * Use this method to delete all states of the FA. This action removes all
	 * transitions too.
	 */
	public void deleteAllTransitions() {
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			Iterator alphIter = this.getAlphabet().iterator();
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				tmpState.deleteTransitions(inp);
			}
		}
	}

	/**
	 * If the state belongs to the FABase, this method will delete it. Also the
	 * transitions from other states to this state will be deleted.
	 * 
	 * @param st
	 *            The state to be deleted.
	 * @exception FAException
	 *                if the given state was the start state.
	 */
	public void deleteState(State state) throws FAException {
		if (state == startState && states.size() > 1) {
			throw new FAException("Can not delete the start state"); //$NON-NLS-1$
		}
		if (state != null && states.containsKey(state)) {
			// first delete the transitions from other states to the given one.
			Hashtable prec = getPrecedingStates(state, null);
			Iterator en = prec.keySet().iterator();
			while (en.hasNext()) {
				State tmp = (State) en.next();

				Iterator iter = alphabet.iterator();
				while (iter.hasNext()) {
					Character inp = (Character) iter.next();
					tmp.deleteTransition(inp, state);
				}
			}
			// ... then, delete the transitions from this state to other
			// states.
			Iterator iter = alphabet.iterator();
			while (iter.hasNext()) {
				Character inp = (Character) iter.next();
				Set nextStates = state.getNextStateSet(inp);
				Iterator stateIter = nextStates.iterator();
				while (stateIter.hasNext()) {
					State to = (State) stateIter.next();
					state.deleteTransition(inp, to);
				}
			}

			// if the given state was the start state, this no longer exists.
			if (state == startState) {
				startState = null;
			}
			states.remove(state);
			notifyFSMModelListeners(FSMEvent.createDeleteStateEvent(state));
		}
	}

	/**
	 * Returns a set of all <code>accept states</code>.
	 * 
	 * @return A set of all <code>accept states</code>;
	 */
	public Set getAcceptingStates() {
		HashSet res = new HashSet();
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmp = (State) iter.next();
			if (tmp.isAccepting()) {
				// (State)states.get(tmp)).isAccepting()) {// == ACCEPTING){
				res.add(tmp);
			}
		}
		return res;
	}

	public Set getAllStates() {
		return states.keySet();
	}

	public Alphabet getAlphabet() {
		return alphabet;
	}

	/**
	 * This method returns the type which specifies how the model would like to
	 * be visualised. See setType().
	 */
	public int getLayoutType() {
		return type;
	}

	/**
	 * Returns the name of the FA.
	 * 
	 * @return The name of the FABase. <code>sampleFA</code> if the FABase has
	 *         it's default name.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns a <code>Set</code> of states which can never be reached in this
	 * FA.
	 * 
	 * @return <code>Set</code> of states.
	 * @exception FAException
	 *                if no <code>start state</code> is defined in this FA.
	 */
	public Set getNotReachableStates() {
		Set reachables = getReachableStates();
		Hashtable statesClone = (Hashtable) this.states.clone();
		Set res = statesClone.keySet();
		res.removeAll(reachables);
		return res;
	}

	/**
	 * Adds to the Hashtable all states which have a transition to the given
	 * state. If result Hashtable parameter is null a new Hashtable will be
	 * built and filled up.
	 * 
	 * @param st
	 *            The state the transitions go to.
	 * @param res
	 *            A Hashtable to fill in the preceding states.
	 * @return The given Hashtable or a new one if res was null.
	 */
	public Hashtable getPrecedingStates(State st, Hashtable res) {
		if (res == null) {
			res = new Hashtable();
		}
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmp = (State) iter.next();

			Iterator alphIter = alphabet.iterator();
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				Set nextStates = tmp.getNextStateSet(inp);
				Iterator stateIter = nextStates.iterator();
				while (stateIter.hasNext()) {
					State nextState = (State) stateIter.next();
					if (nextState == st) {
						res.put(tmp, tmp);
					}
				}
			}
		}
		return res;
	}

	/**
	 * Returns a randomly one of the states of this FA. If no state is defined
	 * an FAException will be thrown.
	 * 
	 * @return The randomly choosen state.
	 * @exception FAException
	 *                if there is no state in this FA.
	 */
	public State getRandomState() throws FAException {
		if (states.size() == 0) {
			throw new FAException("No state found in this FA."); //$NON-NLS-1$
		}
		Object[] statearray = this.getStateSet().toArray();
		return ((State) statearray[(int) (Math.random() * statearray.length)]);
	}

	/**
	 * Returns a <code>Set</code> of states which can be reached in this FA.
	 * 
	 * @return <code>Set</code> of states.
	 * @exception FAException
	 *                if no <code>start state</code> is defined in this FA.
	 */
	public Set getReachableStates() {
		if (this.getStartState() == null) {
			// rl 13.3.02
			// throw new FAException(
			// FAException.STATE_NOT_FOUND,
			// "No start state defined in this FA");
			return null;
		}
		return this.getReachableStates((State) this.getStartState(),
				new HashSet());
	}

	/**
	 * Returns a <code>Set</code> of states which can never be reached from
	 * the given state.
	 * 
	 * @return <code>Set</code> of states.
	 * @exception FAException
	 *                if the given <code>state</code> is not part of this FA.
	 */
	public Set getReachableStates(State state) throws FAException {
		if (state == null || !this.states.containsKey(state)) {
			throw new FAException("The given state is not part of this FA"); //$NON-NLS-1$
		}
		return this.getReachableStates(state, new HashSet());
	}

	private Set getReachableStates(State st, Set marks) {
		if (states.containsKey(st) && marks != null && !marks.contains(st)) {
			marks.add(st);
			Set neighbours = st.getAllNeighbours();
			Iterator iter = neighbours.iterator();
			while (iter.hasNext()) {
				State tmpState = (State) iter.next();
				if (!marks.contains(tmpState)) {
					getReachableStates(tmpState, marks);
				}
			}
		}
		return marks;
	}

	// public void clear();

	public FSMStateInterface getStartState() {
		return startState;
	}

	/**
	 * Returns the <code>State</code> object which corresponds to the given
	 * state label string. Returns <code>null</code> if no state which such a
	 * label exists in this FA. Note: If more than one state have the same
	 * label, the one the method returns is not specified.
	 * 
	 * @return <code>State</code> corresponding to the specified label;
	 *         <code>false</code> if no such state is part of the FABase;
	 */
	public State getState(String label) {
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			if (tmpState.getLabel() != null
					&& tmpState.getLabel().equals(label)) {
				return tmpState;
			}
		}
		return null;
	}

	Hashtable getStates() {
		return states;
	}

	/**
	 * Returns an iterator over all states of the FABase.
	 * 
	 * @return An <code>iterator</code> over all states.
	 */
	/*
	 * public Iterator getAllStates(){ return states.keySet().iterator(); }
	 */

	/**
	 * Returns a Set of all states of the FABase.
	 * 
	 * @return A <code>Set</code> over all states.
	 */
	public Set getStateSet() {
		return states.keySet();
	}

	/**
	 * Returns <code>Set</code> of all trap states of this FA.
	 * 
	 * @return <code>Set</code> of trap states.
	 */
	public Set getTrapStates() {
		HashSet res = new HashSet();
		try {
			Iterator iter = this.getAllStates().iterator();
			while (iter.hasNext()) {
				State tmpState = (State) iter.next();
				if (tmpState != this.getStartState()
						&& !this.hasWayToAcceptingState(tmpState)) {
					res.add(tmpState);
				}
			}
		} catch (FAException fae) {
			Debug.showException(fae);
		}
		return res;
	}

	/**
	 * Returns <code>true</code> if there is any epsilon transition defined in
	 * this FA. If this is true this FA is nondeterministic!
	 * 
	 * @return <code>True</code> if there is any epsilon transition;
	 *         <code>false</code> otherwise.
	 */
	public boolean hasEpsilonTransition() {
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			Set nextStates = tmpState.getNextStateSet(Alphabet.EPSILON);
			if (nextStates.size() > 0) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns <code>true</code> if there is any state with more than one
	 * transition defined on the same input in this FA. If this is true this FA
	 * is nondeterministic!
	 * 
	 * @return <code>True</code> if there is any multiple transition;
	 *         <code>false</code> otherwise.
	 */
	public boolean hasMultipleTransitions() {
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			Iterator alphIter = AlphabetUtil.getDetIterator(getAlphabet());
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				Set nextStates = tmpState.getNextStateSet(inp);
				if (nextStates.size() > 1) {
					return true;
				}
			}
		}
		return false;
	}

	public boolean hasNoOtherTransitions(Collection inputs) {
		State current;
		Iterator iter = getStateSet().iterator();
		while (iter.hasNext()) {
			current = (State) iter.next();
			if (!current.hasNoOtherTransitions(inputs))
				return false;
		}
		return true;
	}

	/**
	 * Returns <code>true</code> if there exists a transition on the given
	 * input symbol between the two states.
	 * 
	 * @return <code>True</code> if the transition exists; <code>false</code>
	 *         otherwise.
	 */
	public boolean hasTransition(State from, State to, Character inp) {
		Set nextStates = from.getNextStateSet(inp);
		return nextStates.contains(to);
	}

	/**
	 * Returns <code>true</code> if there is way from the given state to one
	 * of the <code>accpet states</code>.
	 * 
	 * @return <code>True</code> if their exists a way to an
	 *         <code>accept state</code>;<code>false</code> otherwise.
	 * @exception FAException
	 *                if the given state is not part of the FAModel.
	 */
	public boolean hasWayToAcceptingState(State state) throws FAException {
		Set reachables = this.getReachableStates(state);
		return this.isOneOfTheseAccepting(reachables);
	}

	private void setFA(String name) {
		this.name = name;
		startState = null;
		states = new Hashtable();
		alphabet = Alphabet.createABeAlphabet();
		type = USER_DEFINED;
		listeners = new ArrayList();
	}

	/**
	 * Returns <code>true</code> of the given state is part of the the set of
	 * <code>accept states</code> of the FABase.
	 * 
	 * @return <code>true</code> if and only if the specified object is an
	 *         ACCEPTING state of this FA; <code>false</code> otherwise.
	 * @exception FAException
	 *                if the given state is not part of the FABase.
	 */
	public boolean isAccepting(State state) throws FAException {
		if (state == null || !states.containsKey(state)) {
			throw new FAException("State does not belong to this FABase"); //$NON-NLS-1$
		}
		return state.isAccepting();
	}

	/**
	 * Returns <code>true</code> if this FA is deterministic. So the FA has
	 * not epsilon transitions, no multiple defined transitions but also no
	 * underdefined states. Means: for each state is excactly one transition
	 * defined for each input out of the alphabet.
	 * 
	 * @return <code>True</code> if the FA is deterministic.
	 *         <code>false</code> otherwise.
	 * @exception FAException
	 *                if no start state is defined in the FAModel.
	 */
	public boolean isDeterministic() throws FAException {
		if (this.getStartState() == null) {
			throw new FAException("No start state defined in this FA"); //$NON-NLS-1$
		}
		if (this.hasEpsilonTransition() || this.hasMultipleTransitions()) {
			return false;
		}
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			Iterator alphIter = AlphabetUtil.getDetIterator(getAlphabet());
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				Set nextStates = tmpState.getNextStateSet(inp);
				if (nextStates.size() != 1) {
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Returns <code>true</code> if there exist no way from the
	 * <code>start state</code> to one of the <code>accpet states</code>.
	 * The searching time is linear to the size of the graph.
	 * 
	 * @return <code>True</code> if it exists a way to an
	 *         <code>accept state</code>;<code>false</code> otherwise.
	 * @exception FAException
	 *                if no start state is defined in the FAModel.
	 */
	public boolean isLanguageEmpty() { // rl.13.3.02 throws FAException {
		try {
			Set reachables = this.getReachableStates();
			return !this.isOneOfTheseAccepting(reachables);
		} catch (FAException e) {
			Debug.showException(e);
			return true;
		}
	}

	/**
	 * Returns <code>true</code> if at least one of the states in the given
	 * state set is an <code>accept state</code>.
	 * 
	 * @return <code>true</code> if and only if one of the states in the state
	 *         set is an ACCEPTING state of this FA; <code>false</code>
	 *         otherwise.
	 * @exception FAException
	 *                if one of the given states is not part of the FABase.
	 */
	public boolean isOneOfTheseAccepting(Set someStates) throws FAException {
		Iterator iter = someStates.iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			if (isAccepting(tmpState)) {
				return true;
			}

		}
		return false;
	}

	public void notifyFSMModelListeners(FSMEvent event) {
		for (int i = 0; i < listeners.size(); i++) {
			FSMModelListener fsml = (FSMModelListener) listeners.get(i);
			fsml.notify(event);
		}
	}

	/**
	 * Use this method to (re-)label all states of the FA. The labels begin will
	 * be the given prefix, the endings will be numeric indices. The
	 * <code>start state</code> will have the index 0.
	 * 
	 * @param begin
	 *            The common beginning of the <code>label</code> of all
	 *            states.
	 */
	public void reLabelAllStates(String pref) {
		String prefix = pref;
		if (prefix == null) {
			prefix = ""; //$NON-NLS-1$
		} else if (prefix.length() >= 1) {
			prefix = prefix.substring(0, 1);
			try {
				Integer.parseInt(prefix);
				prefix = ""; //$NON-NLS-1$
			} catch (NumberFormatException nfe) {
			}
		}
		int index = 1;
		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			if (getStartState() == tmpState) {
				tmpState.setLabel(prefix + "0"); //$NON-NLS-1$
			} else {
				tmpState.setLabel(prefix + index);
				index++;
			}
		}
	}

	/**
	 * Removes the given FAListener.
	 * 
	 * @param listener
	 *            A FAListener.
	 */
	public void removeFSMModelListener(FSMModelListener listener) {
		if (listener != null && listeners.contains(listener)) {
			listeners.remove(listener);
		}
	}

	public void removeState(FSMStateInterface state) {
		try {
			if (state instanceof State) {
				deleteState((State) state);
			}
		} catch (FAException e) {
			Debug.showException(e);
		}
	}

	public void setAccepting(State state, boolean flag) {
		if (state == null || !states.containsKey(state)) {
		} else {
			state.setAccepting(flag);
		}
	}

	public void setAlphabet(Alphabet alph) {
		alphabet = alph;
		notifyFSMModelListeners(FSMEvent.createSetAlphabetEvent(alph));
	}

	/**
	 * Replaces this FA by a clone of the specified FA. Works only for arguments
	 * of type BasicFAModel.
	 * 
	 * @see FA#setFA()
	 */
	public void setFA(BasicFA other) {
		this.name = other.getName();
		this.alphabet = other.getAlphabet();
		this.type = other.getLayoutType();

		// add states
		this.clear();
		Hashtable cloneMap = new Hashtable(); // needed to add transitions
		// later
		Iterator itOtherStates = other.getAllStates().iterator();
		while (itOtherStates.hasNext()) {
			State sOther = (State) itOtherStates.next();
			State sClone = (State) sOther.clone();
			cloneMap.put(sOther, sClone);
			this.addState(sClone);
			if (sOther.isStartState()) {
				this.startState = sClone;
			}
		}
		// copy transitions
		itOtherStates = cloneMap.keySet().iterator();
		while (itOtherStates.hasNext()) { // iterate through original states
			State sOther = (State) itOtherStates.next();
			State sClone = (State) cloneMap.get(sOther);
			Iterator itInputs = sOther.getInputs().iterator();
			while (itInputs.hasNext()) {
				Object currInput = itInputs.next();
				Iterator itNextStates = sOther.getNextStateSet(currInput)
						.iterator();
				while (itNextStates.hasNext()) {
					sClone.addTransition(currInput,
							(FSMStateInterface) cloneMap.get(itNextStates
									.next()));
				}
			}
		}
	}

	/**
	 * This method allows to set a type which specifies how the model would like
	 * to be visualised. The valid types are described in the BasicFA interface.
	 * Note: The type is only a default setting. A viewer/editor has not
	 * nescesserly to use it.
	 * 
	 * @param type
	 *            The visualisation type.
	 */
	public void setLayoutType(int type) {
		this.type = type;
	}

	/**
	 * Sets the name of the FABase.
	 * 
	 * @param name
	 *            The name of the FABase.
	 */
	public void setName(String name) {
		this.name = name;
	}

	public void setStartState(FSMStateInterface state) {
		if (state instanceof State)
			setStartState((State) state);
	}

	/**
	 * If the state does not belong to the FABase yet the method will throw an
	 * FAException. If the state is part of the FABase, it will be marked to be
	 * the <code>start state</code>. If the given parameter is null, the old
	 * start state is reset. Note: A FA can only have one
	 * <code>start state</code>!
	 * 
	 * @param state
	 *            The <code>start state</code>
	 * @exception FAException
	 *                if the given state is not part of the FABase.
	 */
	public void setStartState(State state) {
		if (state != null && !states.containsKey(state)) {
			Debug
					.showException(new Exception(
							"BasicFAModel.setStartState(): this state does not belong here. ")); //$NON-NLS-1$
			return;
		}
		if (startState != null && state != startState) {
			startState.setStart(false);
			notifyFSMModelListeners(FSMEvent.createSetStartStateEvent(state));
		}
		startState = state;
		state.setStart(true);
		if (state != null) {
			notifyFSMModelListeners(FSMEvent.createSetStartStateEvent(state));
		}
	}

	protected State stateCloning(BasicFA theClone, State tmp, Hashtable marks,
			boolean doCompl) {
		try {
			State newState = new State((String) tmp.getLabel());
			newState.setPosition(tmp.getX(), tmp.getY());
			theClone.addState(newState);
			marks.put(tmp, newState);

			if (getStartState() == tmp) {
				theClone.setStartState(newState);
			}

			boolean isAccepting = isAccepting(tmp);
			if (doCompl) {
				theClone.setAccepting(newState, !isAccepting);
			} else {
				theClone.setAccepting(newState, isAccepting);
			}

			// adding the transitions ...
			Iterator iter = alphabet.iterator();
			while (iter.hasNext()) {
				Character inp = (Character) iter.next();
				Set nextStates = tmp.getNextStateSet(inp);
				Iterator stateIter = nextStates.iterator();
				while (stateIter.hasNext()) {
					State nextState = (State) stateIter.next();
					State nextStateClone = null;
					if (!marks.containsKey(nextState)) {
						nextStateClone = stateCloning(theClone, nextState,
								marks, doCompl);
					} else {
						nextStateClone = (State) marks.get(nextState);
					}
					newState.addTransition(inp, nextStateClone);
				}
			}
			return newState;
		} catch (FAException e) {
			e.printStackTrace();
			System.exit(0);
			Debug.showException(e);
			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see ch.ethz.ifa.model.BasicFA#nOfStates()
	 */
	public int nOfStates() {
		return states.size();
	}

	public Set getStatesAfter(String inputString) {
		Set currSet = new HashSet();
		State currState = (State) getStartState();
		if (null == currState) {
			return currSet;
		}
		currSet.add(currState);
		// next line adds all states reachable from the startstate with epsilon
		// to currSet.
		currSet.addAll(currState.getEPSReachableStates(null));
		Character currChar;
		Set newSet;
		for (int i = 0; i < inputString.length(); i++) {
			currChar = new Character(inputString.charAt(i));
			newSet = State.getNextStateSet(currSet, currChar);
			// newSet now contains the states reachable from currSet with
			// currChar)
			currSet.clear();
			currSet.addAll(newSet);
			// next loop calculates the states reachable from newSet with
			// EPSILON.
			Iterator iter = newSet.iterator();
			while (iter.hasNext()) {
				State s = (State) iter.next();
				// the method getEPSReachableStates searches rekursively for
				// states reachable by epsilon-paths.
				// it puts reached states into the specified sets and stops, if
				// a reached state is already contained in the first set.
				s.getEPSReachableStates(currSet, currSet, s);
			}
			// currSet = newSet;
		}
		return currSet;
	}

	private static BasicFA cb = null;

	public void copy() {
		System.out.println("sdf");
		cb = (BasicFA) this.clone();
		System.out.println(cb);
	}

	public void paste() {
		setFA(cb);
	}

	public boolean isPasteable() {
		System.out.println("pasteable: " + (cb != null));
		return cb != null;
	}
}