/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JComponent;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;

import ch.ethz.exorciser.Debug;
import ch.ethz.exorciser.Messages;
import ch.ethz.exorciser.Shared;
import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.fsmgui.FSMControllerInterface;
import ch.ethz.exorciser.fsmgui.FSMEditorOperations;
import ch.ethz.exorciser.fsmgui.FSMModelInterface;
import ch.ethz.exorciser.fsmgui.FSMView;
import ch.ethz.exorciser.fsmgui.Transition;

/**
 * @author Reto Lamprecht
 * @author Christian Hfeli
 * @author Vincent Tscherter
 */
public class CustomTransitionPopup implements ActionListener {

    /** The transition that this menu is appearing on. */
    protected Transition concernedTrans;

    protected FSMEditorOperations editorOps;

    protected FSMView fascape;

    protected FSMModelInterface model;

    protected Alphabet alphabet;

    JPopupMenu menu;

    Character epsilon = Alphabet.EPSILON;

    public CustomTransitionPopup(FSMControllerInterface controller) {
        this.editorOps = controller.getFSMEditorOperations();
        this.fascape = controller.getView();
        this.model = controller.getModel();
        if (model.getAlphabet() == null) {
            Debug.showException(new Exception(
                    "CustomTransitionPopup only works with CharAlphabets, not with " //$NON-NLS-1$
                            + model.getAlphabet().toString())); //$NON-NLS-1$
        }
        menu = new JPopupMenu();
    }

    public JPopupMenu getPopup(Transition t) {
        this.concernedTrans = t;
        this.alphabet = model.getAlphabet();
        List existingInputs = new ArrayList();
        Transition already = fascape.getTransition(t.from, t.to);
        if (already != null) existingInputs = already.label.getInputs();
        boolean epsilonEnabled = (alphabet.containsEpsilon() && t.to != t.from);
        boolean emptyTrans = concernedTrans.label.number() < 1;
        menu.removeAll();
        AlphabetStats aStats = new AlphabetStats(alphabet);
        if (aStats.nOfLetters() > 6) {
            aStats.resetLetterCount();
            while (!aStats.allLettersDone()) {
                menu.add(getCharMenu(aStats.getNext7Letters(), existingInputs,
                        null));
            }
        } else
            addChars(aStats.getLetters(), existingInputs, menu);

        if (aStats.nOfNumbers() > 4) {
            menu.add(getCharMenu(aStats.getNumbers(), existingInputs, Messages
                    .getString("FSM.popup.numbers"))); //$NON-NLS-1$
        } else
            addChars(aStats.getNumbers(), existingInputs, menu);

        if (aStats.nOfSpecial() > 4) {
            menu.add(getCharMenu(aStats.getSpecial(), existingInputs, Messages
                    .getString("FSM.popup.special"))); //$NON-NLS-1$
        } else
            addChars(aStats.getSpecial(), existingInputs, menu);
        if (epsilonEnabled) {
            JCheckBoxMenuItem cmi = new JCheckBoxMenuItem(epsilon.toString());
            menu.add(cmi);
            cmi.addActionListener(this);
            cmi.setSelected(existingInputs.contains(epsilon));
        }

        JMenuItem mi = new JMenuItem(Messages.getString("FSM.popup.all")); //$NON-NLS-1$
        menu.add(mi);
        mi.addActionListener(this);

        if (!emptyTrans) {
            menu.addSeparator();
            mi = new JMenuItem(Shared.DELETE);
            menu.add(mi);
            mi.addActionListener(this);
        }

        return menu;
    }

    private String addChars(List v, List existingInputs, JComponent menu_) {
        String chars = ""; //$NON-NLS-1$
        JCheckBoxMenuItem cmi;
        Character current;
        for (int i = 0; i < v.size(); i++) {
            current = (Character) v.get(i);
            cmi = new JCheckBoxMenuItem(current.toString());
            menu_.add(cmi);
            chars += current.toString();
            cmi.addActionListener(this);
            if (existingInputs.contains(current)) {
                cmi.setState(true);
            }
        }
        return chars;
    }

    public JMenu getCharMenu(List v, List existingInputs, String name) {
        JMenu charMenu = new JMenu();
        String title = addChars(v, existingInputs, charMenu);
        if ("".equalsIgnoreCase(name) || null == name) { //$NON-NLS-1$
            charMenu.setText(title);
        } else {
            charMenu.setText(name);
        }
        return charMenu;
    }

    public void actionPerformed(ActionEvent e) {
        if (concernedTrans == null) { return; }

        String item = e.getActionCommand();
        JMenuItem source = (JMenuItem) (e.getSource());
        boolean emptyTrans = concernedTrans.label.number() < 1;

        if (item == null) {
        } else if (Shared.DELETE.equals(item)) {
            if (emptyTrans) {
                fascape.removeTransition(concernedTrans);
            } else {
                editorOps.setTransitionInput(concernedTrans, new Object[0]);
            }
        } else if (Messages.getString("FSM.popup.all").equals(item)) { //$NON-NLS-1$
            Object[] c;
            if (concernedTrans.to == concernedTrans.from
                    && alphabet.contains(Alphabet.EPSILON)) {
                c = new Character[alphabet.size() - 1];
            } else {
                c = new Character[alphabet.size()];
            }
            Iterator iter = alphabet.iterator();
            int i = 0;
            Character current;
            while (iter.hasNext()) {
                current = (Character) iter.next();
                if (!(concernedTrans.to == concernedTrans.from && Alphabet.EPSILON
                        .equals(current))) {
                    c[i++] = current;
                }
            }
            editorOps.setTransitionInput(concernedTrans, c);
        } else if (source instanceof JCheckBoxMenuItem) {
            if (((JCheckBoxMenuItem) source).getState()) {
                Character[] c = { new Character(item.charAt(0))};
                editorOps.addTransitionInput(concernedTrans, c);
            } else {
                editorOps.deleteTransition(concernedTrans.from.model,
                        concernedTrans.to.model, new Character(item.charAt(0)));
            }
        }
    }
}