/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import ch.ethz.exorciser.fsmgui.Alphabet;

/**
 * @author Reto Lamprecht
 * @author Christian Hfeli
 * @author Vincent Tscherter
 */
public class FAModel extends BasicFAModel implements FA {

	/**
	 * Constructs an FAModel. The FAModel will get the initial name
	 * <code>'sampleFA'</code>.
	 */
	public FAModel() {
		super(""); //$NON-NLS-1$
	}

	/**
	 * Constructs an empty FAModel with the given inital name.
	 *
	 * @param   name    The name of the FA
	 */
	public FAModel(String name) {
		super(name);
	}

	/*
	   * Returns a <code>Object</code> which is a clone of
	   * this <code>FA</code>. All states will be cloned too.
	   *
	   * @return      <code>clone</code> of the FA
	   */
	public Object clone() {
		Hashtable marks = new Hashtable();
		FAModel theClone = new FAModel();
		theClone.setAlphabet((Alphabet)this.getAlphabet().clone());
		theClone.setName(this.name);
		theClone.setLayoutType(this.getLayoutType());
		Iterator en = states.keySet().iterator();
		while (en.hasNext()) {
			State tmpState = (State) en.next();
			if (!marks.containsKey(tmpState)) {
				stateCloning(theClone, tmpState, marks, false);
			}
		}
		return theClone;
	}

	/**
	 * This method tests whether the Regular Language accepted by this FA is finite or not.
	 *
	 * @return	Returns <code>true</code> if the Regular Language accepted by this FA is finite, otherwise <code>false</code>. 
	 */
	public boolean isLanguageFinite() throws FAException {
		Character sentinel = new Character('$');
		Iterator iter;
		Stack stack;
		String label;
		State current, next;

		FA thisClone = (FA)this.clone();
		thisClone.makeDetOpt();
		thisClone.detMinimize();
		thisClone.reLabelAllStates("s"); //$NON-NLS-1$
		Alphabet alph = (Alphabet)thisClone.getAlphabet().clone();
		Alphabet epsAlph = (Alphabet)thisClone.getAlphabet().clone();
		epsAlph.add(sentinel);
		thisClone.setAlphabet(epsAlph);

		// Algorithm:
		//
		// Turn all edges round and search paths with circles from all accepting states to start state.
		// If such a circle can be found, the language is infinite.

		// Instead of turning all edges round, we insert sentinel-transitions in the opposite direction for all edges.
		iter = thisClone.getAllStates().iterator();
		while (iter.hasNext()) {
			current = (State)iter.next();
			Iterator alphIter = AlphabetUtil.getDetIterator(alph);
			while (alphIter.hasNext()) {
				Character ch = (Character)alphIter.next();
				if (ch != sentinel) {
					next = current.getNextState(ch);
					if (next != null) {
						next.addTransition(sentinel, current);
					}//if2
				}//if1
			}//while2
		}//while1

		// Start depth-first search from all accepting states
		boolean found = false;
		Iterator accIter = thisClone.getAcceptingStates().iterator();
		while (accIter.hasNext() && !found) {
			// Colour all states white
			// w = while  (not visited), g = gray   (visited), b = black  (on the path)
			iter = thisClone.getAllStates().iterator();
			while (iter.hasNext()) {
				current = (State)iter.next();
				current.setLabel("w" + current.getLabel()); //$NON-NLS-1$
			}//while

			current = (State)accIter.next();
			stack = new Stack();
			stack.push(current);
			while (!stack.empty() && !found) {
				current = (State)stack.peek();
				label = (String)current.getLabel();
				current.setLabel("b" + label.substring(label.indexOf("s"))); //$NON-NLS-1$ //$NON-NLS-2$
				iter = current.getNextStateSet(sentinel).iterator();
				while (iter.hasNext() && !found) {
					next = (State)iter.next();
					label = (String)next.getLabel();
					if (label.startsWith("w")) { //$NON-NLS-1$
						next.setLabel("g" + label.substring(label.indexOf("s"))); //$NON-NLS-1$ //$NON-NLS-2$
						stack.push(next);
					} else if (label.startsWith("b")) { //$NON-NLS-1$
						found = true;
					}//if..else
				}//while3
				if (current.equals(stack.peek())) {
					label = (String)current.getLabel();
					current.setLabel("g" + label.substring(label.indexOf("s"))); //$NON-NLS-1$ //$NON-NLS-2$
					stack.pop();
				}//if
			}//while2
		}//while1

		return (!found);
	}

	public boolean isLanguageAll() throws FAException {
		FA thisClone = (FA) this.clone();
		thisClone.makeDetOpt();
		thisClone.detMinimize();
		int nrOfStates = thisClone.getStateSet().size();
		State tmpSS = (State) thisClone.getStartState();
		if (nrOfStates == 1 && thisClone.isAccepting(tmpSS)) {
			Iterator alphIter =
				AlphabetUtil.getDetIterator(thisClone.getAlphabet());
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				Set nextStates = tmpSS.getNextStateSet(inp);
				if (!nextStates.contains(tmpSS)) {
					return false;
				}
			}
			return true;
		}
		return false;
	}

	/**
	 * Removes all states which can not be reached in this FA.
	 *
	 * @exception  FAException if no start state is defined in this FA.
	 */
	public void removeNotReachableStates() throws FAException {
		Set s = getNotReachableStates();
		Iterator iter = s.iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			this.deleteState(tmpState);
		}
	}

	/**
	 * Removes all trap states from this FA.
	 *
	 * @exception  FAException if no <code>start state</code> is defined in this FA.
	 */
	public void removeTrapStates() throws FAException {
		Iterator iter = this.getTrapStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			this.deleteState(tmpState);
		}
	}

	/**
	 * Returns the <code>type</code> of the FA. Valid types defnined in
	 * <code>FA</code> interface are:
	 *
	 * NOT_A_FA:    The model does not represent a valid FA. There is either no
	 *              states or no start <code>start state</code> defined.
	 * GFA:         The model represents a general nondeterminstic FA.
	 *              That means at least one epsilon transition is defined.
	 * NFA:         The model represents a nondeterminstic FA. More than one
	 *              transition on the same input at the same state is defined.
	 * UFA:         The model represents a nondeterminstic FA. But the FA is
	 *              underdefined. It's easy to complete the FA. All not defined
	 *              tranistion can be directed to a trap state which transforms
	 *              the UFA easily(linear time) to a DFA.
	 * DFA:         The model represents a determinstic FA.
	 * MFA:         The model represents a minimal determinstic FA.
	 *
	 * @return      <code>String</code> the type of the FA.
	 */
	public String getFAType() {
		if (states.size() == 0 || getStartState() == null) {
			return FA.NOT_A_FA;
		}
		if (this.hasEpsilonTransition()) {
			return FA.GFA;
		}
		try {
			if (this.hasMultipleTransitions()) {
				return FA.NFA;
			}
			if (!this.isDeterministic()) {
				return FA.UFA;
			}
			FA faClone = (FA) this.clone();
			faClone.detMinimize();
			if (faClone.getStateSet().size() != this.getStateSet().size()) {
				return FA.DFA;
			}
		} catch (FAException fae) {}
		return FA.MFA;
	}

	/**
	 * This method will transform an UFA to a DFA. If the FA is not an
	 * uncomplete NFA (UFA) the method will throw an exception. If the
	 * FA does not has any trap state a new trap state will be inserted.
	 *
	 * @exception  FAException if the FA is not an UFA.
	 */
	public void complete() throws FAException {
		if (this.isDeterministic()) {
			// The FA is already a complete DFA
			return;
		}
		if (this.hasEpsilonTransition() || this.hasMultipleTransitions()) {
			throw new FAException("The FA must be at least an UFA (underdetermined DFA) to complete it!"); //$NON-NLS-1$
		}
		State trapState = null;
		Iterator trapIter = this.getTrapStates().iterator();
		if (trapIter.hasNext()) {
			// There are already 0..n trap states, take on of them ...
			trapState = (State) trapIter.next();
		} else {
			trapState = new State("trap"); //$NON-NLS-1$
			addState(trapState);
			Iterator alphIter = getAlphabet().iterator();
			// create the transitions to the trap state itself...
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				if (!inp.equals(Alphabet.EPSILON)) {
					trapState.addTransition(inp, trapState);
				}
			}
		}
		Iterator iter = this.getStateSet().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			Iterator alphIter = getAlphabet().iterator();
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				Set nextStates = tmpState.getNextStateSet(inp);
				if (!inp.equals(Alphabet.EPSILON) && nextStates.size() == 0) {
					// add a transition to the trap-state.
					tmpState.addTransition(inp, trapState);
				}
			}
		}
	}

	/**
	 * This method will remove redundant trap states. If the given FA is
	 * nondeterministic the method will just remove all trap states. If the
	 * given FA is deterministic the method will merge trap states, so that
	 * the FA will have at maximum one trap state after execution.
	 *
	 * @exception  FAException if no <code>start state</code> is defined.
	 */
	public void mergeTrapStates() throws FAException {
		boolean isDetFA = this.isDeterministic();
		this.removeTrapStates();
		if (isDetFA) {
			this.complete();
		}
	}

	/**
	 * Builts the complement of FA. The FA must be deterministic, ohterwise
	 * an exception will be thrown. The method inverts the isAccepting value
	 * of all states.
	 *
	 * @exception  FAException if the FA is not deterministic.
	 */
	public void detComplement() throws FAException {
		if (!isDeterministic()) {
			throw new FAException("The FA(" + this.getName() + ") is not deterministic!"); //$NON-NLS-1$ //$NON-NLS-2$
		}
		Hashtable htClone = (Hashtable) this.states.clone();
		Iterator iter = htClone.keySet().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			setAccepting(tmpState, !isAccepting(tmpState));
		}
	}

	/**
	 * Unions the given FA with the FA itself. Both FA must be deterministic, ohterwise
	 * an exception will be thrown.
	 *
	 * @exception  FAException if one of the FA is not deterministic.
	 */
	public void detUnion(FA otherFA) throws FAException {
		FAModel other = (FAModel) otherFA;
		if (!isDeterministic()) {
			throw new FAException("The FA(" + this.getName() + ") is not deterministic!"); //$NON-NLS-1$ //$NON-NLS-2$
		}
		if (!otherFA.isDeterministic()) {
			throw new FAException(
				"The FA(" + otherFA.getName() + ") is not deterministic!"); //$NON-NLS-1$ //$NON-NLS-2$
		}

		this.removeNotReachableStates();
		String oldPrefix = (String) this.getStartState().getLabel();
		FAModel old = (FAModel) this.clone();

		clear();

		// The follwoing datastructure helps to handle the state-pairs.
		StatePairs pairs = new StatePairs();
		State thisStateNext = null;
		State otherStateNext = null;
		State newStateNext = null;
		State thisState = (State) old.getStartState();
		State otherState = (State) other.getStartState();

		State newState =
			new State((String) thisState.getLabel() + (String) otherState.getLabel());
		newState.setPosition(thisState.getX(), thisState.getY());
		pairs.addStatePair(thisState, otherState, newState);
		addState(newState);
		setStartState(newState);
		if (old.isAccepting(thisState) || other.isAccepting(otherState)) {
			setAccepting(newState, true);
		}

		State[] stp = pairs.getUnMarkedStatePair();
		while (stp != null) {
			thisState = stp[0];
			otherState = stp[1];
			newState = stp[2];
			Iterator detAlphIter = AlphabetUtil.getDetIterator(alphabet);
			while (detAlphIter.hasNext()) {
				Character inp = (Character) detAlphIter.next();
				// get the next state following the inp-transition
				thisStateNext = thisState.getNextState(inp);
				otherStateNext = otherState.getNextState(inp);

				// if this state-pairs does not already exists a new state will be created
				if (!pairs.containsStatePair(thisStateNext, otherStateNext)) {
					newStateNext =
						new State(
							(String) thisStateNext.getLabel() + (String) otherStateNext.getLabel());
					newStateNext.setPosition(thisStateNext.getX(), thisStateNext.getY());
					pairs.addStatePair(thisStateNext, otherStateNext, newStateNext);

					// add the newState to the FA ...
					addState(newStateNext);

					// and test if the new State is an accepting one
					if (old.isAccepting(thisStateNext) || other.isAccepting(otherStateNext)) {
						setAccepting(newStateNext, true);
					}
				} else {
					newStateNext = pairs.getCorrespondingState(thisStateNext, otherStateNext);
				}
				// adding the transitions
				// ... so, one looks at the next
				newState.addTransition(inp, newStateNext);
			}
			pairs.mark(thisState, otherState);
			stp = pairs.getUnMarkedStatePair();
		}
		this.reLabelAllStates(oldPrefix);
	}

	/**
	 * Returns true if the FA's are equivalent, means they recognize the same
	 * languages.Both FA must be deterministic, ohterwise an exception will be thrown.
	 *
	 * @exception  FAException if one of the FA is not deterministic.
	 */
	public boolean isEquivalent(FA other) throws FAException {
		FA thisClone = (FA) this.clone();
		thisClone.detComplement();
		thisClone.detUnion(other);
		thisClone.detComplement();
		if (!thisClone.isLanguageEmpty()) {
			return false;
		}

		FA otherClone = (FA) other.clone();
		otherClone.detComplement();
		otherClone.detUnion(this);
		otherClone.detComplement();
		if (!otherClone.isLanguageEmpty()) {
			return false;
		}
		return true;
	}

	/**
	 * Intersects the given FA with the FA itself. Both FA must be deterministic, ohterwise
	 * an exception will be thrown.
	 *
	 * @exception  FAException if one of the FA is not deterministic.
	 */
	public void detIntersection(FA other) throws FAException {
		this.detComplement();
		FA otherClone = (FA) other.clone();
		otherClone.detComplement();
		this.detUnion(otherClone);
		this.detComplement();
	}

    public void detMinus(FA other) throws FAException {
		FA otherClone = (FA) other.clone();
		otherClone.detComplement();
		this.detIntersection(otherClone);
	}
	
	private int findIndexInArray(Object[] array, Object obj) {
		for (int i = 0; i < array.length; i++) {
			if (array[i] == obj) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * Minimizes the given FA. The FA must be deterministic, ohterwise
	 * an exception will be thrown.
	 *
	 * @exception  FAException if the FA is not deterministic.
	 */
	public void detMinimize() throws FAException {
		if (!isDeterministic()) {
			throw new FAException("The FA is not deterministic!"); //$NON-NLS-1$
		}
		Object[] sl = getStateSet().toArray();
		Object[] ab = getAlphabet().toArray();
		boolean[][] tab = new boolean[sl.length][sl.length];
		for (int i = 0; i < sl.length; i++) {
			if (i == 0) { // check for pairwise accepting
				for (int j = 0; j < tab.length; j++) {
					for (int k = 0; k < tab.length; k++) {
						if (j < k) {
							State a = (State) sl[j];
							State b = (State) sl[k];
							tab[j][k] = a.isAccepting() ^ b.isAccepting();
						}
					}
				}
			} else {
				for (int j = 0; j < tab.length; j++) {
					for (int k = 0; k < tab.length; k++) {
						if (j < k && !tab[j][k]) {
							State a = (State) sl[j];
							State b = (State) sl[k];
							for (int l = 0; l < ab.length && !tab[j][k]; l++) {
								Character c = (Character) ab[l];
								State anext = a.getNextState(c);
								State bnext = b.getNextState(c);
								int q = findIndexInArray(sl, anext);
								int r = findIndexInArray(sl, bnext);
								if (q > r) {
									int t = q;
									q = r;
									r = t;
								}
								if (q != r) {
									tab[j][k] = tab[q][r];
								}
							}
						}
					}
				}
			}
		}
		for (int j = 0; j < tab.length; j++) {
			for (int k = 0; k < tab.length; k++) {
				if (j < k && !tab[j][k] && sl[j] != null && sl[k] != null) {
					State a = (State) sl[j];
					State b = (State) sl[k];
					for (int q = 0; q < sl.length; q++) {
						State s = (State) sl[q];
						if (s != null) {
							for (int l = 0; l < ab.length; l++) {
								Character c = (Character) ab[l];
								State next = s.getNextState(c);
								if (next == b) {
									s.removeTransition(c, b);
									s.addTransition(c, a);
								}
							}
						}
					}
					if (b.isStartState()) {
						setStartState(a);
					}
					removeState(b);
					sl[k] = null;
				}

			}
		}
	}

	public StatePairs getEquivalencePairs() throws FAException {
		if (!isDeterministic()) {
			throw new FAException("The FA is not deterministic!"); //$NON-NLS-1$
		}

		Hashtable lists = new Hashtable();
		StatePairs pairs = StatePairs.getStatePairs(this);
		Iterator iter = pairs.iterator();
		while (iter.hasNext()) {
			StatePair stp = (StatePair) iter.next();
			lists.put(stp, new ArrayList());
			if (this.isAccepting(stp.first)
				&& !this.isAccepting(stp.second)
				|| !this.isAccepting(stp.first)
				&& this.isAccepting(stp.second)) {
				stp.marked = true;
			}
		}
		iter = pairs.iterator();
		while (iter.hasNext()) {
			StatePair stp = (StatePair) iter.next();
			if (!stp.marked) {
				boolean allUnmarked = true;
				Iterator alphIter = AlphabetUtil.getDetIterator(alphabet);
				while (alphIter.hasNext()) {
					Character inp = (Character) alphIter.next();
					State firstNext = stp.first.getNextState(inp);
					State secondNext = stp.second.getNextState(inp);
					StatePair next = pairs.getStatePair(firstNext, secondNext);
					if (next != null && next.marked) {
						markStatePair(stp, lists);
						allUnmarked = false;
						break;
					}
				}
				if (allUnmarked) {
					alphIter = AlphabetUtil.getDetIterator(alphabet);
					while (alphIter.hasNext()) {
						Character inp = (Character) alphIter.next();
						State firstNext = stp.first.getNextState(inp);
						State secondNext = stp.second.getNextState(inp);
						StatePair next = pairs.getStatePair(firstNext, secondNext);
						if (next != null && next != stp) {
							((List) lists.get(next)).add(stp);
						}
					}
				}
			}
		}
		return pairs;
	}

	private void markStatePair(StatePair stp, Hashtable lists) {
		stp.marked = true;
		List list = (List) lists.get(stp);
		Iterator liter = list.iterator();
		while (liter.hasNext()) {
			StatePair lstp = (StatePair) liter.next();
			liter.remove();
			markStatePair(lstp, lists);
		}
	}

	public void removeEpsilon() throws FAException {
		String faType = this.getFAType();
		if (faType.equals(FA.NOT_A_FA)) {
			throw new FAException("Can not remove epsilons: This is not a valid FA"); //$NON-NLS-1$
		}
		if (!faType.equals(FA.GFA)) {
			return; // no epsilon transitions at all at the moment...
		}
		// First, remove all not reachable states and merge the trap states.
		this.removeNotReachableStates();
		this.removeTrapStates();

		Hashtable statesClone = (Hashtable) states.clone();
		Iterator iter = statesClone.keySet().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			Iterator alphIter = AlphabetUtil.getDetIterator(alphabet);
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();

				// Determine all states(reachables) which can be reached from
				// tmpState by following 0..n epsilon transitions and then
				// reading the input symbol 'inp'.
				Set reachables = tmpState.getEPSReachableStates(inp);

				// Add a new transition from tmpState to all the found reachable
				// states on the given input 'inp'.
				Iterator reachIter = reachables.iterator();
				while (reachIter.hasNext()) {
					State nextState = (State) reachIter.next();
					tmpState.addTransition(inp, nextState);
				}
			}
			// The set of accept states should contain all states
			// from wich one follows 0..n eps-transitions, will leed to
			// an accepting state.
			if (this.isAccepting(tmpState)) {
				Set epsPrec = tmpState.getEPSPreceders(this.getAllStates().iterator());
				Iterator epsPrecIter = epsPrec.iterator();
				while (epsPrecIter.hasNext()) {
					State precState = (State) epsPrecIter.next();
					this.setAccepting(precState, true);
				}
			}
		}

		// Deleting all epsilon transitions...
		iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			tmpState.deleteTransitions(Alphabet.EPSILON);
		}
		this.removeNotReachableStates();
		this.removeTrapStates();
	}

	private String calcLabel(HashSet subSet) {
		StringBuffer sb = new StringBuffer();
		Iterator iter = subSet.iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			String label = (String) tmpState.getLabel();
			sb.append(label);
		}
		if (sb.length() == 0) {
			sb.append("E"); //$NON-NLS-1$
		}
		return sb.toString();
	}

	private State combine(FA otherFA) throws FAException {
		State otherStartState = null;
		//add all the states from otherFA to this one...
		Hashtable pairs = new Hashtable();
		Iterator iter = otherFA.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			State newState = new State((String) tmpState.getLabel());
			newState.setPosition(tmpState.getX(), tmpState.getY());
			if (otherFA.getStartState() == tmpState) {
				otherStartState = newState;
			}

			this.addState(newState);
			boolean isAccepting = otherFA.isAccepting(tmpState);
			this.setAccepting(newState, isAccepting);
			pairs.put(tmpState, newState);
		}

		Iterator pIter = pairs.keySet().iterator();
		while (pIter.hasNext()) {
			State tmpState = (State) pIter.next();
			// adding the transitions ...
			Iterator alphIter = alphabet.iterator();
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();
				Set nextStates = tmpState.getNextStateSet(inp);
				Iterator stateIter = nextStates.iterator();
				while (stateIter.hasNext()) {
					State nextState = (State) stateIter.next();
					((State) pairs.get(tmpState)).addTransition(
						inp,
						((State) pairs.get(nextState)));
				}
			}
		}
		return otherStartState;
	}

	/**
	 * Makes a union on the FA's. The FA will be a GFA after this operation.
	 *
	 * @exception  FAException if the model does not represent a valid FA.
	 */
	public void nDetUnion(FA otherFA) throws FAException {
		if (!getAlphabet().contains(Alphabet.EPSILON)) {
			throw new FAException("The Alphabet does not allow EPSILON-transitions!"); //$NON-NLS-1$
		}
		State otherStartState = combine(otherFA);
		// create a new start state.
		State newStartState = new State("st"); //$NON-NLS-1$
		this.addState(newStartState);
		newStartState.addTransition(Alphabet.EPSILON, this.getStartState());
		newStartState.addTransition(Alphabet.EPSILON, otherStartState);
		this.setStartState(newStartState);
	}

	/**
	 * Makes a concatenation on the FA's. The FA will be a GFA after this operation.
	 *
	 * @exception  FAException if the model does not represent a valid FA.
	 */
	public void nDetConcatenation(FA otherFA) throws FAException {
		if (!getAlphabet().contains(Alphabet.EPSILON)) {
			throw new FAException("The Alphabet does not allow EPSILON-transitions!"); //$NON-NLS-1$
		}
		Iterator iter = this.getAcceptingStates().iterator();
		List oldAcc = new ArrayList();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			oldAcc.add(tmpState);
			this.setAccepting(tmpState, false);
		}
		State otherStartState = combine(otherFA);
		for (int i = 0; i < oldAcc.size(); i++) {
			State tmpState = (State) oldAcc.get(i);
			tmpState.addTransition(Alphabet.EPSILON, otherStartState);
		}
	}

	/**
	 * Makes a kleen star operation on the FA's. The FA will be a GFA after this operation.
	 *
	 * @exception  FAException if the model does not represent a valid FA.
	 */
	public void nDetStar() throws FAException {
		if (!getAlphabet().contains(Alphabet.EPSILON)) {
			throw new FAException("The Alphabet does not allow EPSILON-transitions!"); //$NON-NLS-1$
		}
		Iterator iter = this.getAcceptingStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			tmpState.addTransition(Alphabet.EPSILON, this.getStartState());
		}
		State newStartState = new State("st"); //$NON-NLS-1$
		this.addState(newStartState);
		newStartState.addTransition(Alphabet.EPSILON, this.getStartState());
		this.setStartState(newStartState);
		this.setAccepting(newStartState, true);
	}

	/**
	 * Makes the given FA deterministic without building unnecessary states.
	 *
	 * @exception  FAException if the model does not represent a valid FA.
	 */
	public void makeDetOpt() throws FAException {
		this.removeEpsilon();
		String oldStartStateLabel = (String) this.getStartState().getLabel();

		List newTrans = new ArrayList(); // for the new transitions
		Hashtable stateSets = new Hashtable(); // {state set; representor}
		Hashtable newStates = new Hashtable(); // {new state; is accept state}
		LinkedList toDo = new LinkedList(); // the list of unhandled states sets

		Iterator iter = this.getAllStates().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			HashSet newHS = new HashSet();
			newHS.add(tmpState);
			stateSets.put(newHS, tmpState);
			toDo.addLast(newHS);
		}

		while (!toDo.isEmpty()) {
			HashSet tmpSet = (HashSet) toDo.getFirst();
			State tmpState = (State) stateSets.get(tmpSet);

			Iterator alphIter = AlphabetUtil.getDetIterator(alphabet);
			while (alphIter.hasNext()) {
				Character inp = (Character) alphIter.next();

				Iterator tmpSetIter = tmpSet.iterator();
				HashSet nextStateSet = new HashSet();
				while (tmpSetIter.hasNext()) {
					State stateOfSet = (State) tmpSetIter.next();
					Set tmpSubSet = stateOfSet.getNextStateSet(inp);
					nextStateSet.addAll(tmpSubSet);
				}

				State nextState = (State) stateSets.get(nextStateSet);
				if (nextState == null) {
					nextState = new State(calcLabel(nextStateSet));
					stateSets.put(nextStateSet, nextState);
					toDo.addLast(nextStateSet);

					if (this.isOneOfTheseAccepting(nextStateSet)) {
						newStates.put(nextState, Boolean.TRUE);
					} else {
						newStates.put(nextState, Boolean.FALSE);
					}
				}
				newTrans.add(new Object[] { tmpState, nextState, inp });
			}
			toDo.removeFirst();
		}
		this.deleteAllTransitions();
		iter = newStates.keySet().iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			this.addState(tmpState);
			if (((Boolean) newStates.get(tmpState)).booleanValue()) {
				this.setAccepting(tmpState, true);
			}
		}

		for (int i = 0; i < newTrans.size(); i++) {
			Object[] objs = (Object[]) newTrans.get(i);
			State from = (State) objs[0];
			from.addTransition((Character) objs[2], (State) objs[1]);
		}
		this.removeNotReachableStates();
		this.reLabelAllStates(oldStartStateLabel);
	}

	public String findWitnessForNonEquivalence(FA other) {
		StringBuffer witness = new StringBuffer(""); //$NON-NLS-1$
		try {
			FAModel thisClone = (FAModel) this.clone();
			FAModel otherClone = (FAModel) other.clone();

			thisClone.makeDetOpt();
			thisClone.detMinimize();
			State thisStart = (State) thisClone.getStartState();

			otherClone.makeDetOpt();
			otherClone.detMinimize();
			State otherStart = (State) otherClone.getStartState();

			HashSet levelPairs = new HashSet();
			HashSet nextLevelPairs = new HashSet();
			StatePairs allPairs = new StatePairs();

			if (thisClone.isAccepting(thisStart) != otherClone.isAccepting(otherStart)) {
				return new String(new char[] { Alphabet.EPSILON.charValue()});
			}
			allPairs.addStatePair(thisStart, otherStart, null);
			Object[] newPair =
				new Object[] { thisStart, otherStart, Alphabet.EPSILON, null };
			nextLevelPairs.add(newPair);

			while (nextLevelPairs.size() != 0) {
				levelPairs = nextLevelPairs;
				nextLevelPairs = new HashSet();
				Iterator lpIter = levelPairs.iterator();
				while (lpIter.hasNext()) {
					Object[] pair = (Object[]) lpIter.next();
					Iterator alphIter =
						AlphabetUtil.getDetIterator(thisClone.getAlphabet());
					while (alphIter.hasNext()) {
						Character inp = (Character) alphIter.next();
						State thisNext = ((State) pair[0]).getNextState(inp);
						State otherNext = ((State) pair[1]).getNextState(inp);
						if (thisClone.isAccepting(thisNext) != otherClone.isAccepting(otherNext)) {
							// found the witness...
							witness.insert(0, inp);
							Object[] tmpPair = pair;
							while (tmpPair[3] != null) {
								witness.insert(0, ((Character) tmpPair[2]).charValue());
								tmpPair = (Object[]) tmpPair[3];
							}
							return witness.toString();
						}
						if (!allPairs.containsStatePair(thisNext, otherNext)) {
							allPairs.addStatePair(thisNext, otherNext, null);
							newPair = new Object[] { thisNext, otherNext, inp, pair };
							nextLevelPairs.add(newPair);
						}
					}
				}
			}
		} catch (FAException fae) {}
		return null;
	}




	
//	public boolean isWordInLanguage(String word) throws FAException {
//		char[] cArr = word.toCharArray();
//		State tmpState = (State) this.getStartState();
//		for (int i = 0; i < cArr.length; i++) {
//			Character inp = new Character(cArr[i]);
//			State nextState = tmpState.getNextState(inp);
//			tmpState = nextState;
//		}
//		if (this.isAccepting(tmpState)) {
//			return true;
//		}
//		return false;
//	}


}