/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import ch.ethz.exorciser.Debug;
import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.fsmgui.FSMModelInterface;
import ch.ethz.exorciser.fsmgui.FSMStateInterface;

public class FSMIO {

	static private final String NEWLINE = "\r\n"; //$NON-NLS-1$
	static private final String TAB = "    "; //$NON-NLS-1$
	static private final String XMLPROLOG = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"; //$NON-NLS-1$
	//	static private final String ENCODING = "UTF-8";

	static private final String FSM = "fsm"; //$NON-NLS-1$
	static private final String ALPHABET = "alphabet"; //$NON-NLS-1$
	static private final String SYMBOL = "symbol"; //$NON-NLS-1$
	static private final String STATESET = "states"; //$NON-NLS-1$
	static private final String STATE = "state"; //$NON-NLS-1$
	static private final String TRANSITIONSET = "transitions"; //$NON-NLS-1$
	static private final String TRANSITION = "transition"; //$NON-NLS-1$

	public static void saveFSM(FSMModelInterface m, OutputStream os) {
		try {
			Writer w = new OutputStreamWriter(os, "UTF-8"); //$NON-NLS-1$
			w.write(XMLPROLOG + NEWLINE);
			w.write("<" + FSM + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
			w.write("<" + ALPHABET + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
			Iterator i = m.getAlphabet().iterator();
			while (i.hasNext()) {
				Object symbol = i.next();
				w.write(TAB + "<" + SYMBOL + " value='" + symbol + "'/>" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
			}
			i = m.getAllStates().iterator();
			Map sn = new HashMap();
			Map ns = new HashMap();
			int j = 0;
			while (i.hasNext()) {
				Object state = i.next();
				sn.put(j + "", state); //$NON-NLS-1$
				ns.put(state, j + ""); //$NON-NLS-1$
				j++;
			}
			w.write("</" + ALPHABET + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
			w.write("<" + STATESET + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
			i = sn.keySet().iterator();
			while (i.hasNext()) {
				Object key = i.next();
				FSMStateInterface s = (FSMStateInterface) sn.get(key);
				w.write(
					TAB
						+ "<" //$NON-NLS-1$
						+ STATE
						+ " id='" //$NON-NLS-1$
						+ key
						+ "' label='" //$NON-NLS-1$
						+ s.getLabel()
						+ "' x='" //$NON-NLS-1$
						+ s.getX()
						+ "' y='" //$NON-NLS-1$
						+ s.getY()
						+ "' initial='" //$NON-NLS-1$
						+ s.isStartState()
						+ "' final='" //$NON-NLS-1$
						+ s.isAccepting()
						+ "'/>" //$NON-NLS-1$
						+ NEWLINE);
			}
			i = sn.keySet().iterator();
			w.write("</" + STATESET + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
			w.write("<" + TRANSITIONSET + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
			while (i.hasNext()) {
				Object key = i.next();
				FSMStateInterface s = (FSMStateInterface) sn.get(key);
				Iterator a = m.getAlphabet().iterator();
				while (a.hasNext()) {
					Object sym = a.next();
					Iterator n = s.getNextStateSet(sym).iterator();
					while (n.hasNext()) {
						w.write(
							TAB
								+ "<" //$NON-NLS-1$
								+ TRANSITION
								+ " from='" //$NON-NLS-1$
								+ key
								+ "' to='" //$NON-NLS-1$
								+ ns.get(n.next())
								+ "' symbol='" //$NON-NLS-1$
								+ sym
								+ "'/>" //$NON-NLS-1$
								+ NEWLINE);
					}
				}
			}
			w.write("</" + TRANSITIONSET + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
			w.write("</" + FSM + ">"); //$NON-NLS-1$ //$NON-NLS-2$
			w.flush();
		} catch (IOException e) {
			Debug.showException(e);
		}
	}

	static FA fa;
	static Map ids;
	// static Alphabet alpha;

	public static FA loadFSM(InputStream is) {
		try {
			fa = new FAModel();
			fa.setAlphabet(new Alphabet());
			ids = new HashMap();
			DefaultHandler handler = new DefaultHandler() {
				public void startElement(String namespaceURI, String localName, String qName, Attributes atts) {
					String name = localName;
					if ("".equals(name)) { //$NON-NLS-1$
						name = qName;
					}
					if (name.equals(SYMBOL)) {
						fa.getAlphabet().add(new Character(atts.getValue("value").charAt(0))); //$NON-NLS-1$
					}
					if (name.equals(STATE)) {
						State s = new State();
						s.setLabel(atts.getValue("label")); //$NON-NLS-1$
						s.setPosition(
							Double.valueOf(atts.getValue("x")).doubleValue(), //$NON-NLS-1$
							Double.valueOf(atts.getValue("y")).doubleValue()); //$NON-NLS-1$
						fa.addState(s);
						fa.setAccepting(s, atts.getValue("final").equals("true")); //$NON-NLS-1$ //$NON-NLS-2$
						if (atts.getValue("initial").equals("true")) { //$NON-NLS-1$ //$NON-NLS-2$
							fa.setStartState(s);
						}
						ids.put(atts.getValue("id"), s); //$NON-NLS-1$
					}
					if (name.equals(TRANSITION)) {
						State from = (State) ids.get(atts.getValue("from")); //$NON-NLS-1$
						State to = (State) ids.get(atts.getValue("to")); //$NON-NLS-1$
						Character sym = new Character(atts.getValue("symbol").charAt(0)); //$NON-NLS-1$
						from.addTransition(sym, to);
					}
				}
				public void endElement(String namespaceURI, String localName, String qName) {
					String name = localName;
					if ("".equals(name)) { //$NON-NLS-1$
						name = qName;
					}
					if (name.equals(ALPHABET)) {
					}
				}
			};
			SAXParser saxParser = SAXParserFactory.newInstance().newSAXParser();
			saxParser.parse(is, handler);
			return fa;
		} catch (ParserConfigurationException e) {
			Debug.showException(e);
		} catch (SAXException e) {
			Debug.showException(e);
		} catch (FactoryConfigurationError e) {
			e.printStackTrace();
		} catch (IOException e) {
			Debug.showException(e);
		}
		return null;
	}
}