/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;

import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.event.CaretEvent;

import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.fsmgui.FSMControllerInterface;
import ch.ethz.exorciser.fsmgui.FSMEditorOperations;
import ch.ethz.exorciser.fsmgui.FSMEvent;
import ch.ethz.exorciser.fsmgui.FSMModelInterface;
import ch.ethz.exorciser.fsmgui.FSMView;

/**
 * @author Reto Lamprecht
 * @author Christian Hfeli
 * @author Vincent Tscherter
 */
public class IFACompleteEditor extends JPanel implements MouseMotionListener {

	// this static member references the newest instance of IFACompleteEditor. It allows a MouseListener to call setFA() on that IFACompleteEditor.
	//public static IFACompleteEditor THE_REF;

	public static int SHOW_INPUTTRACKER = 1;
	public static int SHOW_MATRIX = 2;
	public static int SHOW_DIAGRAM = 4;
	public static int SHOW_BOTH = 4 + 2;
	public static int VIEWER = 8;
	public static int ALPHABET_EDITABLE = 16;

	InputTrackPanel inputTracker;
	boolean inputTrackerShowing;
	MatrixPanel matrixPanel;
	boolean matrixShowing;
	FSMView fascape;
	boolean diagramShowing;
	JSplitPane matrixDiagramSplit;
	boolean matrixAndDiagramShowing;

	boolean alphabetEditable = true;

	IFAMVCFactory mvcFactory;
	FSMModelInterface model;

	public IFACompleteEditor() {
		this(SHOW_DIAGRAM);
	}

	public IFACompleteEditor(int config) {

		Frame frame = new Frame();
		setLayout(new BorderLayout());

		if ((config & VIEWER) > 0) {
			mvcFactory = new ViewerMVCFactory(frame);
			matrixPanel = new ViewerMatrixPanel(this);
		} else {
			mvcFactory = new IFAMVCFactory(frame);
			matrixPanel = new MatrixPanel(this);
		}
		fascape = (FSMView) mvcFactory.getView();
		fascape.addMouseMotionListener(this);

		model = mvcFactory.getModel();
		((IFAFSMController) mvcFactory.getController()).createDefaultPopup(this, frame);

		inputTracker = new InputTrackPanel(this, InputTrackPanel.SMALL, 20);

		matrixPanel.setPreferredSize(new Dimension(100, 100));

		matrixDiagramSplit =
			new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, true, matrixPanel, fascape);
		matrixDiagramSplit.setDividerLocation(0.5);
		matrixDiagramSplit.setDividerSize(6);
		matrixDiagramSplit.setForeground(Color.white);
		matrixDiagramSplit.setBackground(Color.white);

		if (config == SHOW_INPUTTRACKER) {
			showInputTracker(true);
		}
		if (config == SHOW_BOTH) {
			showMatrixAndDiagram();
		} else if (config == SHOW_MATRIX) {
			showMatrix();
		} else if (config == SHOW_DIAGRAM) {
			showDiagram();
		} else {
			showDiagram();
		}

	}

	public void setMyBackground(Color bgrColor) {

		//mvcFactory.getView().setBackground(bgrColor);
		fascape.setMyBackground(bgrColor);
		matrixPanel.setBackground(bgrColor);
		inputTracker.setMyBackground(bgrColor);
	}

	public FSMModelInterface getModel() {
		return mvcFactory.getModel();
	}

	public FSMControllerInterface getController() {
		return mvcFactory.getController();
	}

	public FSMEditorOperations getFSMEditorOperations() {
		return mvcFactory.getController().getFSMEditorOperations();
	}

	public void setFA(FA fa) {
		if (null == fa || fa == model)
			return;
		this.model = fa;
		mvcFactory.setModel(fa);
		matrixPanel.setEditTarget((IFAFSMController) getController());
		inputTracker.setEditTarget(getModel());
		matrixPanel.notify(FSMEvent.createSetAlphabetEvent((Alphabet) null));
		inputTracker.notify(FSMEvent.createSetAlphabetEvent((Alphabet) null));

	}

	public FA getFA() {
		return (FA) mvcFactory.getModel();
	}

	public void mouseDragged(MouseEvent e) {
	}
	public void mouseMoved(MouseEvent e) {
		if (matrixPanel.anyCellsMarked()) {
			inputTracker.caretUpdate((CaretEvent) null);
		} //model.notifyFSMModelListeners(FSMEvent.createUnMarkAllEvent());
	}

	void activateInputTracker() {
		inputTracker.caretUpdate((javax.swing.event.CaretEvent) null);
	}

	private void validateAll() {
		model.notifyFSMModelListeners(FSMEvent.createUnMarkAllEvent());
		if (null != getParent())
			getParent().validate();
		else
			validate();
		repaint();
	}

	public void showInputTracker(boolean b) {
		if (b == inputTrackerShowing)
			return;

		inputTrackerShowing = b;
		if (b) {
			inputTracker.input.setText(""); //$NON-NLS-1$
			add(inputTracker, BorderLayout.SOUTH);
			inputTracker.caretUpdate((javax.swing.event.CaretEvent) null);
		} else {
			remove(inputTracker);
			//model.notifyFSMModelListeners(FSMEvent.createUnMarkAllEvent())
		}
		validateAll();
	}

	public boolean isInputTrackerShowing() {
		return inputTrackerShowing;
	}

	public void showMatrix() {
		if (matrixShowing)
			return;
		matrixShowing = true;
		diagramShowing = false;
		matrixAndDiagramShowing = false;
		remove(fascape);
		remove(matrixDiagramSplit);
		add(matrixPanel, BorderLayout.CENTER);
		validateAll();
	}

	public boolean isMatrixShowing() {
		return matrixShowing;
	}

	public void showDiagram() {
		if (diagramShowing)
			return;
		diagramShowing = true;
		matrixShowing = false;
		matrixAndDiagramShowing = false;
		remove(matrixPanel);
		remove(matrixDiagramSplit);
		add(fascape, BorderLayout.CENTER);
		validateAll();
	}

	public boolean isDiagramShowing() {
		return diagramShowing;
	}

	public void showMatrixAndDiagram() {
		if (matrixAndDiagramShowing)
			return;
		matrixAndDiagramShowing = true;
		diagramShowing = false;
		matrixShowing = false;
		remove(fascape);
		remove(matrixPanel);
		matrixDiagramSplit =
			new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, true, matrixPanel, fascape);
		matrixDiagramSplit.setDividerLocation(0.5);
		matrixDiagramSplit.setDividerSize(6);
		matrixDiagramSplit.setForeground(Color.white);
		matrixDiagramSplit.setBackground(Color.white);
		add(matrixDiagramSplit, BorderLayout.CENTER);
		matrixDiagramSplit.setDividerLocation(0.5);
		matrixDiagramSplit.validate();
		validateAll();
	}

	public boolean isMatrixAndDiagramShowing() {
		return matrixAndDiagramShowing;
	}

}