/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RoundRectangle2D;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.event.MouseInputListener;

import ch.ethz.exorciser.fsmgui.ComponentRenderer;
import ch.ethz.exorciser.fsmgui.FSMEditorOperations;
import ch.ethz.exorciser.fsmgui.FSMEvent;
import ch.ethz.exorciser.fsmgui.FSMModelInterface;
import ch.ethz.exorciser.fsmgui.FSMModelListener;

/**
 * @author Reto Lamprecht
 * @author Christian Hfeli
 * @author Vincent Tscherter
 */
public class MatrixPanel extends JPanel implements FSMModelListener,
        MouseInputListener {

    /**
     * if true, the cells are rebuilt from the model before painting them.
     * changed by FSMEvents that change table dimensions in notify().
     */
    private boolean REBUILD_TABLE = true;

    /**
     * if true, the size off cells are rebuilt from the model before painting
     * them. changed by FSMEvents that change cell content in notify().
     */
    private boolean RELAYOUT_TABLE = true;

    /**
     * true if any cell (and corresponding states and transitions in diagram)
     * are marked.
     */
    private boolean CELLS_MARKED = false;

    public boolean anyCellsMarked() {
        return CELLS_MARKED;
    }

    private boolean MUTE = false;

    public void setMute(boolean b) {
        MUTE = b;
    }

    Color BGR = new Color(255, 255, 255);

    private IFACompleteEditor ifaCmplEd;

    private FSMEditorOperations editorOps;

    private FSMModelInterface model;

    /*
     * all rows, one for each state. Each accesible from rows by the
     * corresponding state.
     */
    Hashtable rows = new Hashtable();

    Hashtable rowRects = new Hashtable();

    /* state in lexographical order of their labels. */
    State[] rowOrder = new State[0];

    /* number of columns */
    int nofCols;

    Cell lastSel;

    Cell grabbedCell;

    Cell draggedCell;

    Hashtable oldStateMark;

    Hashtable oldTransMarks;

    CellPopup cellPopup;

    DefaultMatrixPopup defaultMatrixPopup;

    ZoomNDragablePane zdPane = new ZoomNDragablePane();

    Rectangle tableBounds = new Rectangle(0, 0, 0, 0);

    /// Layout Information ///
    /* space around input symbol labels. */
    Insets inputInsets = new Insets(7, 12, 7, 12);

    /* space around state labels. */
    Insets stateInsets = new Insets(7, 12, 7, 12);

    /* Position of upper left corner of tabel (without labels). */
    int offsetX, offsetY;

    /* gap between cells */
    int cellGapX = 3;

    int cellGapY = 3;

    /* width of each colmn */
    int[] colWidths;

    public MatrixPanel(IFACompleteEditor ifaCmplEd) {
        this.ifaCmplEd = ifaCmplEd;
        this.editorOps = ifaCmplEd.getFSMEditorOperations();
        this.model = ifaCmplEd.getModel();
        this.model.addFSMModelListener(this);
        addMouseListener(this);
        addMouseMotionListener(this);

        cellPopup = new CellPopup(editorOps);
        defaultMatrixPopup = new DefaultMatrixPopup(ifaCmplEd);

        zdPane.posTranslation(false);
        this.setBackground(BGR);
    }

    public void setEditTarget(IFAFSMController controller) {
        this.editorOps = controller.getFSMEditorOperations();
        this.model.removeFSMModelListener(this);
        this.model = controller.getModel();
        this.model.addFSMModelListener(this);

        this.cellPopup.setEditTarget(controller.getFSMEditorOperations());
        this.defaultMatrixPopup.setEditTarget(controller);

        REBUILD_TABLE = true;
        repaint();
    }

    public void mouseMoved(MouseEvent e) {

        Cell c = getCellAt(zdPane.getModelX(e.getX()), zdPane.getModelY(e
                .getY()));
        if (null != c) {
            if (lastSel != c) {
                CELLS_MARKED = true;
                markDiagram(c);
                selectCell(c);
                lastSel = c;
            }
        } else {
            if (null != lastSel) {
                selectAllCells(Cell.NEUTRAL);
                CELLS_MARKED = false;
                lastSel = null;
            }
            if (!tableBounds.contains(e.getX(), e.getY())) {
                ifaCmplEd.activateInputTracker();
            }
        }
        repaint();
    }

    int zoomRef;

    int transRef;

    public void mousePressed(MouseEvent e) {
        int modif = e.getModifiers();

        // handle right mouseclicks resp. Ctrl-clicks
        if ((modif & InputEvent.BUTTON3_MASK) > 0
                || (((modif & InputEvent.BUTTON1_MASK) > 0) & ((modif & InputEvent.CTRL_MASK) > 0))) {
            Cell cell = getCellAt(zdPane.getModelX(e.getX()), zdPane
                    .getModelY(e.getY()));
            if (cell != null) {
                JPopupMenu popup = cellPopup.getPopup(cell, rowOrder);
                if (popup != null) popup.show(this, e.getX(), e.getY());
            } else
                defaultMatrixPopup.getPopup().show(this, e.getX(), e.getY());
        }
        // handle middle mouseclicks resp. Shift-clicks
        else if ((modif & InputEvent.BUTTON2_MASK) > 0
                || (((modif & InputEvent.BUTTON1_MASK) > 0) & ((modif & InputEvent.SHIFT_MASK) > 0))) {
            if (e.getClickCount() > 1) {
                zdPane.scale = 1.0;
                zdPane.dx = 0.0;
                zdPane.dy = 0.0;
            } else {
                zoomRef = e.getY();
                zdPane.setScaleRef(0, 0);
            }
        } else {
            grabbedCell = getCellAt(zdPane.getModelX(e.getX()), zdPane
                    .getModelY(e.getY()));
            if (null != grabbedCell)
                grabbedCell.grab(zdPane.getModelX(e.getX()), zdPane.getModelY(e
                        .getY()));
            else
                zdPane.setTranslationRef(0, e.getY());
        }
    }

    public void mouseDragged(MouseEvent e) {
        int modif = e.getModifiers();
        limitMousePos(e);
        // handle right mouseclicks resp. Ctrl-clicks
        if ((modif & InputEvent.BUTTON3_MASK) > 0
                || (((modif & InputEvent.BUTTON1_MASK) > 0) & ((modif & InputEvent.CTRL_MASK) > 0))) {
        }
        // handle middle mouseclicks resp. Shift-clicks
        else if ((modif & InputEvent.BUTTON2_MASK) > 0
                || (((modif & InputEvent.BUTTON1_MASK) > 0) & ((modif & InputEvent.SHIFT_MASK) > 0))) {
            zdPane.scale(e.getY() - zoomRef);
        } else if (grabbedCell != null) {
            draggedCell = new Cell(grabbedCell.state, grabbedCell.input);
            draggedCell.setSelected(true);
            draggedCell.setRect(zdPane.getModelX(e.getX())
                    - grabbedCell.grabOffsX, zdPane.getModelY(e.getY())
                    - grabbedCell.grabOffsY, (int) grabbedCell.rect.getWidth(),
                    (int) grabbedCell.rect.getHeight());
        } else {
            zdPane.translate(0, e.getY());
        }
        mouseMoved(e);
        repaint();
    }

    public void mouseReleased(MouseEvent e) {
        Cell c = getCellAt(zdPane.getModelX(e.getX()), zdPane.getModelY(e
                .getY()));
        if (null != draggedCell && null != c) {
            List trans = new ArrayList();
            Iterator iter = draggedCell.state
                    .getNextStateSet(draggedCell.input).iterator();
            while (iter.hasNext()) {
                trans.add(c.state);
                trans.add(iter.next());
                trans.add(c.input);
            }
            editorOps.addTransitions(trans);
        }
        draggedCell = null;
        repaint();
    }

    public void mouseClicked(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    /**
     * Limits mouse position to values inside fascape.
     */
    protected void limitMousePos(MouseEvent e) {
        if (e.getX() < 0) e.translatePoint(-e.getX(), 0);
        if (e.getY() < 0) e.translatePoint(0, -e.getY());
        if (e.getX() > this.getWidth())
                e.translatePoint(this.getWidth() - e.getX(), 0);
        if (e.getY() > this.getHeight())
                e.translatePoint(0, this.getHeight() - e.getY());
    }

    private Cell getCellAt(int x, int y) {
        Object[] row;
        Cell c;
        for (int i = 0; i < rowOrder.length; i++) {
            row = (Object[]) rows.get(rowOrder[i]);
            for (int j = 0; j < row.length; j++) {
                c = (Cell) row[j];
                if (c.contains(x, y)) { return c;
                //  j = row.length;
                //  i = rows.size();
                }
            }
        }
        return null;
    }

    /* Doesn't deselect anything */
    public void selectCell(State state, Character input) {
        Object[] row = (Object[]) rows.get(state);
        if (row != null) {
            for (int i = 0; i < row.length; i++) {
                if (input.equals(((Cell) row[i]).input)) {
                    ((Cell) row[i]).setSelected(true);
                }
            }
        }
    }

    public void setRowActive(State s) {
        Object[] row;
        row = (Object[]) rows.get(s);
        if (row != null) {
            for (int j = 0; j < row.length; j++) {
                ((Cell) row[j]).setActive(true);
            }
        }
    }

    public void deselectAllCells() {
        lastSel = null;
        Iterator ri = rows.keySet().iterator();
        while (ri.hasNext()) {
            selectRow((State) ri.next(), Cell.NEUTRAL);
        }
    }

    /* Deselects rest and selects referred state too. */
    private void selectCell(Cell c) {
        if (null == c) return;
        selectAllCells(Cell.NEUTRAL);
        Iterator iter = c.state.getNextStateSet(c.input).iterator();
        while (iter.hasNext()) {
            selectRow((State) iter.next(), Cell.REFERRED);
        }
        c.setSelected(true);
    }

    private void selectAllCells(int mode) {
        Iterator ri = rows.keySet().iterator();
        while (ri.hasNext()) {
            selectRow((State) ri.next(), mode);
        }
    }

    private void selectRow(State s, int mode) {
        Object[] row;
        row = (Object[]) rows.get(s);
        if (row != null) {
            for (int j = 0; j < row.length; j++) {
                ((Cell) row[j]).setTempAppearance(mode);
            }
        }
    }

    public void markDiagram(Cell c) {

        if (null != oldStateMark)
                model.notifyFSMModelListeners(FSMEvent
                        .createMarkEvent(oldStateMark));
        if (null != oldTransMarks)
                model.notifyFSMModelListeners(FSMEvent
                        .createMarkEvent(oldTransMarks));
        ifaCmplEd.activateInputTracker();
        oldStateMark = new Hashtable();
        oldTransMarks = new Hashtable();

        Hashtable newStateMark = new Hashtable();
        oldStateMark.put(c.state, ComponentRenderer.MARK_NOT_MARKED);
        newStateMark.put(c.state, ComponentRenderer.MARK_GREEN);

        Hashtable newTransMarks = new Hashtable();
        Iterator iter = c.state.getNextStateSet(c.input).iterator();
        while (iter.hasNext()) {
            Object[] trans = new Object[2];
            trans[0] = c.state;
            trans[1] = (State) iter.next();
            oldTransMarks.put(trans, ComponentRenderer.MARK_NOT_MARKED);
            newTransMarks.put(trans, ComponentRenderer.MARK_GREEN);
            oldStateMark.put(trans[1], ComponentRenderer.MARK_NOT_MARKED);
            newStateMark.put(trans[1], ComponentRenderer.MARK_BLUE);
        }

        model.notifyFSMModelListeners(FSMEvent.createMarkEvent(newStateMark));
        model.notifyFSMModelListeners(FSMEvent.createMarkEvent(newTransMarks));
    }

    public static State[] orderStates(Set stateSet) {
        //Hashtable h = new Hashtable();
        // sort labels first
        Iterator iter = stateSet.iterator();
        String[] labels = new String[stateSet.size()];
        for (int i = 0; iter.hasNext(); i++) {
            labels[i] = (String) ((State) iter.next()).getLabel();
        }
        java.util.Arrays.sort(labels);
        // now fill an ordered array of states
        iter = stateSet.iterator();
        State[] rowOrder = new State[stateSet.size()];
        State state;
        while (iter.hasNext()) {
            state = (State) iter.next();
            for (int i = 0; i < labels.length; i++) {
                if (labels[i] == state.getLabel()) {
                    while (rowOrder[i] != null) {
                        i++;
                    }
                    rowOrder[i] = state;
                    break;
                }
            }
        }
        /*
         * State[] rowOrder = new State[stateSet.size()]; for (int i = 0;
         * iter.hasNext(); i++) { rowOrder[i] = (State)iter.next(); }
         * java.util.Arrays.sort(rowOrder); /* state = (State)iter.next();
         * String s = state.getLabel(); h.put(s, state); } TreeSet o = new
         * TreeSet(h.keySet()); State[] rowOrder = new State[o.size()]; iter =
         * o.iterator(); int count = 0; while (iter.hasNext()) {
         * rowOrder[count++] = (State)h.get(iter.next()); }
         */
        return rowOrder;
    }

    private int getLongestStateLabelLength(Graphics g) {
        FontMetrics fm = g.getFontMetrics();
        int max = 0;
        Set stateSet = model.getAllStates();
        if (null != stateSet) {
            Iterator iter = stateSet.iterator();
            while (iter.hasNext()) {
                max = Math.max(max, fm.stringWidth((String) ((State) iter
                        .next()).getLabel()));
            }
        }
        return max;
    }

    private void updateWidthsNHeights(Graphics g) {
        FontMetrics fm = g.getFontMetrics();
        colWidths = new int[nofCols];
        Object[] row;
        Cell c;
        Rectangle2D sbounds;
        for (int i = 0; i < rowOrder.length; i++) {
            row = (Object[]) rows.get(rowOrder[i]);
            for (int j = 0; j < row.length; j++) {
                c = (Cell) row[j];
                sbounds = fm.getStringBounds(c.getRefString(), g);
                colWidths[j] = Math.max(colWidths[j], (int) sbounds.getWidth()
                        + Cell.insets.left + Cell.insets.right);
                Cell.height = Math.max(Cell.height, (int) sbounds.getHeight()
                        + Cell.insets.bottom + Cell.insets.top);
            }
        }
    }

    private void layoutTable(Graphics g) {
        RELAYOUT_TABLE = false;
        // wegnehmen, wenn model dynamisch upgedated wird...
        updateWidthsNHeights(g);
        FontMetrics fm = g.getFontMetrics();
        offsetX = getLongestStateLabelLength(g) + stateInsets.left
                + stateInsets.right;
        offsetY = fm.getHeight() + inputInsets.top + inputInsets.bottom;
        rowRects = new Hashtable();
        Object[] row;
        Cell c;
        int y = offsetY;
        for (int i = 0; i < rowOrder.length; i++) {
            row = (Object[]) rows.get(rowOrder[i]);
            int x = offsetX;
            for (int j = 0; j < row.length; j++) {
                c = (Cell) row[j];
                c.setAppearance(Cell.NEUTRAL);
                if (c.state.isAccepting()) c.setAccepting(true);
                if (c.state.isStartState()) c.setStart(true);
                c.setRect(x, y, colWidths[j], Cell.height);
                x += colWidths[j] + cellGapX;
            }
            rowRects.put(rowOrder[i], new RoundRectangle2D.Double(cellGapX, y,
                    x - cellGapX * 2, Cell.height, Cell.roundness,
                    Cell.roundness));
            y += Cell.height + cellGapY;
        }
        // update table bounds:
        tableBounds.x = offsetX;
        tableBounds.y = offsetY;
        tableBounds.height = rowOrder.length * Cell.height;
        tableBounds.width = 0;
        for (int j = 0; j < colWidths.length; j++) {
            tableBounds.width += colWidths[j];
        }
    }

    private void buildTable() {
        REBUILD_TABLE = false;
        RELAYOUT_TABLE = true;
        nofCols = model.getAlphabet().size();
        int colCount = 0;
        State s;
        Character input;
        rows = new Hashtable();
        Iterator alph;
        //Iterator nextStates;
        Set stateSet = model.getAllStates();
        if (null != stateSet) {
            Iterator sts = stateSet.iterator();
            while (sts.hasNext()) {
                Object[] row = new Object[nofCols];
                colCount = 0;
                s = (State) sts.next();
                rows.put(s, row);
                alph = model.getAlphabet().iterator();
                while (alph.hasNext()) {
                    input = (Character) alph.next();
                    Cell cell = new Cell(s, input);
                    row[colCount++] = cell;
                }
            }
        }
        rowOrder = orderStates(model.getAllStates());
    }

    public Dimension getPreferredSize() {
        Dimension tbounds = new Dimension(10, 25);
        //tbounds.height = rows.size() * Cell.height;
        if (null == colWidths || 0 == colWidths.length) {
            tbounds.width = stateInsets.left + stateInsets.right + 24
                    + model.getAlphabet().size()
                    * (inputInsets.left + inputInsets.right + 12);
        } else {
            for (int i = 0; i < colWidths.length; i++) {
                tbounds.width += colWidths[i];
            }
            return tbounds;
        }
        return tbounds;
    }

    /*
     * public Dimension getMinimumSize() { return new
     * Dimension(getPreferredSize().width, 25); }
     */

    public void paintComponent(Graphics g) {
        if (MUTE) return;
        //g.setClip(getBounds().x,getBounds().y, getBounds().width,
        // getBounds().height);
        super.paintComponent(g);

        if (REBUILD_TABLE) buildTable();
        if (RELAYOUT_TABLE) layoutTable(g);
        //g.drawString("#"+this.model.getName(), 20, 20);

        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);
        zdPane.transform(g2);

        // draw alphabet
        Iterator alph = model.getAlphabet().iterator();
        int w = offsetX;
        String input;
        int j = 0;
        g.setColor(Color.black);
        while (rows.size() > 0 && alph.hasNext()) {
            input = ((Character) alph.next()).toString();
            g.drawString(input, w + inputInsets.left, inputInsets.top
                    + g.getFontMetrics().getAscent());
            w += colWidths[j++] + cellGapX;
        }

        // draw cells

        g2.setStroke(Cell.narrowStroke);
        int h = offsetY;
        Object[] row;
        State currState;
        for (int i = 0; i < rowOrder.length; i++) {
            currState = rowOrder[i];
            row = (Object[]) rows.get(currState);
            if (1 == ((Cell) row[0]).referred) {
                g.setColor(Cell.getBoxColor(0, ((Cell) row[0]).referred, 0));
                g2.fill((RoundRectangle2D.Double) rowRects.get(rowOrder[i]));
                // RoundRect(cellGapX, h, offsetX + Cell.roundness, Cell.height,
                // Cell.roundness, Cell.roundness);
                g.setColor(Color.black);
            }
            if (1 == ((Cell) row[0]).active) {
                g.setColor(Cell.getBoxColor(0, 0, ((Cell) row[0]).active));
                g2.fill((RoundRectangle2D.Double) rowRects.get(rowOrder[i]));
                // RoundRect(cellGapX, h, offsetX + Cell.roundness, Cell.height,
                // Cell.roundness, Cell.roundness);
                g.setColor(Color.black);
            }
            g.drawString((String) currState.getLabel(), stateInsets.left, h
                    + stateInsets.top + g.getFontMetrics().getAscent());
            for (j = 0; j < row.length; j++) {
                ((Cell) row[j]).paint(g2);
            }
            h += Cell.height + cellGapY;
        }

        // draw draggedCell
        if (null != draggedCell) {
            g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
                    0.70f));
            //g2.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION,
            // RenderingHints.VALUE_ANTIALIAS_ON);
            draggedCell.paint(g2);
        }

    }

    public void notify(FSMEvent fae) {
        switch (fae.getId()) {
        case FSMEvent.ALPHABET_CHANGED:
            REBUILD_TABLE = true;
            break;
        case FSMEvent.STATE_ADDED:
            REBUILD_TABLE = true;
            break;
        case FSMEvent.STATE_DELETED:
            REBUILD_TABLE = true;
            break;
        case FSMEvent.STATE_POS_CHANGED:
            break;
        case FSMEvent.STATE_LABEL_CHANGED:
            REBUILD_TABLE = true;
            break;
        case FSMEvent.STATE_MODE_CHANGED:
            RELAYOUT_TABLE = true;
            break;
        case FSMEvent.START_STATE_CHANGED:
            RELAYOUT_TABLE = true;
            break;
        case FSMEvent.CLEAR_ALL:
            REBUILD_TABLE = true;
            break;
        case FSMEvent.TRANSITION_ADDED:
            REBUILD_TABLE = true;
            break;
        case FSMEvent.TRANSITION_DELETED:
            REBUILD_TABLE = true;
            break;
        case FSMEvent.FSM_NAME_CHANGED:
            break;
        case FSMEvent.MARK:
            break;
        case FSMEvent.UNMARK_ALL:
            deselectAllCells();
            break;
        }
        repaint();
    }

}