/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import ch.ethz.exorciser.fsmgui.Alphabet;

/**
 * Class that generates non minimal finite automata. Its abstract since it
 * contains only static methods, .
 * 
 * @author Alex Imfeld
 * @author Reto Lamprecht
 * @author Vincent Tscherter
 */

public abstract class NonMinFAGenerator {

	public static final int MAX_TRIES = 1000;
	public static Alphabet alphabet = Alphabet.create01Alphabet();
	public static int minimalNofStates = 6;
	public static int additionalRandomNofStates = 2;
	public static final double ACCEPTING_STATE_RATIO = 0.3;
	public static int minimalNofEquivalentStates = 1;
	public static int additionalRandomNofEquivalentStates = 2;

	public static FA generate() {
		int numStates = minimalNofStates + (int) (Math.random() * (1 + additionalRandomNofStates));
		int numEquiv = minimalNofEquivalentStates + (int) (Math.random() * (1 + additionalRandomNofEquivalentStates));

		boolean again;
		FA fa = null;
		do {
			again = false;
			try {
				fa = generate(numStates, ACCEPTING_STATE_RATIO, numEquiv, alphabet);
			} catch (FAException e) {

				again = true;
			}
		} while (again);
		return fa;

	}

	private static FA generate(int numStates, double acceptingStateRatio, int numEquiv, Alphabet alph) throws FAException {
		FA fa = null;
		//	try {
		fa = getRandomMinDFA(numStates - numEquiv, acceptingStateRatio, alph);
		Hashtable equivalentStates = new Hashtable();
		for (Iterator stateIter = fa.getAllStates().iterator(); stateIter.hasNext();) {
			State cur = (State) stateIter.next();
			List states = new ArrayList();
			states.add(cur);
			equivalentStates.put(cur, states);
		}
		for (int i = 0; i < numEquiv; i++) {
			List transitionsTo = new ArrayList();
			State origState = null;
			int cnt = 0;
			while (transitionsTo.size() < 2) {
				transitionsTo.clear();
				do {
					origState = fa.getRandomState();
				} while (fa.getTrapStates().contains(origState));
				for (Iterator stateIter = fa.getAllStates().iterator(); stateIter.hasNext();) {
					State cur = (State) stateIter.next();
					for (Iterator alphIter = alph.iterator(); alphIter.hasNext();) {
						Character inp = (Character) alphIter.next();
						if (fa.hasTransition(cur, origState, inp)) {
							transitionsTo.add(new Transition(cur, origState, inp));
						}
					}
				}
				if (++cnt > MAX_TRIES) {
					throw new FAException("No more state with two or more ingoing transitions found..."); //$NON-NLS-1$
				}
			}
			// determine transitions leaving the state
			List transitionsFrom = new ArrayList();
			for (Iterator alphIter = alph.iterator(); alphIter.hasNext();) {
				Character inp = (Character) alphIter.next();
				State nextState = origState.getNextState(inp);
				if (nextState != null) {
					transitionsFrom.add(new Transition(origState, nextState, inp));
				}
			}
			// add equivalent state
			State equivState = new State();
			fa.addState(equivState);
			equivState.setAccepting(origState.isAccepting());
			// update hashtable of equivalences
			for (Iterator iter = ((List) ((ArrayList) equivalentStates.get(origState)).clone()).iterator(); iter.hasNext();) {
				Object cur = iter.next();
				List states = (List) equivalentStates.get(cur);
				states.add(equivState);
				equivalentStates.put(cur, states);
			}
			equivalentStates.put(equivState, ((ArrayList) equivalentStates.get(origState)).clone());
			// delete all transitions to state
			for (int j = 0; j < transitionsTo.size(); j++) {
				Transition trans = (Transition) transitionsTo.get(j);
				trans.getFromState().deleteTransition(trans.getInput(), origState);
			}
			int index;
			Transition trans;
			// one transition has to go to state, but the transition must not
			// be a self-transition
			index = (int) (Math.random() * transitionsTo.size());
			trans = (Transition) transitionsTo.get(index);
			while (trans.getFromState() == origState) {
				index = (index + 1) % transitionsTo.size();
				trans = (Transition) transitionsTo.get(index);
			}
			trans.getFromState().addTransition(trans.getInput(), origState);
			transitionsTo.remove(index);
			// one transition has to go to equivState
			index = (int) (Math.random() * transitionsTo.size());
			trans = (Transition) transitionsTo.get(index);
			trans.getFromState().addTransition(trans.getInput(), equivState);
			transitionsTo.remove(index);
			// other transitions to state may go to any equivalent state
			for (int j = 0; j < transitionsTo.size(); j++) {
				trans = (Transition) transitionsTo.get(j);
				List states = (List) equivalentStates.get(origState);
				index = (int) (Math.random() * states.size());
				trans.getFromState().addTransition(trans.getInput(), (State) states.get(index));
			}
			// handle transitions from equivState
			for (int j = 0; j < transitionsFrom.size(); j++) {
				trans = (Transition) transitionsFrom.get(j);
				List states = (List) equivalentStates.get(trans.getToState());
				index = (int) (Math.random() * states.size());
				equivState.addTransition(trans.getInput(), (State) states.get(index));
			}
		}
		fa.reLabelAllStates("s"); //$NON-NLS-1$
		return fa;
	}

	/**
	 * Generate a random dfa that is minimal.
	 * 
	 * @param alph
	 *            Specifies the alphabet that should be used.
	 * @param numStates
	 *            How many states should the fa contain.
	 * @param acceptingStateRatio
	 *            How many states should be accepting.
	 */
	public static FA getRandomMinDFA(int numStates, double acceptingStateRatio, Alphabet alph) throws FAException {
		FA res = FAGenerator.getRandomDFA(numStates, acceptingStateRatio, alph);
		FA resTmp = (FA) res.clone();
		resTmp.detMinimize();
		while (res.getStateSet().size() != resTmp.getStateSet().size()) {
			res = FAGenerator.getRandomDFA(numStates, acceptingStateRatio, alph);
			resTmp = (FA) res.clone();
			resTmp.detMinimize();
		}
		return res;
	}

}