/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.fsmgui.FSMEvent;
import ch.ethz.exorciser.fsmgui.FSMStateInterface;

/**
 * @author Reto Lamprecht
 * @author Christian Hfeli
 * @author Vincent Tscherter
 */
public class State implements Cloneable, FSMStateInterface {

	private String label;
	private double x, y;
	private boolean ACCEPTING;
	private boolean START;
	private Hashtable transitions;
	private BasicFA fa;

	/**
	 * Constructs a new State object without a label-identification.
	 *
	 * @param       label     The label of the State object
	 */
	public State() {
		set("", Double.NaN, Double.NaN); //$NON-NLS-1$
	}

	/**
	 * Constructs a new State object with a label.
	 *
	 * @param       label     The label of the State object
	 */
	public State(String label) {
		set(label, Double.NaN, Double.NaN);
	}

	/**
	 * Constructs a new State object with a label and
	 * a given position.
	 *
	 * @param       label     The label of the State object
	 * @param       x         The x coordintate of the position
	 * @param       y         The y coordintate of the position
	 */
	public State(String label, double x, double y) {
		set(label, x, y);
	}

	private void set(String label, double x, double y) {
		this.label = label;
		this.x = x;
		this.y = y;
		transitions = new Hashtable();
		fa = null;
	}

	/** Clones a state without its transitions, because they should probably lead to other clones.
	 * */
	public Object clone() {
		State clone = new State(this.label, this.x, this.y);
		clone.fa = this.fa;
		clone.setAccepting(this.ACCEPTING);
		clone.setStart(this.START);
		return clone;
	}

	//////////////////////////////////////////////
	// ch.ethz.fsmgui.FSMStateInterface methods //
	//////////////////////////////////////////////
	//                                          //

	public void addTransition(Object input, FSMStateInterface target) {
		if (input instanceof Character)
			addTransition((Character) input, (State) target);
	}

	public void removeTransition(Object input, FSMStateInterface target) {
		if (input instanceof Character)
			deleteTransition((Character) input, (State) target);
	}

	public Set getInputs() {
		return transitions.keySet();
	}

	public Set getNextStateSet(Object input) {
		if (input instanceof Character) {
			return getNextStateSet((Character) input);
		}
		return null;
	}

	//public void setPosition(double x, double y);

	//public double getX();

	//public double getY();

	public void setLabel(Object label) {
		if (label instanceof String)
			setLabel((String) label);
	}

	public Object getLabel() {
		return label;
	}

	public void setAccepting(boolean b) {
		//this.fa.setAccepting(this, b);
		ACCEPTING = b;
		fa.notifyFSMModelListeners(FSMEvent.createSetAcceptingEvent(this, b));
	}

	public boolean isAccepting() {
		return ACCEPTING;
	}

	public void setStart(boolean b) {
		START = b;
	}

	public boolean isStartState() {
		return START;
	}

	//                                          //
	//////////////////////////////////////////////

	public void setFA(BasicFA fa) {
		this.fa = fa;
	}

	/**
	 * Adds a transition to an other State based on a input
	 *
	 * @param      input     The needed input for the transition
	 * @param      st        The state the transition should leed to.
	 */
	public void addTransition(Character input, State st) {
		if (fa != null && !fa.getAlphabet().contains(input)) {
			return;
		}
		Set nextStates = (Set) transitions.get(input);
		if (nextStates == null) {
			nextStates = new HashSet();
			transitions.put(input, nextStates);
		}

		if (!nextStates.contains(st)) {
			nextStates.add(st);
			if (fa != null) {
				fa.notifyFSMModelListeners(FSMEvent.createAddTransitionEvent(this, st, input));
			}
		}
	}

	public void deleteTransition(Character input, State st) {
		Set nextStates = (Set) transitions.get(input);
		if (nextStates != null && nextStates.contains(st)) {
			nextStates.remove(st);
			if (fa != null) {
				fa.notifyFSMModelListeners(
					FSMEvent.createDeleteTransitionEvent(this, st, input));
			}
		}
	}

	public void deleteTransitions(Character input) {
		Set nextStates = (Set) transitions.get(input);
		if (nextStates != null) {
			Iterator stateIter = nextStates.iterator();
			while (stateIter.hasNext()) {
				State tmpState = (State) stateIter.next();
				stateIter.remove();
				if (fa != null) {
					fa.notifyFSMModelListeners(
						FSMEvent.createDeleteTransitionEvent(this, tmpState, input));
				}
			}
		}
	}

	public void setPositionSilently(double x, double y) {
		this.x = x;
		this.y = y;
	}

	public void setPosition(double x, double y) {
		this.x = x;
		this.y = y;
		if (fa != null) {
			fa.notifyFSMModelListeners(FSMEvent.createStatePosChangedEvent(this));
		}
	}

	public void setLabel(String label) {
		this.label = label;
		if (fa != null) {
			fa.notifyFSMModelListeners(FSMEvent.createSetStateLabelEvent(this, label));
		}
	}

	public Set getTransitionLabels() {
		return transitions.keySet();
	}

	public double getX() {
		return x;
	}
	public double getY() {
		return y;
	}
	//public String getLabel(){return label;}

	public boolean hasTransitionOnInput(Character input) {
		return (getNextStateSet(input).size() != 0);
	}

	/** Returns true if no transition with a different symbol than those specified leads out of this state. */
	public boolean hasNoOtherTransitions(Collection inputs) {
		if (inputs == null)
			return true;
		Iterator iter = transitions.keySet().iterator();
		Character current;
		while (iter.hasNext()) {
			current = (Character) iter.next();
			if (getNextStateSet(current).size() != 0 && !inputs.contains(current))
				return false;
		}
		return true;
	}

	public Set getAllNeighbours() {
		HashSet res = new HashSet();
		Iterator iter = transitions.keySet().iterator();
		while (iter.hasNext()) {
			Character tmpInp = (Character) iter.next();
			Set tmpSet = getNextStateSet(tmpInp);
			Iterator stateIter = tmpSet.iterator();
			while (stateIter.hasNext()) {
				res.add(stateIter.next());
			}
		}
		return res;
	}

	/**
	 * Returns a set of states which can be reached from this state
	 * by following 0..n eps-transitions and then reading the given input (may be null).
	 * 
	 *
	 * @param      inp          The input.
	 *
	 * @return     HashSet      The set of the founded states.
	 */
	public Set getEPSReachableStates(Character inp) {
		Set epsReachableOnes =
			getEPSReachableStates(new HashSet(), new HashSet(), this);
		if (inp == null) {
			return epsReachableOnes;
		}
		HashSet res = new HashSet();
		Iterator iter = epsReachableOnes.iterator();
		while (iter.hasNext()) {
			State tmpState = (State) iter.next();
			res.addAll(tmpState.getNextStateSet(inp));
		}
		return res;
	}

	public Set getEPSReachableStates(Set visited, Set res, State state) {
		Set epsNN = state.getNextStateSet(Alphabet.EPSILON);
		Iterator nnIter = epsNN.iterator();
		while (nnIter.hasNext()) {
			State tmpState = (State) nnIter.next();
			if (!visited.contains(tmpState)) {
				visited.add(tmpState);
				res.add(tmpState);
				getEPSReachableStates(visited, res, tmpState);
			}
		}
		return res;
	}

	/**
	 * Returns a set of states, which can be followed by 0..n eps-transitions
	 * and will leed to this state.
	 *
	 * @param      inp          The input.
	 *
	 * @return     HashSet      The set of the founded states.
	 */
	public Set getEPSPreceders(Iterator allStatesIter) {
		HashSet res = new HashSet();
		while (allStatesIter.hasNext()) {
			State tmpState = (State) allStatesIter.next();
			if (tmpState != this) {
				Set epsReachableOnes =
					getEPSReachableStates(new HashSet(), new HashSet(), tmpState);
				if (epsReachableOnes.contains(this)) {
					res.add(tmpState);
				}
			}
		}
		return res;
	}

	public State getNextState(Character input) {
		if (input == null)
			return null;
		Set nextStates = (Set) transitions.get(input);
		if (nextStates != null && !nextStates.isEmpty()) {
			Iterator iter = nextStates.iterator();
			return (State) iter.next();
		}
		return null;
	}

	public Set getNextStateSet(Character input) {
		HashSet res = new HashSet();
		Set nextStates = (Set) transitions.get(input);
		if (nextStates != null) {
			res.addAll(nextStates);
		}
		return res;
	}

	public static Set getNextStateSet(Set actSet, Character inp) {
		HashSet res = new HashSet();
		Iterator iter = actSet.iterator();
		while (iter.hasNext()) {
			State tmp = (State) iter.next();
			Set s = tmp.getNextStateSet(inp);
			res.addAll(s);
		}
		return res;
	}
}