/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.ifa;

import java.awt.Graphics2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;

/**
 * Usage for scale: Call setScaleRef(x, y) in mousePressed() with the current
 * mouse position. Call scale(y) in mouseDragged() with the y-coordinate of the
 * current mouse position. Call transform(Graphics2D) before painting.
 * 
 * Analog usage for translation.
 * 
 * @author Reto Lamprecht
 * @author Christian Hfeli
 * @author Vincent Tscherter
 */
public class ZoomNDragablePane {

	boolean NEGTRANS = true;
	boolean POSTRANS = true;

	double scale = 1;
	double dx = 0.0, dy = 0.0;

	/** The scale factor at the beginning of a scale drag. */
	double refScale;
	double refDx, refDy;

	/*
	 * the distance in pixels that a user has to drag the mouse to double or
	 * half the scale factor.
	 */
	double dScaleDist = 150.0;

	/* The scale factor is limited */
	double maxScale = 8.0;
	double minScale = 0.125;

	/** coordinates where a zoom or tanslation gesture started. */
	double transRefX, transRefY;
	double scaleRefX, scaleRefY;
	//double scaleCenterX, scaleCenterY;

	public ZoomNDragablePane() {
	}

	/**
	 * Sets the distance in pixels that a user has to drag the mouse to double
	 * or half the scale factor.
	 */
	public void setDoubleScaleDistance(int distance) {
		dScaleDist = distance;
	}

	/** Sets the limits of the scale factor */
	public void setScaleRange(double min, double max) {
		minScale = min;
		maxScale = max;
	}

	public void negTranslation(boolean b) {
		NEGTRANS = b;
	}

	public void posTranslation(boolean b) {
		POSTRANS = b;
	}

	/** Sets the coordinates, where a translation gesture started. */
	public void setTranslationRef(int x, int y) {
		transRefX = x;
		transRefY = y;
		refDx = dx;
		refDy = dy;
	}

	/** Sets the coordinates, where a translation gesture started. */
	public void setTranslationRef(double x, double y) {
		transRefX = x;
		transRefY = y;
		refDx = dx;
		refDy = dy;
	}

	/**
	 * calculates the new translation values according to the specified mouse
	 * position.
	 */
	public void translate(int x, int y) {
		dx = refDx + x - transRefX;
		dy = refDy + y - transRefY;
		if (!NEGTRANS && dx < 0)
			dx = 0;
		if (!NEGTRANS && dy < 0)
			dy = 0;
		if (!POSTRANS && dx > 0)
			dx = 0;
		if (!POSTRANS && dy > 0)
			dy = 0;
	}

	/***************************************************************************
	 * Sets the center of scale. public void setScaleCenter(int x, int y) {
	 * CENTER_SPECIFIED = true; scaleCenterX = x; scaleCenterY = y; } * Removes
	 * the center of scale. public void removeScaleCenter(int x, int y) {
	 * CENTER_SPECIFIED = false; }
	 * 
	 * Sets the coordinates, where a scale gesture started. *
	 **************************************************************************/
	public void setScaleRef(int x, int y) {
		scaleRefX = x;
		scaleRefY = y;
		refScale = scale;
	}

	/** Sets the coordinates, where a scale gesture started. */
	public void setScaleRef(double x, double y) {
		scaleRefX = x;
		scaleRefY = y;
		refScale = scale;
	}

	/**
	 * calculates the new scale value according to the specified mouse
	 * position.
	 */
	public void scale(int y) {
		// calc scale from dragged distance
		double newScale;
		if (scaleRefY < y) {
			newScale = refScale * (1 + 2 * (y - scaleRefY) / dScaleDist);
		} else {
			newScale = refScale * (1 - 0.5 * (scaleRefY - y) / dScaleDist);
		}
		if (newScale < minScale)
			newScale = minScale;
		if (newScale > maxScale)
			newScale = maxScale;

		Point2D modScaleRef = getModelCoords(scaleRefX, scaleRefY);
		double diffX = scaleRefX - (modScaleRef.getX() * newScale + dx);
		double diffY = scaleRefY - (modScaleRef.getY() * newScale + dy);
		dx = dx + diffX;
		dy = dy + diffY;
		scale = newScale;
	}

	/** returns model coordinates of the specified view coordinates. */
	public Point2D getModelCoords(double x, double y) {
		double modX = (x - dx) / scale;
		double modY = (y - dy) / scale;
		return new Point2D.Double(modX, modY);
	}

	/** returns model coordinates of the specified view coordinates. */
	public int getModelX(int x) {
		double modX = (x - dx) / scale;
		return (int) modX;
	}

	/** returns model coordinates of the specified view coordinates. */
	public int getModelY(int y) {
		double modY = (y - dy) / scale;
		return (int) modY;
	}

	/** returns view coordinates of the specified model coordinates. */
	public Point2D getViewCoords(double x, double y) {
		double viewX = x * scale + dx;
		double viewY = y * scale + dy;
		return new Point2D.Double(viewX, viewY);
	}

	/**
	 * Returns a GeneralPath representing built from the points in the
	 * specifies array, transformed according to the current scale and
	 * translation.
	 */
	public GeneralPath makeTransformedPath(float[] coords) {
		GeneralPath path = new GeneralPath();
		path.reset();
		path.moveTo(
			(float) (coords[0] * scale + dx),
			(float) (coords[1] * scale + dy));
		for (int i = 2; i < coords.length - 1; i += 2) {
			path.lineTo(
				(float) (coords[i] * scale + dx),
				(float) (coords[i + 1] * scale + dy));
		}
		path.closePath();
		return path;
	}

	/**
	 * Transforms a Graphics2D Objext according to the current scale and
	 * translation.
	 */
	public void transform(Graphics2D g) {
		g.translate((int) dx, (int) dy);
		g.scale(scale, scale);
	}
}