/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.markov;

import javax.swing.JEditorPane;

import ch.ethz.exorciser.Messages;
import ch.ethz.exorciser.Notifications;
import ch.ethz.exorciser.Shared;
import ch.ethz.exorciser.markov.grammar.Grammar;
import ch.ethz.exorciser.markov.runtime.CheckerExecution;

/**
 * @author Vincent Tscherter
 */
public class Checker extends Thread {

   private Object[] alphabeth;

   private int[] word;

   private final int MAX = 20;

   private int l, feedbackStep;

   private boolean random = false;

   private long test = 0;

   private boolean error = false;

   private Grammar answer;

   private Exercise ex;

   private Notifications notifications;

   private Markov markov;

   static String TITLE = Messages.getString("Markov.checker.checking.title"); //$NON-NLS-1$

   static String NO_DIFF = Messages.getString("Markov.checker.no_diff.title"); //$NON-NLS-1$

   public Checker (Markov markov, boolean random) {
      answer = markov.editor.getGrammar();
      ex = markov.exercise;
      notifications = markov.notifications;
      this.markov = markov;
      if (markov.editor.getGrammar().getProductions().size() > 15) {
         feedbackStep = 1000;
      } else feedbackStep = 1000;
      this.random = random;
   }

   private JEditorPane ep;

   public void run () {
      //
      notifications.removeAll();
      ep = notifications.addNotification(TITLE, ""); //$NON-NLS-1$ //$NON-NLS-2$
      if (answer.getProductions().size() == 0) {
         notifications.addErrorNotification(Messages
               .getString("Markov.checker.noting_to_check"), //$NON-NLS-1$
               Messages.getString("Markov.checker.no_productions")); //$NON-NLS-1$
         return;
      }
      // check if alphabeth is correct
      for (int i = 0; i < answer.getAlphabet().size(); i++) {
         if (!ex.solution.getAlphabet().contains(answer.getAlphabet().get(i))) {
            notifications.addErrorNotification(Messages
                  .getString("Markov.checker.illegal_symbol.title"), //$NON-NLS-1$
                  Messages.getString("Markov.checker.illegal_symbol") //$NON-NLS-1$
                        .replaceAll("\\$X", answer.getAlphabet().get(i) //$NON-NLS-1$
                              .toString()));
            return;
         }
      }
      for (int i = 0; i < ex.solution.getAlphabet().size(); i++) {
         if (!answer.getAlphabet().contains(ex.solution.getAlphabet().get(i))) {
            notifications.addErrorNotification(Messages
                  .getString("Markov.checker.incomplete.aphabeth.title"), //$NON-NLS-1$
                  Messages.getString("Markov.checker.incomplete.aphabeth") //$NON-NLS-1$
                        .replaceAll("\\$X", //$NON-NLS-1$
                              ex.solution.getAlphabet() //$NON-NLS-1$
                                    .get(i).toString())); //$NON-NLS-2$
            return;
         }
      }
      // check inputs exhaustively
      alphabeth = answer.getAlphabet().toArray();
      initGenerator();
      String input = nextWord();
      String outputUser, outputSolution, outputInputChecker;
      while ((input != null) && !interupt) {
         if (ex.checker.getProductions().size() > 0) {
            outputInputChecker = checkInput(input);
         } else {
            outputInputChecker = ""; //$NON-NLS-1$
         }
         if (outputInputChecker.equals("") && !interupt) { //$NON-NLS-1$
            CheckerExecution ce = new CheckerExecution(answer, input,
                  ex.constraints);
            outputUser = ce.run();
            ce = new CheckerExecution(ex.solution, input, ex.constraints);
            outputSolution = ce.run();
            if (!(outputUser.equals(outputSolution))) {
               notifications.removeAll();
               error = true;
               if (input.equals("")) //$NON-NLS-1$
               notifications.addErrorNotification(Shared.ERROR, Messages
                     .getString("Markov.checker.check_empty_string")); //$NON-NLS-1$
               else notifications.addErrorNotification(Shared.ERROR, Messages
                     .getString("Markov.checker.check_input") //$NON-NLS-1$
                     .replaceAll("\\$INPUT", //$NON-NLS-1$
                           input));
               return;
            }
         }
         input = nextWord();
      }
      if (!interupt) {
         notifications.removeAll();
         notifications.addCorrectNotification(NO_DIFF, Messages.getString(
               "Markov.checker.no_diff_l") //$NON-NLS-1$
               .replaceAll("\\$L", l + "")); //$NON-NLS-1$//$NON-NLS-2$
         markov.t = null;
      }
   }

   private String checkInput (String input) {
      CheckerExecution ce = new CheckerExecution(ex.checker, input,
            ex.constraints);
      return ce.run();
   }

   public void interrupt () {
      stop = true;
      interupt = true;
      if (!error) {
         notifications.removeAll();
         if (random) {
            notifications.addNotification(NO_DIFF, Messages.getString(
                  "Markov.checker.no_diff_n") //$NON-NLS-1$
                  .replaceAll("\\$N", test + "")); //$NON-NLS-1$ //$NON-NLS-2$
         } else {
            notifications.addNotification(NO_DIFF, Messages.getString(
                  "Markov.checker.no_diff_l") //$NON-NLS-1$
                  .replaceAll("\\$L", l + "")); //$NON-NLS-1$ //$NON-NLS-2$
         }
      }
      super.interrupt();
   }

   private void initGenerator () {
      l = 0;
      word = new int[MAX];
      stop = false;
      int i = 0;
      while (i < word.length) {
         word[i] = -1;
         i++;
      }
   }

   private boolean stop;

   private boolean interupt;

   private String nextWord () {
      if (stop || interupt) { return null; }
      String s = ""; //$NON-NLS-1$
      if (random) {
         int i = 0;
         int max = (int) (Math.random() * MAX);
         while (i < max) {
            s += alphabeth[(int) (Math.random() * alphabeth.length)];
            i++;
         }
         test++;
         if (test % feedbackStep == 0) {
            ep.setText(Notifications.createFeedback(TITLE, Messages.getString(
                  "Markov.checker.checking_n") //$NON-NLS-1$
                  .replaceAll("\\$N", test + ""))); //$NON-NLS-1$ //$NON-NLS-2$
         }
      } else {
         int i = 0;
         while (i < word.length && word[i] != -1) {
            s += alphabeth[word[i]];
            i++;
         }
         // INC WORD
         i = 0;
         int c = 1;
         while (i < word.length && c == 1) {
            if (word[i] == -1) {
               l++;
               ep.setText(Notifications.createFeedback(TITLE, Messages
                     .getString("Markov.checker.checking_l") //$NON-NLS-1$
                     .replaceAll("\\$L", //$NON-NLS-1$
                           l + ""))); //$NON-NLS-1$
            }
            word[i]++;
            if (word[i] == alphabeth.length) {
               word[i] = 0;
            } else {
               c = 0;
            }
            i++;
         }
         if (c == 1) {
            stop = true;
         }
      }
      return s;
   }
}