/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.markov.grammar;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Reto Lamprecht
 * @author Vincent Tscherter
 */
public class Grammar {

	public static char EPSILON = (char) ('\u03b1' + 4);

	private List variables = new ArrayList();

	private List terminals = new ArrayList();

	protected List productions = new ArrayList();

	public List getMarkers() {
		return this.variables;
	}

	private boolean addVariable(Object o) {
		if (!variables.contains(o)) {
			variables.add(o);
			return true;
		}
		return false;
	}

	private void removeVariable(Object o) {
		variables.remove(o);
		removeSymbolFromProductions(o);
	}

	public boolean isInMarkers(Object o) {
		return variables.contains(o);
	}

	public List getAlphabet() {
		return this.terminals;
	}

	private boolean addTerminal(Object o) {
		if (!terminals.contains(o)) {
			terminals.add(o);
			return true;
		}
		return false;
	}

	private void removeTerminal(Object o) {
		terminals.remove(o);
		removeSymbolFromProductions(o);
	}

	public boolean isInAlphabeth(Object o) {
		return terminals.contains(o);
	}

	public List getProductions() {
		return this.productions;
	}

	private boolean addProduction(int i, Production p) {
		if (!productions.contains(p) && i <= productions.size()) {
			productions.add(i, p);
			return true;
		}
		return false;
	}

	private void removeProduction(Production p) {
		productions.remove(p);
	}

	private List listeners = new ArrayList();

	public void addGrammarListener(GrammarListener gl) {
		listeners.add(gl);
	}

	public void removeGrammarListener(GrammarListener gl) {
		listeners.remove(gl);
	}

	private void notifyAlgorithmListeners(GrammarChange gc) {
		for (int i = 0; i < listeners.size(); i++) {
			((GrammarListener) listeners.get(i)).notify(gc);
		}
	}

	private void notifyAnythingChangedListeners() {
		for (int i = 0; i < listeners.size(); i++) {
			((GrammarListener) listeners.get(i)).notifyAnythingChanged();
		}
	}

	public Grammar() {
		this(new Object[0], new Object[0], new Object[0]);
	}

	public Grammar(Object[] asMarkers, Object[] asAlphabeth,
			Object[] asProductions) {
		int i = 0;
		for (i = 0; i < asMarkers.length; i++)
			this.variables.add(asMarkers[i]);
		for (i = 0; i < asAlphabeth.length; i++)
			this.terminals.add(asAlphabeth[i]);
		for (i = 0; i < asProductions.length; i++) {
			this.productions.add(new Production((String) asProductions[i]));
		}
	}

	public void executeGrammarChange(GrammarChange gc) {
		boolean propagate = true;
		notifyAnythingChangedListeners();
		switch (gc.getId()) {
		case GrammarChange.CHANGE_PRODUCTION:
			gc.getProduction().setLeft(gc.getLeft());
			gc.getProduction().setRight(gc.getRight());
			if ("".equals(gc.getProduction().leftAsString()) //$NON-NLS-1$
					&& "".equals(gc.getProduction().rightAsString())) { //$NON-NLS-1$
				executeGrammarChange(GrammarChange.createRemoveProduction(this,
						gc.getProduction()));
				propagate = false;
			}
			break;
		case (GrammarChange.CHANGE_PRODUCTION_LEFT):
			gc.getProduction().setLeft(gc.getLeft());
			if ("".equals(gc.getProduction().leftAsString()) //$NON-NLS-1$
					&& "".equals(gc.getProduction().rightAsString())) { //$NON-NLS-1$
				executeGrammarChange(GrammarChange.createRemoveProduction(this,
						gc.getProduction()));
				propagate = false;
			}
			break;
		case GrammarChange.CHANGE_PRODUCTION_RIGHT:
			gc.getProduction().setRight(gc.getRight());
			if ("".equals(gc.getProduction().leftAsString()) //$NON-NLS-1$
					&& "".equals(gc.getProduction().rightAsString())) { //$NON-NLS-1$
				executeGrammarChange(GrammarChange.createRemoveProduction(this,
						gc.getProduction()));
				propagate = false;
			}
			break;
		case GrammarChange.ADD_PRODUCTION:
			addProduction(gc.getProdIndex(), gc.getProduction());
			break;
		case GrammarChange.REMOVE_PRODUCTION:
			removeProduction(gc.getProduction());
			break;
		case GrammarChange.MOVE_PRODUCTION:
			moveProduction(gc.getFrom(), gc.getTo());
			break;
		case GrammarChange.CLEAR_ALL:
			productions.clear();
			break;
		case GrammarChange.ADD_VAR:
			addVariable(gc.getVariable());
			break;
		case GrammarChange.REMOVE_VAR:
			removeVariable(gc.getVariable());
			break;
		case GrammarChange.ADD_TERM:
			addTerminal(gc.getTerminal());
			break;
		case GrammarChange.REMOVE_TERM:
			removeTerminal(gc.getTerminal());
			break;
		case GrammarChange.RENAME_SYM:
			renameSymbol(gc.getOldSym(), gc.getNewSym());
			break;
		case GrammarChange.SET_STOP_PRODUCTION:
			gc.getProduction().addAttribute("STOP"); //$NON-NLS-1$
			break;
		case GrammarChange.SET_NONSTOP_PRODUCTION:
			gc.getProduction().removeAttribute("STOP"); //$NON-NLS-1$
			break;
		default:
			break;
		}
		if (propagate)
			notifyAlgorithmListeners(gc);
	}

	private void renameSymbol(Object o, Object n) {
		List vt;
		if (isInMarkers(o))
			vt = variables;
		else if (isInAlphabeth(o))
			vt = terminals;
		else
			return;
		int i = vt.indexOf(o);
		vt.remove(o);
		if (!vt.contains(n)) {
			vt.add(i, n);
		}
		renameSymbolsInProductions(o, n);
	}

	private void renameSymbolsInProductions(Object oldS, Object newS) {
		for (int i = 0; i < productions.size(); i++) {
			Production current = (Production) productions.get(i);
			current.renameOccurrencesOf(oldS, newS);
		}
	}

	private void removeSymbolFromProductions(Object o) {
		for (int i = 0; i < productions.size(); i++) {
			Production current = (Production) productions.get(i);
			current.removeOccurrencesOf(o);
			if (current.isEmpty())
				productions.remove(current);
		}
	}

	private void moveProduction(int from, int to) {
		if (from == to || from > productions.size() || to > productions.size())
			return;
		productions.add(to, productions.remove(from));
	}

	public Object clone() {
		String[] p = new String[productions.size()];
		for (int i = 0; i < p.length; i++) {
			p[i] = ((Production) productions.get(i)).toString();
		}
		return new Grammar(variables.toArray(), terminals.toArray(), p);
	}
}