/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.markov.grammar;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import ch.ethz.exorciser.Debug;

public class GrammarIO {

   static private final String NEWLINE = "\r\n"; //$NON-NLS-1$

   static private final String TAB = "    "; //$NON-NLS-1$

   static private final String XMLPROLOG = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"; //$NON-NLS-1$

   static private final String GRAMMAR_TAG = "grammar"; //$NON-NLS-1$

   static private final String ALPHABET_TAG = "alphabet"; //$NON-NLS-1$

   static private final String MARKERS_TAG = "markers"; //$NON-NLS-1$

   static private final String SYMBOL_TAG = "symbol"; //$NON-NLS-1$

   static private final String PRODUCTION_TAG = "production"; //$NON-NLS-1$

   static private final String LEFT_TAG = "left"; //$NON-NLS-1$

   static private final String RIGHT_TAG = "right"; //$NON-NLS-1$

   static private final String VALUE_ATTRUBUTE = "value"; //$NON-NLS-1$

   public static void saveGrammar (Grammar g, OutputStream os) {
      Writer w = null;
      try {
         w = new OutputStreamWriter(os, "UTF-8"); //$NON-NLS-1$
         List v;
         String s = ""; //$NON-NLS-1$
         w.write(XMLPROLOG + NEWLINE);
         w.write("<" + GRAMMAR_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
         v = g.getAlphabet();
         w.write(TAB + "<" + ALPHABET_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
         for (int i = 0; i < v.size(); i++) {
            w.write(TAB + TAB + "<" //$NON-NLS-1$
                  + SYMBOL_TAG + " " //$NON-NLS-1$
                  + VALUE_ATTRUBUTE + "=\"" //$NON-NLS-1$
                  + v.get(i) + "\"/>" //$NON-NLS-1$
                  + NEWLINE);
         }
         w.write(TAB + "</" + ALPHABET_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
         w.write(TAB + "<" + MARKERS_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
         v = g.getMarkers();
         for (int i = 0; i < v.size(); i++) {
            w.write(TAB + TAB + "<" //$NON-NLS-1$
                  + SYMBOL_TAG + " " //$NON-NLS-1$
                  + VALUE_ATTRUBUTE + "=\"" //$NON-NLS-1$
                  + v.get(i) + "\"/>" //$NON-NLS-1$
                  + NEWLINE);
         }
         w.write(TAB + "</" + MARKERS_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
         //
         v = g.getProductions();
         for (int i = 0; i < v.size(); i++) {
            w.write(TAB + "<" + PRODUCTION_TAG); //$NON-NLS-1$
            Production prod = (Production) v.get(i);
            if (prod.hasAttribute("STOP")) { //$NON-NLS-1$
               w.write(" stop=\"true\">" + NEWLINE); //$NON-NLS-1$
            } else {
               w.write(" stop=\"false\">" + NEWLINE); //$NON-NLS-1$
            }
            s = prod.leftAsString();
            w.write(TAB + TAB + "<" + LEFT_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
            for (int j = 0; j < s.length(); j++) {
               w.write(TAB + TAB + TAB + "<" //$NON-NLS-1$
                     + SYMBOL_TAG + " value=\"" //$NON-NLS-1$
                     + s.substring(j, j + 1) + "\"/>" //$NON-NLS-1$
                     + NEWLINE);
            }
            w.write(TAB + TAB + "</" + LEFT_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
            s = prod.rightAsString();
            w.write(TAB + TAB + "<" + RIGHT_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
            for (int j = 0; j < s.length(); j++) {
               w.write(TAB + TAB + TAB + "<" //$NON-NLS-1$
                     + SYMBOL_TAG + " " //$NON-NLS-1$
                     + VALUE_ATTRUBUTE + "=\"" //$NON-NLS-1$
                     + s.substring(j, j + 1) + "\"/>" //$NON-NLS-1$
                     + NEWLINE);
            }
            w.write(TAB + TAB + "</" + RIGHT_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
            w.write(TAB + "</" + PRODUCTION_TAG + ">" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
         }
         w.write("</" + GRAMMAR_TAG + ">"); //$NON-NLS-1$ //$NON-NLS-2$
         w.flush();
      } catch (IOException e) {
         Debug.showException(e);
      } finally {
         if (w != null) {
            try {
               w.close();
            } catch (IOException e1) {
               e1.printStackTrace();
            }
         }
      }
   }

   public static String LatexValue (Object obj) {
      if (obj.equals("\u03B1")) { //$NON-NLS-1$
      return "\\alpha"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B2")) { //$NON-NLS-1$
      return "\\beta"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B3")) { //$NON-NLS-1$
      return "\\gamma"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B4")) { //$NON-NLS-1$
      return "\\delta"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B5")) { //$NON-NLS-1$
      return "\\zeta"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B6")) { //$NON-NLS-1$
      return "\\eta"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B7")) { //$NON-NLS-1$
      return "\\theta"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B8")) { //$NON-NLS-1$
      return "\\vartheta"; //$NON-NLS-1$
      }
      if (obj.equals("\u03B9")) { //$NON-NLS-1$
      return "\\iota"; //$NON-NLS-1$
      }
      if (obj.equals("\u03BA")) { //$NON-NLS-1$
      return "\\kappa"; //$NON-NLS-1$
      }
      if (obj.equals("\u03BB")) { //$NON-NLS-1$
      return "\\lambda"; //$NON-NLS-1$
      }
      if (obj.equals("\u03BC")) { //$NON-NLS-1$
      return "\\mu"; //$NON-NLS-1$
      }
      if (obj.equals("\u03BD")) { //$NON-NLS-1$
      return "\\nu"; //$NON-NLS-1$
      }
      if (obj.equals("\u03BE")) { //$NON-NLS-1$
      return "\\xio"; //$NON-NLS-1$
      }
      if (obj.equals("\u03BF")) { //$NON-NLS-1$
      return "o"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C0")) { //$NON-NLS-1$
      return "\\pi"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C1")) { //$NON-NLS-1$
      return "\\rho"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C2")) { //$NON-NLS-1$
      return "\\sigma"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C3")) { //$NON-NLS-1$
      return "\\varsigma"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C4")) { //$NON-NLS-1$
      return "\\tau"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C5")) { //$NON-NLS-1$
      return "\\upsilon"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C6")) { //$NON-NLS-1$
      return "\\phi"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C7")) { //$NON-NLS-1$
      return "\\chi"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C8")) { //$NON-NLS-1$
      return "\\psi"; //$NON-NLS-1$
      }
      if (obj.equals("\u03C9")) { //$NON-NLS-1$
      return "\\omega"; //$NON-NLS-1$
      }
      if (obj.equals("#")) { //$NON-NLS-1$
      return "\\sharp"; //$NON-NLS-1$
      }
      if (obj.equals("}")) { //$NON-NLS-1$
      return "\\}"; //$NON-NLS-1$
      }
      if (obj.equals("{")) { //$NON-NLS-1$
      return "\\{"; //$NON-NLS-1$
      }
      return obj + ""; //$NON-NLS-1$
   }

   public static void saveGrammarToLatex (Grammar g, OutputStream os) {
      Writer w = null;
      try {
         w = new OutputStreamWriter(os, "UTF-8"); //$NON-NLS-1$
         List v = new ArrayList();
         String s = ""; //$NON-NLS-1$
         w.write("% Generated LaTex Code / Exorciser / V. Tscherter" //$NON-NLS-1$
               + NEWLINE);
         v = g.getAlphabet();
         w.write("${\\cal A}=\\{"); //$NON-NLS-1$
         if (v.size() > 0) {
            w.write("" + LatexValue(v.get(0))); //$NON-NLS-1$
         }
         for (int i = 1; i < v.size(); i++) {
            w.write(", " + LatexValue(v.get(i))); //$NON-NLS-1$
         }
         w.write("\\}$ \\\\" + NEWLINE); //$NON-NLS-1$
         w.write("${\\cal M}=\\{"); //$NON-NLS-1$
         v = g.getMarkers();
         if (v.size() > 0) {
            w.write("" + LatexValue(v.get(0))); //$NON-NLS-1$
         }
         for (int i = 1; i < v.size(); i++) {
            w.write(", " + LatexValue(v.get(i))); //$NON-NLS-1$
         }
         w.write("\\}$ \\\\" + NEWLINE + "$\\begin{array}{rrcl}" + NEWLINE); //$NON-NLS-1$ //$NON-NLS-2$
         //
         v = g.getProductions();
         for (int i = 0; i < v.size(); i++) {
            w.write(i + ": & "); //$NON-NLS-1$
            Production prod = (Production) v.get(i);
            s = prod.leftAsString();
            if (s.length() == 0) {
               w.write(" \\varepsilon"); //$NON-NLS-1$
            }
            for (int j = 0; j < s.length(); j++) {
               w.write(" " + LatexValue(s.substring(j, j + 1))); //$NON-NLS-1$
            }
            w.write(" & "); //$NON-NLS-1$
            if (prod.hasAttribute("STOP")) { //$NON-NLS-1$
               w.write(" \\dashv "); //$NON-NLS-1$
            } else {
               w.write(" \\rightarrow "); //$NON-NLS-1$
            }
            w.write(" & "); //$NON-NLS-1$
            s = prod.rightAsString();
            if (s.length() == 0) {
               w.write(" \\varepsilon"); //$NON-NLS-1$
            }
            for (int j = 0; j < s.length(); j++) {
               w.write(" " + LatexValue(s.substring(j, j + 1))); //$NON-NLS-1$
            }
            w.write(" "); //$NON-NLS-1$
            w.write("\\\\" + NEWLINE); //$NON-NLS-1$
         }
         w.write("\\end{array}$" + NEWLINE); //$NON-NLS-1$
         w.flush();
      } catch (IOException e) {
         Debug.showException(e);
      } finally {
         if (w != null) {
            try {
               w.close();
            } catch (IOException e1) {
               e1.printStackTrace();
            }
         }
      }
   }

   static List alphabet = new ArrayList();

   static List markers = new ArrayList();

   static List productions = new ArrayList();

   public static Grammar loadGrammar (InputStream is) {
      productions.clear();
      alphabet.clear();
      markers.clear();
      try {
         DefaultHandler handler = new DefaultHandler() {

            private boolean alpha = false;

            private boolean marks = false;

            private String s;

            private String arrow;

            public void startElement (String namespaceURI, String localName,
                  String qName, Attributes atts) {
               String name = localName;
               if ("".equals(name)) { //$NON-NLS-1$
                  name = qName;
               }
               if (name.equals(ALPHABET_TAG)) {
                  alpha = true;
               } else if (name.equals(MARKERS_TAG)) {
                  marks = true;
               } else if (name.equals(SYMBOL_TAG) && alpha) {
                  alphabet.add(atts.getValue(VALUE_ATTRUBUTE));
               } else if (name.equals(SYMBOL_TAG) && marks) {
                  markers.add(atts.getValue(VALUE_ATTRUBUTE));
               } else if (name.equals(PRODUCTION_TAG)) {
                  if (atts.getValue("stop").equals("true")) { //$NON-NLS-1$ //$NON-NLS-2$
                     arrow = "-|"; //$NON-NLS-1$
                  } else {
                     arrow = "->"; //$NON-NLS-1$
                  }
               } else if (name.equals(LEFT_TAG)) {
                  s = ""; //$NON-NLS-1$
               } else if (name.equals(RIGHT_TAG)) {
                  s += arrow + " "; //$NON-NLS-1$
               } else if (name.equals(SYMBOL_TAG)) {
                  s += atts.getValue(VALUE_ATTRUBUTE) + " "; //$NON-NLS-1$
               }
            }

            public void endElement (String namespaceURI, String localName,
                  String qName) {
               String name = localName;
               if ("".equals(name)) { //$NON-NLS-1$
                  name = qName;
               }
               if (name.equals(ALPHABET_TAG)) {
                  alpha = false;
               } else if (name.equals(MARKERS_TAG)) {
                  marks = false;
               } else if (name.equals(PRODUCTION_TAG)) {
                  productions.add(s);
               }
            }
         };
         SAXParser saxParser = SAXParserFactory.newInstance().newSAXParser();
         saxParser.parse(is, handler);
         return new Grammar(markers.toArray(), alphabet.toArray(), productions
               .toArray());
      } catch (ParserConfigurationException e) {
         Debug.showException(e);
      } catch (SAXException e) {
         Debug.showException(e);
      } catch (FactoryConfigurationError e) {
         e.printStackTrace();
      } catch (IOException e) {
         Debug.showException(e);
      }
      return null;
   }
}