/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.markov.grammar;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.StringTokenizer;

/**
 * @author Reto Lamprecht
 * @author Vincent Tscherter
 */
public class Production {

   private Object[] left = null;

   private Object[] right = null;

   public Object[] getLeft () {
      return left;
   }

   public Object[] getRight () {
      return right;
   }

   private String leftAsString = ""; //$NON-NLS-1$

   private String rightAsString = ""; //$NON-NLS-1$

   public String leftAsString () {
      return leftAsString;
   }

   public String rightAsString () {
      return rightAsString;
   }

   /**
    * You can assign arbitrary Objects as attributes to a Production. There are
    * two kinds of attributes: key-only and key-value attributes. attributes of
    * the former kind can be set by the method addAttribute(key) and queried by
    * the method hasAttribute(). attributes of the latter kind can be set by the
    * method setAttribute(key, value) and queried by the method
    * getAttribute(key). An example of a key-only attribute is the String "STOP"
    * for a markov production that stops. An example of a key-value attribute
    * are the key "MARK" and a value of type Color.
    */
   private Hashtable attributes = new Hashtable();

   public Object setAttribute (Object key, Object value) {
      return attributes.put(key, value);
   }

   public Object getAttribute (Object key) {
      return attributes.get(key);
   }

   public void addAttribute (Object key) {
      attributes.put(key, "attribute has no value"); //$NON-NLS-1$
   }

   public boolean hasAttribute (Object key) {
      return attributes.containsKey(key);
   }

   public Object removeAttribute (Object key) {
      return attributes.remove(key);
   }

   private List vTmp = new ArrayList(); // a Vector for any occasion

   /**
    * The String representation of a Production looks like e.g.: E -> [ E ] All
    * Symbols of the left an right side need to be separted by spaces!
    */
   public Production (String sProd) {
      List _left = new ArrayList();
      List _right = new ArrayList();
      StringTokenizer stok = new StringTokenizer(sProd);
      String sTmp = ""; //$NON-NLS-1$
      boolean bLeft = true;
      while (stok.hasMoreTokens()) {
         sTmp = stok.nextToken();
         if (sTmp.equalsIgnoreCase("->")) { //$NON-NLS-1$
            bLeft = false;
            continue;
         }
         if (sTmp.equalsIgnoreCase("-|")) { //$NON-NLS-1$
            bLeft = false;
            this.addAttribute("STOP"); //$NON-NLS-1$
         } else {
            if (bLeft) {
               _left.add(sTmp.trim());
               leftAsString += sTmp.trim();
            } else {
               _right.add(sTmp.trim());
               rightAsString += sTmp.trim();
            }
         }
      }
      this.left = _left.toArray();
      this.right = _right.toArray();
   }

   public Production (Object[] left, Object[] right) {
      this.left = left;
      this.right = right;
      this.leftAsString = objArrayToString(left);
      this.rightAsString = objArrayToString(right);
   }

   public Production (String left, String right) {
      this.leftAsString = left;
      this.rightAsString = right;
      this.left = stringToObjArray(left);
      this.right = stringToObjArray(right);
   }

   public void setLeft (String left) {
      this.leftAsString = left;
      this.left = stringToObjArray(left);
   }

   public void setRight (String right) {
      this.rightAsString = right;
      this.right = stringToObjArray(right);
   }

   void removeOccurrencesOf (Object o) {
      vTmp.clear();
      for (int i = 0; i < left.length; i++) {
         if (!left[i].equals(o)) {
            vTmp.add(left[i]);
         }
      }
      left = vTmp.toArray();
      leftAsString = objArrayToString(left);
      vTmp.clear();
      for (int i = 0; i < right.length; i++) {
         if (!right[i].equals(o)) {
            vTmp.add(right[i]);
         }
      }
      right = vTmp.toArray();
      rightAsString = objArrayToString(right);
   }

   void renameOccurrencesOf (Object oldS, Object newS) {
      vTmp.clear();
      for (int i = 0; i < left.length; i++) {
         if (!left[i].equals(oldS)) {
            vTmp.add(left[i]);
         } else vTmp.add(newS);
      }
      left = vTmp.toArray();
      leftAsString = objArrayToString(left);
      vTmp.clear();
      for (int i = 0; i < right.length; i++) {
         if (!right[i].equals(oldS)) {
            vTmp.add(right[i]);
         } else vTmp.add(newS);
      }
      right = vTmp.toArray();
      rightAsString = objArrayToString(right);
   }

   public boolean isEmpty () {
      return (right.length + left.length) == 0;
   }

   private String objArrayToString (Object[] obja) {
      StringBuffer sb = new StringBuffer();
      for (int i = 0; i < obja.length; i++) {
         sb.append(obja[i].toString());
      }
      return sb.toString();
   }

   private Object[] stringToObjArray (String s) {
      Object[] obja = new Object[s.length()];
      for (int i = 0; i < s.length(); i++) {
         obja[i] = s.substring(i, i + 1);
      }
      return obja;
   }

   public Object clone () {
      Production prod = new Production(this.leftAsString, this.rightAsString);
      if (this.hasAttribute("STOP")) { //$NON-NLS-1$
         prod.addAttribute("STOP"); //$NON-NLS-1$
      }
      return prod;
   }

   public String toString () {
      String sTmp = ""; //$NON-NLS-1$
      for (int i = 0; i < left.length; i++) {
         sTmp += " " + left[i]; //$NON-NLS-1$
      }
      if (this.hasAttribute("STOP")) //$NON-NLS-1$
      sTmp += " -|"; //$NON-NLS-1$
      else sTmp += " ->"; //$NON-NLS-1$
      for (int i = 0; i < right.length; i++) {
         sTmp += " " + right[i]; //$NON-NLS-1$
      }
      return sTmp.trim();
   }
}