/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */

package ch.ethz.exorciser.markov.grammar.view;

import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.geom.RoundRectangle2D;

import ch.ethz.exorciser.markov.grammar.Grammar;
import ch.ethz.exorciser.markov.grammar.Production;

/**
 * this class displays a production of a grammar.
 * 
 * @author		Reto Lamprecht
 */
public class ProductionUI {

	public class Arrow {
		ProductionUI parent;

		Arrow(ProductionUI parent) {
			this.parent = parent;
		}
		public Production getProduction() {
			return parent.model;
		}
		public ProductionUI getProductionUI() {
			return parent;
		}
		public boolean isStop() {
			return parent.model.hasAttribute(ProductionEditor.ATTR_STOP);
		}

	}
	/** space between arrow and right side of production. */
	protected static int arrow_right_hgap = 4;

	/** space between index and left side of production. */
	private	static int index_left_hgap = 15;

	protected static Insets insets = new Insets(3, 4, 3, 4);
	/** space between left side of production and arrow. */
	protected static int left_arrow_hgap = 4;

	private static int roundness = 5;
	private Arrow arrow;
	protected int arrowW = 18;

	protected int arrowX; // never mind this variable. x-coord to draw '->' in production
	private Rectangle bounds = new Rectangle(0, 0, 1, 1);

	protected boolean dragged;

	private Grammar grammar;

	int index;
	int indexX;

	protected SymbolList left;
	Production model;
	protected SymbolList right;
	protected RoundRectangle2D rrect = new RoundRectangle2D.Double(0, 0, 1, 1, roundness, roundness);

	public ProductionUI(Grammar grammar, Production model) {
		this.grammar = grammar;
		this.model = model;
		left = new SymbolList(grammar, this, SymbolList.LEFT_SIDE_OF_PROD, model.getLeft());
		right = new SymbolList(grammar, this, SymbolList.RIGHT_SIDE_OF_PROD, model.getRight());
		arrow = new Arrow(this);
	}

	public Object clone() {
		ProductionUI clone = new ProductionUI(grammar, (Production) this.model.clone());
		clone.index = this.index;
		return clone;
	}
	public void drag(boolean b) {
		dragged = b;
	}
	public int getHeight() {
		return bounds.height;
	}
	public int getIndex() {
		return index;
	}

	public String getLeftFromView() {
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < left.symbols.size(); i++) {
			sb.append(((Symbol) left.symbols.get(i)).toString());
		}
		return sb.toString();
	}

	public int getLength() {
		return left.size() + right.size();

	}
	public Production getModel() {
		return this.model;
	}

	public String getRightFromView() {
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < right.symbols.size(); i++) {
			sb.append(((Symbol) right.symbols.get(i)).toString());
		}
		return sb.toString();
	}

	public Symbol getSymbol(int i) {
		if (1 > 0 && i < left.size()) {
			return left.getSymbol(i);
		}
		i -= left.size();
		if (i > 0 && i < right.size()) {
			return right.getSymbol(i);
		}
		return null;
	}
	public int getWidth() {
		return bounds.width;
	}
	public int getX() {
		return bounds.x;
	}
	public int getY() {
		return bounds.y;
	}

	public Object objectHitBy(int x, int y) {
		Object obj = left.objectHitBy(x, y);
		if (null != obj) {
			return obj;
		}
		obj = right.objectHitBy(x, y);
		if (null != obj) {
			return obj;
		}
		if (bounds.contains(x, y) && x > left.bounds.x + left.bounds.width && x < right.bounds.x + right.bounds.width) {
			return arrow;
		}
		if (bounds.contains(x, y)) {
			return this;
		}
		return null;
	}

	public void paint(Renderer r, Graphics2D g) {
		r.paintProduction(this, g);
	}

	void removeOccurrencesOf(Object o) {
		left.removeObject(o);
		right.removeObject(o);
	}

	void renameOccurrencesOf(Object oldO, Object newO) {
		left.renameObject(oldO, newO);
		right.renameObject(oldO, newO);
	}

	/**
	 * Called from ProductionEditor-method notify(GrammarChange) to display GrammarChanges of type CHANGE_PRODUCTION_LEFT.
	 * Creates new Objects for all Symbols. Dont use for rename or delete symbols.
	 */
	void setLeft(String left) {
		if (!left.equals(getLeftFromView())) {
			this.left.applySymbols(left);
		}
	}

	/**
	 * Called from ProductionEditor-method notify(GrammarChange) to display GrammarChanges of type CHANGE_PRODUCTION_RIGHT.
	 * Creates new Objects for all Symbols. Dont use for rename or delete symbols.
	 */
	void setRight(String right) {
		if (!right.equals(getRightFromView())) {
			this.right.applySymbols(right);
		}
	}

	void updateBounds(Renderer r, int x, int y) {
		bounds.x = x;
		bounds.y = y;
		int width = insets.left;
		int height = 0;
		// production number
		indexX = bounds.x + width;
		FontMetrics fm = ProductionEditor.fm;
		width += fm.stringWidth(index + " :") + index_left_hgap; //$NON-NLS-1$
		height = Math.max(height, fm.getHeight());
		// left
		left.updateBounds(r, x + width, y + insets.top);
		width += left.getWidth() + left_arrow_hgap;
		height = Math.max(height, left.getHeight());
		// arrow
		arrowX = bounds.x + width;
		width += arrowW + arrow_right_hgap;
		//width += fm.stringWidth("-->") + arrow_right_hgap;
		height = Math.max(height, fm.getHeight());
		// right
		right.updateBounds(r, x + width, y + insets.top);
		width += right.getWidth();
		height = Math.max(height, right.getHeight());
		// finish
		bounds.width = width + insets.right;
		bounds.height = height + insets.top + insets.bottom;
		rrect.setFrame(bounds);
	}

}