/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.markov.grammar.view;

import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.geom.RoundRectangle2D;
import java.util.ArrayList;
import java.util.List;

import ch.ethz.exorciser.markov.grammar.Grammar;
import ch.ethz.exorciser.markov.grammar.Production;

/**
 * this class displays a list of symbols. it is used as a variable- or
 * terminallist and als left or right sides of productions.
 * 
 * @author Reto Lamprecht
 */
public class SymbolList {

	// role indicators for Symbollists
	public static int LEFT_SIDE_OF_PROD = 432;

	public static int RIGHT_SIDE_OF_PROD = 433;

	public static int VARIABLE_LIST = 434;

	public static int TERMINAL_LIST = 435;

	private Grammar grammar;

	private int role;

	public boolean isLeft() {
		return role == LEFT_SIDE_OF_PROD;
	}

	public boolean isRight() {
		return role == RIGHT_SIDE_OF_PROD;
	}

	public boolean isVariableList() {
		return role == VARIABLE_LIST;
	}

	public boolean isTerminalList() {
		return role == TERMINAL_LIST;
	}

	private ProductionUI parent;

	public Production getParentProduction() {
		if (null != parent) {
			return this.parent.getModel();
		}
		return null;
	}

	public ProductionUI getParentProductionUI() {
		return this.parent;
	}

	public void setParentProductionUI(ProductionUI parent) {
		this.parent = parent;
	}

	public List symbols = new ArrayList();

	public int size() {
		return symbols.size();
	}

	static int roundness = 5;

	RoundRectangle2D rrect = new RoundRectangle2D.Double(0, 0, 1, 1, roundness,
			roundness);

	public Rectangle bounds = new Rectangle(0, 0, 1, 1);

	public int getWidth() {
		return bounds.width;
	}

	public int getHeight() {
		return bounds.height;
	}

	public int getX() {
		return bounds.x;
	}

	public int getY() {
		return bounds.y;
	}

	public int textLineY;

	public int getTextLineY() {
		return textLineY;
	}

	int openingBracketsX, closingBracketsX;

	static Insets insets = new Insets(3, 7, 3, 7);

	/** space between symbols. */
	int symbols_hgap = 3;

	/** space inside symbolList if its empty */
	int empty_hgap = 18;

	Symbol lastRemoved, lastAdded; // to avoid duplicate removes and adds

	public SymbolList(Grammar grammar, int role, List symbols) {
		this.grammar = grammar;
		this.role = role;
		for (int i = 0; i < symbols.size(); i++) {
			this.symbols.add(new Symbol(grammar, symbols.get(i), this));
		}
	}

	public SymbolList(Grammar grammar, ProductionUI parent, int role,
			Object[] symbols) {
		this.grammar = grammar;
		this.parent = parent;
		this.role = role;
		this.symbols = new ArrayList();
		if (symbols != null) {
			for (int i = 0; i < symbols.length; i++) {
				this.symbols.add(new Symbol(grammar, symbols[i], this));
			}
		}
	}

	void applySymbols(Object[] symbols_) {
		// this.symbols.clear();
		int i;
		for (i = 0; i < symbols_.length && i < this.symbols.size(); i++) {
			Symbol current = (Symbol) this.symbols.get(i);
			current.setModel(symbols_[i]);
		}
		for (; i < symbols_.length; i++) {
			this.symbols.add(new Symbol(this.grammar, symbols_[i], this));
		}
		for (; i < this.symbols.size(); i++) {
			this.symbols.remove(i);
		}
	}

	void applySymbols(String symbols_) {
		// this.symbols.clear();
		int i;
		for (i = 0; i < symbols_.length() && i < this.symbols.size(); i++) {
			Symbol current = (Symbol) this.symbols.get(i);
			current.setModel(symbols_.substring(i, i + 1));
		}
		for (; i < symbols_.length(); i++) {
			this.symbols.add(new Symbol(this.grammar, symbols_.substring(i,
					i + 1), this));
		}
		for (; i < this.symbols.size(); i++) {
			this.symbols.remove(i);
		}
	}

	public void addSymbol(Symbol s) {
		addSymbol(symbols.size(), s);
	}

	public void addSymbol(int i, Symbol s) {
		if (s == lastAdded)
			return;
		s.setParentSymbolList(this);
		this.symbols.add(i, s);
		lastAdded = s;
		lastRemoved = null;
	}

	public Symbol getSymbol(int i) {
		if (i < symbols.size()) {
			return (Symbol) symbols.get(i);
		}
		return null;
	}

	public void removeSymbol(Symbol s) {
		if (s == lastRemoved)
			return;
		if (this.symbols.remove(s)) {
			lastRemoved = s;
			lastAdded = null;
		}
	}

	public void removeObject(Object o) {
		for (int i = 0; i < symbols.size(); i++) {
			Symbol current = (Symbol) symbols.get(i);
			if (current.getModel().equals(o)) {
				symbols.remove(current);
				// return;
				i--;
			}
		}
	}

	/**
	 * Sets the model of all Symbols in this list with model == oldO to newO.
	 * Called from ProducitonUI-method renameOccurrencesOf() (for productions)
	 * and ProductionEditor-method renameSym() (for varList and termList)
	 */
	public void renameObject(Object oldO, Object newO) {
		for (int i = 0; i < symbols.size(); i++) {
			Symbol current = (Symbol) symbols.get(i);
			if (current.getModel().equals(oldO)) {
				current.setModel(newO);
			}
		}
	}

	StringBuffer buf = new StringBuffer();

	public String toString() {
		buf.delete(0, buf.length());
		for (int i = 0; i < symbols.size(); i++) {
			buf.append(((Symbol) symbols.get(i)).toString());
		}
		return buf.toString();
	}

	/** Returns the index of the specified symbol. */
	public int indexOf(Symbol sym) {
		for (int i = 0; i < symbols.size(); i++) {
			Symbol current = (Symbol) symbols.get(i);
			if (current.equals(sym))
				return i;
		}
		return -1;
	}

	/** Returns the index of the symbol at the the specified pixel x-coordinate. */
	public int indexAt(int x) {
		for (int i = 0; i < symbols.size(); i++) {
			Symbol current = (Symbol) symbols.get(i);
			if (x < current.getX() + current.getWidth())
				return i;
		}
		return symbols.size();
	}

	void updateBounds(Renderer r, int x, int y) {
		r.updateSymListBounds(this, x, y);
	}

	public void paint(Renderer r, Graphics2D g) {
		r.paintSymbolList(this, g);
		Symbol current;
		for (int i = 0; i < symbols.size(); i++) {
			current = (Symbol) symbols.get(i);
			current.paint(r, g);
		}
	}

	public Object objectHitBy(int x, int y) {
		Symbol current;
		for (int i = 0; i < symbols.size(); i++) {
			current = (Symbol) symbols.get(i);
			if (current.contains(x, y)) {
				return current;
			}
		}
		if (bounds.contains(x, y)) {
			return this;
		}
		return null;
	}
}