/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.markov.runtime;

import java.util.ArrayList;
import java.util.List;

import ch.ethz.exorciser.Messages;
import ch.ethz.exorciser.Notifications;
import ch.ethz.exorciser.markov.grammar.Grammar;
import ch.ethz.exorciser.markov.grammar.Production;
import ch.ethz.exorciser.markov.runtime.ui.Renderer;

/**
 * @author Thomas Briner
 * @author Daniel Mller
 */
public class Visualisation extends Thread {

   private static boolean FORWARD = true;

   //	private static boolean BACKWARD = false;
   private List productions;

   private List history; // remembered Mutations

   private String actString;

   private String givenInput;

   private Notifications notifications;

   private Constraints constraints;

   private int stepLimit;

   private int stringLimit;

   private int prodNr;

   private int speed;

   private boolean direction = FORWARD;

   private boolean finished = false;

   private boolean stop;

   private boolean step;

   private final int DELAY = 20;

   public Visualisation (Grammar grammar, String string, String input,
         Constraints constr, Notifications notifications, List history) {
      this.history = history;
      changeSetting(grammar, string, input, constr, notifications);
   }

   public Visualisation (Grammar grammar, String string, String input,
         Notifications notifications, List history) {
      this.history = history;
      changeSetting(grammar, string, input, null, notifications);
   }

   public List getHistory () {
      return history;
   }

   public void changeSetting (Grammar grammar_, String string, String input,
         Constraints constr, Notifications n) {
      setGrammar(grammar_);
      setFeedback(n);
      if (input != null) {
         setGivenInput(input);
         if (string == null) actString = input;
         else actString = string;
      }
      if (constr != null) setConstraints(constr);
   }

   public void changeGrammar (Grammar grammar_) {
      setGrammar(grammar_);
   }

   public void setConstraints (Constraints c) {
      if (c != null) this.constraints = c;
      else this.constraints = new Constraints();
      stringLimit = this.constraints
            .getRelativeStringLimit(givenInput.length());
      stepLimit = this.constraints.getRelativeStepLimit(givenInput.length());
   }

   private void setGrammar (Grammar grammar) {
      productions = grammar.getProductions();
   }

   private void setFeedback (Notifications n) {
      this.notifications = n;
   }

   public void setGivenInput (String input) {
      this.givenInput = input;
   }

   public String getActString () {
      return this.actString;
   }

   public void reset () {
      history.clear();
      actString = givenInput;
      notifyReset();
      finished = false;
      stop = true;
   }

   public void speedChanged (int speed_) {
      this.speed = speed_;
   }

   public void setDirection (boolean dir) {
      this.direction = dir;
   }

   public void run () {
      prodNr = 0;
      stop = false;
      while (!stop) {
         if (direction) {
            stepForward();
         } else {
            stepBackward();
         }
      }
      this.notifySetViewDown();
   }

   public void stepForward () {
      if (finished) { // ###
         notifications.addWarningNotification(Messages
               .getString("Markov.feedback.already_done.title"), //$NON-NLS-1$
               Messages.getString("Markov.feedback.already_done")); //$NON-NLS-1$
         notifications.getRootPane().repaint();
         stop = true;
      } else {
         if (productions.size() == 0) {
            stop = true;
            notifications.addWarningNotification(Messages
                  .getString("Markov.feedback.empty.algorithm.title"), //$NON-NLS-1$
                  Messages.getString("Markov.feedback.empty.algorithm")); //$NON-NLS-1$
         } else if (history.size() > stepLimit) {
            stop = true;
            notifications.addErrorNotification(Messages
                  .getString("Markov.feedback.step_limit_exceeded.title"), //$NON-NLS-1$
                  Messages.getString("Markov.feedback.step_limit_exceeded") //$NON-NLS-1$
                        .replaceAll("\\$STEPLIMIT", constraints //$NON-NLS-1$
                              .getRelativeStepLimitAsHTML()));
         } else if (actString.length() > stringLimit) {
            stop = true;
            notifications.addErrorNotification(Messages
                  .getString("Markov.feedback.string_limit_exceeded.title"), //$NON-NLS-1$
                  Messages.getString("Markov.feedback.string_limit_exceeded") //$NON-NLS-1$
                        .replaceAll("\\$STRINGLIMIT", constraints //$NON-NLS-1$
                              .getRelativeStringLimitAsHTML()));
         } else {
            int pos = applicable(prodNr);
            // checks if there is a pattern match
            if (pos > -1) {
               applicateProduction(prodNr, pos);
               if ((speed <= 85) || (history.size() % 5 == 0)) {
                  this.notifySetViewDown();
               }
               if (step) { // if stepwise: change highlighting
                  if (history.size() > 1) {
                     ((Production) productions.get(((Mutation) history
                           .get(history.size() - 2)).getProdNr()))
                           .removeAttribute("MARK"); //$NON-NLS-1$
                  }
                  stop = true;
                  ((Production) productions.get(prodNr)).setAttribute("MARK", //$NON-NLS-1$
                        Renderer.HIGHLIGHT);
               } else {
                  try {
                     sleep(DELAY * (100 - speed));
                  } catch (Exception e) {}
               }
               // is it a stop - production?
               if (((Production) productions.get(prodNr)).hasAttribute("STOP")) { //$NON-NLS-1$
                  if (history.size() > stepLimit) {
                     stop = true;
                     notifications
                           .addErrorNotification(
                                 Messages
                                       .getString("Markov.feedback.step_limit_exceeded.title"), //$NON-NLS-1$
                                 Messages.getString(
                                       "Markov.feedback.step_limit_exceeded") //$NON-NLS-1$
                                       .replaceAll("\\$STEPLIMIT", constraints //$NON-NLS-1$
                                             .getRelativeStepLimitAsHTML()));
                  } else if (actString.length() > stringLimit) {
                     stop = true;
                     notifications
                           .addErrorNotification(
                                 Messages
                                       .getString("Markov.feedback.string_limit_exceeded.title"), //$NON-NLS-1$
                                 Messages
                                       .getString(
                                             "Markov.feedback.string_limit_exceeded") //$NON-NLS-1$
                                       .replaceAll(
                                             "\\$STRINGLIMIT", //$NON-NLS-1$
                                             constraints
                                                   .getRelativeStringLimitAsHTML()));
                  } else {
                     notifications.addNotification(Messages
                           .getString("Markov.feedback.algorithm_done"), ""); //$NON-NLS-1$ //$NON-NLS-2$
                  }
                  finished = true;
                  stop = true;
               }
               prodNr = 0;
            } else {
               notifications.getRootPane().repaint(); // not so nice ...
               prodNr++;
            }
            if (prodNr >= productions.size()) {
               // no more production applicable
               finished = true;
               stop = true;
               notifications.addNotification(Messages
                     .getString("Markov.feedback.algorithm_done"), ""); //$NON-NLS-1$ //$NON-NLS-2$
            }
         }
      }
   }

   private int applicable (int productionNumber) { // get position of the first
      // pattern match
      int i = actString
            .indexOf(((Production) productions.get(productionNumber))
                  .leftAsString());
      return i;
   }

   private void applicateProduction (int productionNumber, int pos) {
      String leftPart, rightPart, oldMiddlePart, newMiddlePart;
      oldMiddlePart = ((Production) productions.get(productionNumber))
            .leftAsString();
      newMiddlePart = ((Production) productions.get(productionNumber))
            .rightAsString();
      leftPart = actString.substring(0, pos);
      rightPart = actString.substring(pos + oldMiddlePart.length(), actString
            .length());
      actString = leftPart.concat(newMiddlePart.concat(rightPart));
      Mutation mutation = new Mutation(productionNumber, pos);
      // remember the change
      history.add(mutation);
      notifyOut(pos, pos + oldMiddlePart.length());
      // set red lines in visualisation "from"
      notifyNewLine(actString); // set the new string
      notifyIn(pos, pos + newMiddlePart.length()); // set red lines "to"
      notifyProductionNumber(productionNumber);
      // set the number of applicated production
   }

   public void stepBackward () {
      String leftPart, rightPart, oldMiddlePart, newMiddlePart;
      if (history.size() > 0) {
         int pos = ((Mutation) history.get(history.size() - 1)).getPosition();
         int productionNumber = ((Mutation) history.get(history.size() - 1))
               .getProdNr();
         newMiddlePart = ((Production) productions.get(productionNumber))
               .leftAsString();
         oldMiddlePart = ((Production) productions.get(productionNumber))
               .rightAsString();
         leftPart = actString.substring(0, pos);
         rightPart = actString.substring(pos + oldMiddlePart.length(),
               actString.length());
         actString = leftPart.concat(newMiddlePart.concat(rightPart));
         notifyRemoveLastLine();
         notifyOut(-1, -1);
         history.remove(history.size() - 1);
         if ((speed <= 85) || (history.size() % 5 == 0))
               this.notifySetViewDown();
         if (step) { // change highlighting
            ((Production) productions.get(productionNumber))
                  .removeAttribute("MARK"); //$NON-NLS-1$
            if (history.size() > 0) {
               ((Production) productions.get(((Mutation) history.get(history
                     .size() - 1)).getProdNr())).setAttribute("MARK", //$NON-NLS-1$
                     Renderer.HIGHLIGHT);
            }
            stop = true;
         } else {
            try {
               sleep(DELAY * (100 - speed));
            } catch (Exception e) {}
         }
         finished = false;
      } else {
         stop = true;
         notifyReset();
      }
   }

   public void clearHighlight () { // remove all highlighting
      for (int i = 0; i < productions.size(); i++) {
         ((Production) productions.get(i)).removeAttribute("MARK"); //$NON-NLS-1$
      }
   }

   public boolean isFinished () {
      return finished;
   }

   public void setFinished (boolean finished) {
      this.finished = finished;
   }

   public void setStep (boolean b) {
      stop = false;
      step = b;
   }

   public void pause () {
      stop = true;
   }

   private List listeners = new ArrayList();

   public void addSequenceListener (Listener listener) {
      listeners.add(listener);
   }

   protected void notifyReset () {
      int i = 0;
      while (i < listeners.size()) {
         ((Listener) listeners.get(i)).handleReset();
         i++;
      }
   }

   protected void notifyNewLine (String line) {
      int i = 0;
      while (i < listeners.size()) {
         ((Listener) listeners.get(i)).handleNewLine(line);
         i++;
      }
   }

   protected void notifyRemoveLastLine () {
      int i = 0;
      while (i < listeners.size()) {
         ((Listener) listeners.get(i)).handleRemoveLastLine();
         i++;
      }
   }

   protected void notifyOut (int begin, int end) {
      int i = 0;
      while (i < listeners.size()) {
         ((Listener) listeners.get(i)).handleOut(begin, end);
         i++;
      }
   }

   protected void notifySetViewDown () {
      int i = 0;
      while (i < listeners.size()) {
         ((Listener) listeners.get(i)).handleSetViewDown();
         i++;
      }
   }

   protected void notifyIn (int begin, int end) {
      int i = 0;
      while (i < listeners.size()) {
         ((Listener) listeners.get(i)).handleIn(begin, end);
         i++;
      }
   }

   protected void notifyProductionNumber (int number) {
      int i = 0;
      while (i < listeners.size()) {
         ((Listener) listeners.get(i)).handleProductionNumber(number);
         i++;
      }
   }

   public interface Listener {

      public void handleOut (int begin, int end);

      public void handleIn (int begin, int end);

      public void handleProductionNumber (int number);

      public void handleReset ();

      public void handleNewLine (String line);

      public void handleRemoveLastLine ();

      public void handleSetViewDown ();

      public class Adapter implements Listener {

         public void handleIn (int begin, int end) {}

         public void handleNewLine (String line) {}

         public void handleOut (int begin, int end) {}

         public void handleProductionNumber (int number) {}

         public void handleReset () {}

         public void handleRemoveLastLine () {}

         public void handleSetViewDown () {}
      }
   }
}