/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.nondet.npda;

import java.util.Iterator;
import java.util.LinkedList;

import ch.ethz.exorciser.Debug;
import ch.ethz.exorciser.treebrowser.Edge;
import ch.ethz.exorciser.treebrowser.Node;

/**
 * 
 * The Thread that implements Movie mode. Starts at a given root and moves
 * through the tree breadth first at configurable speed
 * 
 * @author Oliver Biderbost
 */
public class MovieThread extends Thread {

    protected static final boolean DEBUG = true; // emit console warnings

    protected static final int MIN_SLEEP = 0;

    protected static int _count = 0;

    protected LinkedList _queue = null;

    protected boolean _running = false;

    protected int _sleeptime = MIN_SLEEP;

    /**
     * Constructor
     */
    public MovieThread() {
        super("MovieThread" + (++_count)); //$NON-NLS-1$
        _queue = new LinkedList();
    }

    /**
     * Method startMovie.
     * 
     * @param root
     * @param speed
     * 
     * starts the traversal at root with given speed 0 <=speed <=1
     */
    public void startMovie(Node root, float speed) {
        setSpeed(speed);
        synchronized (_queue) {
            _queue.clear();
            _queue.addLast(root);
            _running = true;
        }
        if (!isAlive())
            start(); // start thread if not already done
    }

    /**
     * Method stopMovie.
     * 
     * stop the movie and reset
     */
    public void stopMovie() {
        synchronized (_queue) {
            _queue.clear();
            _running = false;
        }
    }

    /**
     * Method pauseMovie.
     * 
     * pause the Movie resume with resumeMovie()
     */
    public void pauseMovie() {
        _running = false;
    }

    /**
     * Method resumeMovie.
     * 
     * resume the movie after pauseMovie
     */
    public void resumeMovie() {
        _running = true;
    }

    /**
     * Method setSpeed.
     * 
     * @param speed
     * 
     * sets sleep params according to speed 0->2000ms, 1->0ms
     */
    protected void setSpeed(float speed) {
        if (speed < 0.0)
            speed = 0.0f;
        if (speed > 1.0)
            speed = 1.0f;
        _sleeptime = (int) ((1.0 - speed) * 2000);
    }

    /**
     * Method run.
     * 
     * the run method of this Thread class does the actual work, loops
     * infinitely
     */
    public void run() {
        NPDAConfiguration cur;
        int sleeptime = _sleeptime;
        while (true) {
            while (_running) {
                try {
                    synchronized (_queue) {
                        cur = null;
                        if (!_queue.isEmpty())
                            cur = (NPDAConfiguration) _queue.removeFirst();
                        if (cur != null) {
                            if (!cur.isHidden())
                                sleeptime = 0; // not new, skip sleep
                            else
                                sleeptime = _sleeptime;
                            handleNode(cur);
                        }
                    }
                    if (_queue.isEmpty()) {
                        _running = false;
                    }
                } catch (OutOfMemoryError e) {
                    _queue.clear();
                    _running = false;
                }
                try { // sleep according to speed settings
                    Thread.sleep(sleeptime + MIN_SLEEP);
                } catch (InterruptedException e) {
                    Debug.showException(e);
                }
            }
            try { // not running -> long sleep
                Thread.sleep(500);
            } catch (InterruptedException e) {
                Debug.showException(e);
            }
        }
    }

    /**
     * Method handleNode.
     * 
     * @param node
     * 
     * unhides the node and inserts its children into the queue
     */
    protected void handleNode(NPDAConfiguration node) {
        Edge e;
        node.unhide();

        Iterator i = node.getChilds().iterator();
        while (i.hasNext()) {
            e = (Edge) i.next();
            _queue.addLast(e.target());
        }
    }
}