/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.nondet.npda;

import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import ch.ethz.exorciser.nondet.npda.automata.NPDA;
import ch.ethz.exorciser.treebrowser.Edge;
import ch.ethz.exorciser.treebrowser.Node;

/**
 * Representation of an NPDA configuration during execution Incorporates state,
 * stack and input string information *
 * 
 * @author Oliver Biderbost
 */
public class NPDAConfiguration extends Node {

   protected boolean _done = false; // are my sucessors calculated

   protected NPDA _npda = null; // reference to the npda

   protected String _stack = null; // my stack

   protected int _wordPos = 0; // my position on the input word

   protected int _state = -1; // my state in the npda

   /**
    * Constructor NPDAConfiguration.
    * 
    * @param npda
    * @param stack
    * @param pos
    * @param state
    *           Stack should be a new copy, npda should be a shared reference.
    *           Call init immediately after, before any other methods
    */
   public NPDAConfiguration (NPDA npda, String stack, int pos, int state) {
      _npda = npda;
      _stack = stack;
      _wordPos = pos;
      _state = state;
      _done = false;
   }

   /**
    * Method init.
    * 
    * @param parent
    *           Init as normal inner node below a given parent. Should be called
    *           before anything other
    */
   public void init (NPDATransition parent) {
      _parent = parent;
      _tree = ((NPDAConfiguration) _parent.source())._tree;
      _depth = ((NPDAConfiguration) _parent.source())._depth + 1;
      ((NPDATree) _tree).register(this);
      _hidden = true;
   }

   /**
    * Method init.
    * 
    * @param tree
    *           Init as root node of tree, should be called uniquely by the tree
    *           itself Should be called before anything other
    */
   public void init (NPDATree tree) {
      _parent = null;
      _tree = tree;
      _depth = 0;
      _hidden = false;
   }

   /**
    * @see nondeterministic.npda.Node#getChilds()
    */
   public List getChilds () {
      if (_childs == null) {
         transitions();
      }
      return _childs;
   }

   public Edge getParent () {
      return _parent;
   }

   public boolean isAccepting () {
      return _npda.accept(_state);
   }

   public boolean done () {
      return _done;
   }

   /**
    * Method stack.
    * 
    * @return String Returns the stack
    */
   public String stack () {
      return _stack;
   }

   /**
    * Method stack.
    * 
    * @param s
    *           set Stack to st
    */
   public void stack (String st) {
      _stack = st;
   }

   public int state () {
      return _state;
   }

   /**
    * Method wordPos.
    * 
    * @return int Returns the current position in the word
    */
   public int wordPos () {
      return _wordPos;
   }

   /**
    * Method wordPos.
    * 
    * @param p
    *           Set the current position in the word to p
    */
   public void wordPos (int p) {
      _wordPos = p;
   }

   public String word () {
      return _npda.word();
   }

   public String getStackString () {
      return _stack;
   }

   public List getNodePath () {
      LinkedList l = new LinkedList();
      Node cur = this;
      l.addFirst(cur);
      while (cur.getParent() != null) {
         cur = cur.getParent().source();
         l.addFirst(cur);
      }
      return l;
   }

   public List getEdgePath () {
      LinkedList l = new LinkedList();
      Node cur = this;
      while (cur.getParent() != null) {
         l.addFirst(cur.getParent());
         cur = cur.getParent().source();
      }
      return l;
   }

   public String toString () {
      return "[" + _state + ",'" + _npda.word().substring(_wordPos) + "'," //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
            + _stack + "]"; //$NON-NLS-1$
   }

   public void focus () {
      unhide();
      _visited = true;
      Iterator i = getChilds().iterator();
      while (i.hasNext())
         ((NPDAConfiguration) ((Edge) i.next()).target()).unhide();
      ((NPDATree) _tree)._engine.focus(this);
   }

   /**
    * Method unhide. set visibility according to state. either visible or
    * highlighted
    */
   public void unhide () {
      if (isHidden()) {
         if (isAccepting()) {
            setHighlight(true);
            Iterator iter = getEdgePath().iterator();
            while (iter.hasNext()) {
               ((Edge) iter.next()).setHighlight(true);
            }
         } else {
            setHide(false);
         }
      }
   }

   /**
    * @see nondeterministic.npda.Paintable
    */
   public void paint (Graphics g, Rectangle r) {
      Painter.paint(this, g, r);
   }

   public void paintDetails (Graphics g, Rectangle r) {
      focus();
      Painter.paintDetails(this, g, r);
   }

   public void setHide (boolean yes) {
      _hidden = yes;
      if (_hidden) _highlighted = false;
   }

   public void setHighlight (boolean yes) {
      _highlighted = yes;
      if (_highlighted) _hidden = false;
   }

   public void showAll () {
      unhide();
      if (_done) {
         Iterator i = getChilds().iterator();
         while (i.hasNext())
            ((Edge) i.next()).target().showAll();
      }
   }

   /**
    * Method transitions. calculates the possible transitions and creates
    * corresponding children
    */
   private void transitions () {
      List l = new LinkedList();
      Rule r;
      NPDAConfiguration child;
      NPDATransition trans;
      //		Stack s;
      //// int pos;
      Iterator i = _npda.rules(_state).iterator();
      while (i.hasNext()) {
         r = (Rule) i.next();
         if (r.matchConfiguration(this)) {
            child = new NPDAConfiguration(_npda, _stack, _wordPos, r.to());
            r.applyToConfiguration(child);
            trans = new NPDATransition(this, child, r);
            child.init(trans);
            l.add(trans);
         }
      }
      _childs = l;
      _done = true;
   }
}