/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.nondet.npda;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import ch.ethz.exorciser.nondet.npda.automata.NPDA;
import ch.ethz.exorciser.nondet.npda.automata.PalindromeNPDA;
import ch.ethz.exorciser.nondet.npda.automata.ParenthesisByGrammarNPDA;
import ch.ethz.exorciser.nondet.npda.automata.ParenthesisNPDA;
import ch.ethz.exorciser.nondet.npda.automata.RegexExampleNPDA;
import ch.ethz.exorciser.treebrowser.Edge;
import ch.ethz.exorciser.treebrowser.Node;
import ch.ethz.exorciser.treebrowser.Tree;

/**
 * Implementation of a generator. Generates a Tree of Nodes representing
 * Execution states of a given NPDA. Calculates a certain ammount of
 * Configurations in advance, to look for solutions and provide Tips to the
 * user. Also provides an interface to a movie component that traverses the Tree
 * at a given speed, making its nodes visible.
 * 
 * @author Oliver Biderbost
 */
public class NPDAEngine extends Generator { //implements Runnable {

   //	protected static final boolean DEBUG = false; // emit console warnings
   protected static final int INIT_MAX_DEPTH = 20; // experiment

   protected static final int INIT_MAX_NODES = 200000; // experiment: 230 000
                                                       // max

   protected NPDA _npda = null;

   protected LinkedList _queue = null;

   //	protected Thread _worker = null;
   //	protected boolean _running = false;
   protected int _curmaxnodes = INIT_MAX_NODES;

   protected int _curmaxdepth = INIT_MAX_DEPTH;

   protected int _curdepth = 0;

   protected NPDAConfiguration _oldfocus = null;

   protected MovieThread mt = null;

   protected NPDAEngine (String word, GeneratorParam genParam) {
      _npda = (NPDA) genParam;
      _queue = new LinkedList();
      _tree = new NPDATree(_npda.start(word), this);
      mt = new MovieThread();
   }

   public static Generator create (String word, GeneratorParam genParam) {
      return new NPDAEngine(word, genParam);
   }

   /**
    * Method getParams.
    * 
    * @return List returns the list of possible creation parameters
    */
   public static List getParams () {
      List l = new ArrayList();
      l.add(new PalindromeNPDA());
      l.add(new ParenthesisNPDA());
      l.add(new ParenthesisByGrammarNPDA());
      l.add(new RegexExampleNPDA());
      return l;
   }

   /**
    * Method reset.
    * 
    * @param word
    *           resets the engine to a new word
    */
   public void reset (String word) {
      stopMovie();
      synchronized (_queue) {
         _queue.clear();
      }
      _tree = new NPDATree(_npda.start(word), this);
   }

   public void reset (String word, GeneratorParam p) {
      stopMovie();
      _npda = (NPDA) p;
      synchronized (_queue) {
         _queue.clear();
      }
      _tree = new NPDATree(_npda.start(word), this);
   }

   public Tree getTree () {
      return _tree;
   }

   public Edge getTip (Node n) { // fixme: untested
      List nodeList = null;
      Iterator nodeIter;
      Iterator paths = ((NPDATree) _tree).accepting().iterator();
      if (!paths.hasNext()) return null; // no solutions found yet
      while (paths.hasNext()) {
         nodeList = ((NPDAConfiguration) paths.next()).getNodePath();
         if (nodeList.contains(n)) { // n is on path
            nodeIter = nodeList.iterator();
            while (nodeIter.hasNext())
               if (nodeIter.next() == n && nodeIter.hasNext())
               // found: // return transit to next
                     return ((Node) nodeIter.next()).getParent();
         }
      }
      return n.getParent(); // not on any path -> go up
   }

   /*
    * public String getSolutionInfo() { if (((NPDATree)
    * (_tree)).accepting().size() == 0) return "No solution found yet. No tipps
    * available. " + _tree.numNodes() + " nodes checked."; else return "" +
    * ((NPDATree) (_tree)).accepting().size() + " solution(s) found."; }
    */
   synchronized protected void restart () {}

   public void focus (NPDAConfiguration c) {
      if (_oldfocus != c) startAt(c);
      _oldfocus = c;
   }

   public void startAt (Node c) {
      synchronized (_queue) {
         _queue.clear();
         _queue.addLast(c);
         _curdepth = c.depth();
         _curmaxdepth = _curdepth + INIT_MAX_DEPTH;
         restart();
      }
   }

   protected int handleNode (NPDAConfiguration node) {
      Edge e;
      Iterator i = node.getChilds().iterator();
      while (i.hasNext()) {
         e = (Edge) i.next();
         _queue.addLast(e.target());
      }
      return node.depth();
   }

   public String defaultWord () {
      return _npda.defaultWord();
   }

   public String randomWord () {
      return _npda.randomWord();
   }

   public void startMovie (float speed) {
      mt.startMovie(_tree.root(), speed);
   }

   public void stopMovie () {
      mt.stopMovie();
   }

   public void resumeMovie () {
      mt.resumeMovie();
   }

   public void setMovieSpeed (float speed) {
      mt.setSpeed(speed);
   }
}