/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.nondet.npda.automata;

import java.awt.Image;
import java.util.ArrayList;
import java.util.List;

import ch.ethz.exorciser.nondet.npda.GeneratorParam;
import ch.ethz.exorciser.nondet.npda.NPDAConfiguration;
import ch.ethz.exorciser.nondet.npda.Rule;

/**
 * NPDA class used for parsing words This is the superclass of all NPDAs To
 * implement a new automaton subclass this and implement methods specific to
 * your NPDA
 * 
 * @author Oliver Biderbost
 */
public abstract class NPDA implements GeneratorParam {

   protected static final int MAX_GEN_WORD_LEN = 20;

   protected int _numStates = 0;

   protected int _startState = -1;

   protected boolean[] _accepting = null;

   protected List[] _rules = null;

   protected String _word;

   /**
    * Constructor
    * 
    * @param numStates
    */
   protected NPDA (int numStates) {
      _numStates = numStates;
      _accepting = new boolean[_numStates];
      _rules = new List[_numStates];
      for (int i = 0; i < _numStates; i++)
         _rules[i] = new ArrayList();
   }

   /**
    * Method name
    * 
    * @return String returns the name of this NPDA override in subclasses
    */
   public abstract String name ();

   /**
    * Method description
    * 
    * @return String returns a description for this NPDA override in subclasses
    */
   public abstract String description ();

   /**
    * Method start
    * 
    * @param word
    * @return NPDAConfiguration creates a starting configuration for the given
    *         input string invalidates all Configurations already created
    *         override in subclasses
    */
   public NPDAConfiguration start (String word) {
      _word = word + "$"; //$NON-NLS-1$
      return new NPDAConfiguration(this, "$", 0, _startState); //$NON-NLS-1$
   }

   /**
    * Method defaultWord. returns a default Word that can be parsed override in
    * subclasses
    */
   public abstract String defaultWord ();

   /**
    * Method randomWord
    * 
    * @param maxlen
    *           returns a random Word to be used for parsing override in
    *           subclasses
    */
   public abstract String randomWord ();

   /**
    * Method image
    * 
    * @return Image returns an image of the npda override in subclasses
    */
   public abstract Image image ();

   /**
    * Method addRule.
    * 
    * @param rule
    *           adds a rule to the list of rules if valid
    */
   protected void addRule (Rule rule) {
      if (validState(rule.from()) && validState(rule.to()))
            _rules[rule.from()].add(rule);
   }

   /**
    * Method addAccept.
    * 
    * @param i
    *           adds state i to the set of accepting states
    */
   protected void addAccept (int i) {
      if (validState(i)) _accepting[i] = true;
   }

   /**
    * Method startState.
    * 
    * @param i
    *           sets i to be the starting state
    */
   protected void startState (int i) {
      if (validState(i)) _startState = i;
   }

   /**
    * Method rules.
    * 
    * @param state
    * @return List get the rules for a given state
    */
   public List rules (int i) {
      if (validState(i)) return _rules[i];
      return null;
   }

   /**
    * Method rules.
    * 
    * @param state
    * @return boolean check if state is valid
    */
   protected boolean validState (int state) {
      return (state < _numStates && state >= 0);
   }

   /**
    * Method accept.
    * 
    * @param _state
    * @return boolean is state i accepting
    */
   public boolean accept (int i) {
      return _accepting[i];
   }

   /**
    * Method word
    * 
    * @return String returns the word being parsed
    */
   public String word () {
      return _word;
   }

   /**
    * Method coin
    * 
    * @return boolean returns random boolean values
    */
   protected boolean coin () {
      return (Math.random() < 0.5);
   }
}