/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.regex.Pattern;

import ch.ethz.exorciser.Debug;
import ch.ethz.exorciser.Messages;
import ch.ethz.exorciser.Notifications;
import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.ifa.FA;
import ch.ethz.exorciser.ifa.FASerializer;

/**
 * @author Vincent Tscherter
 */
public class FAFAChecker {

    private static boolean condition(FA solution, FA answer, String id)
            throws Exception {
        if (id.equals("NO_STATES")) { //$NON-NLS-1$
            return answer.getAllStates().size() == 0;
        }
        if (id.equals("EQUIVALENT")) { //$NON-NLS-1$
            return answer.getAllStates().size() != 0
                    && answer.findWitnessForNonEquivalence(solution) == null;
        }
        if (id.equals("HAS_ACCEPTING_STATE")) { //$NON-NLS-1$
            return answer.getAcceptingStates().size() > 0;
        }
        if (id.equals("HAS_EPSILON_TRANSITIONS")) { //$NON-NLS-1$
            return answer.getFAType().equals(FA.GFA);
        }
        if (id.equals("DETERMINISTIC")) { //$NON-NLS-1$
            return !answer.hasMultipleTransitions()
                    && !answer.hasEpsilonTransition();
        }
        throw new IllegalArgumentException("id '" + id + "' not recognised"); //$NON-NLS-1$ //$NON-NLS-2$
    }

    private static void feedback(FA solution, FA answer, String description,
            Notifications notifications) throws Exception {
        String[] t = Pattern.compile(":").split(description); //$NON-NLS-1$
        if (t.length >= 2) {
            String title = t[1];
            String text = ""; //$NON-NLS-1$
            if (t.length > 2) {
                text = t[2];
            }
            for (int i = 3; i < t.length; i++) {
                text += ":" + t[i]; //$NON-NLS-1$
            }
            if (text.indexOf("$WITNESS") >= 0) { //$NON-NLS-1$
                text = text
                        .replaceAll(
                                "\\$WITNESS", //$NON-NLS-1$
                                "" //$NON-NLS-1$
                                        + answer
                                                .findWitnessForNonEquivalence(solution));
            }
            if (text.indexOf("$EPSILON") >= 0) { //$NON-NLS-1$
                text = text.replaceAll("\\$EPSILON", Alphabet.EPSILON //$NON-NLS-1$
                        .toString());
            }
            if (text.indexOf("$ANSWER") >= 0) { //$NON-NLS-1$
                text = text.replaceAll("\\$ANSWER", //$NON-NLS-1$
                        FASerializer.createSerial(answer));
            }
            if (t[0].equals("WARNING")) { //$NON-NLS-1$
                notifications.addWarningNotification(title, text);
            } else if (t[0].equals("HINT")) { //$NON-NLS-1$
                notifications.addHintNotification(title, text);
            } else if (t[0].equals("ERROR")) { //$NON-NLS-1$
                notifications.addErrorNotification(title, text);
            } else if (t[0].equals("CORRECT")) { //$NON-NLS-1$
                notifications.addCorrectNotification(title, text);
            } else {
                throw new IllegalArgumentException(
                        "feedback type not recognised"); //$NON-NLS-1$
            }
        } else {
            throw new IllegalArgumentException(
                    "feedbackdesc not of type (ERROR|WARNING|CORRECT):title:text"); //$NON-NLS-1$
        }
    }

    public static void check(FA solution, FA answer, String feedbackscheme,
            Notifications notifications) {
        if (!answer.getAlphabet().equals(solution.getAlphabet())) {
            notifications.addErrorNotification(Messages
                    .getString("RE2FA.feedback.wrong_alphabeth.title"), //$NON-NLS-1$
                    Messages.getString("RE2FA.feedback.wrong_alphabeth") //$NON-NLS-1$
                            .replaceFirst("\\$EPSILON", //$NON-NLS-1$
                                    Alphabet.EPSILON.toString()));
        } else {
            try {

                BufferedReader reader = new BufferedReader(
                        new InputStreamReader(ClassLoader
                                .getSystemResourceAsStream(feedbackscheme)));
                String line = reader.readLine();
                boolean stop = false;
                while (line != null && !stop) {
                    if (line.length() > 0 && line.substring(0, 1).equals(":")) { //$NON-NLS-1$
                        boolean condition;
                        //                        System.out.println(line.trim());
                        if (line.substring(1, 2).equals("!")) { //$NON-NLS-1$
                            condition = !condition(solution, answer, line
                                    .substring(2).trim());
                        } else {
                            condition = condition(solution, answer, line
                                    .substring(1).trim());
                        }
                        line = reader.readLine();
                        while (line.length() > 0
                                && line.substring(0, 1).equals(":")) { //$NON-NLS-1$
                            //                            System.out.println(line.trim());
                            if (line.substring(1, 2).equals("!")) { //$NON-NLS-1$
                                condition = condition
                                        && !condition(solution, answer, line
                                                .substring(2).trim());
                            } else {
                                condition = condition
                                        && condition(solution, answer, line
                                                .substring(1).trim());
                            }
                            line = reader.readLine();
                        }
                        //                        System.out.println(">"+condition);
                        if (condition && line.length() > 0
                                && (line.substring(0, 1).equals("=") //$NON-NLS-1$
                                || line.substring(0, 1).equals("+"))) { //$NON-NLS-1$
                            feedback(solution, answer,
                                    line.substring(1).trim(), notifications);
                            if (line.substring(0, 1).equals("=")) { //$NON-NLS-1$
                                stop = true;
                            }
                        } else if (condition) {
                            throw new IllegalArgumentException(
                                    "feedback not of type (+|=)feedbackdesc"); //$NON-NLS-1$
                        }
                    }
                    line = reader.readLine();
                }

            } catch (Exception exception) {
                Debug.showException(exception);
            }
        }
    }

}