/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.minfa;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;

import ch.ethz.exorciser.Debug;
import ch.ethz.exorciser.fsmgui.ComponentRenderer;
import ch.ethz.exorciser.fsmgui.FSMEvent;
import ch.ethz.exorciser.fsmgui.FSMModelInterface;
import ch.ethz.exorciser.fsmgui.FSMModelListener;
import ch.ethz.exorciser.ifa.AcceptSign;
import ch.ethz.exorciser.ifa.BasicFA;
import ch.ethz.exorciser.ifa.FAException;
import ch.ethz.exorciser.ifa.State;

/**
 * Code adapted from original InputTrackPanel. InputTracker is able to deal with
 * several TableModels and marks them all according to input-string.
 * 
 * @see ch.ethz.exorciser.ifa.InputTrackPanel
 * @see ch.ethz.ifa.mdfa.table.model.TableModel
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class InputTracker extends JPanel implements CaretListener,
      FSMModelListener {

   private JTextField input_;

   private List models_;

   // the master of this component
   private ActivityController master_;

   private AcceptSign acceptSign_;

   /**
    * Create an InputTracker.
    * 
    * @param master
    *           Specify a master that controls activity of this InputTracker.
    */
   public InputTracker (ActivityController master) {
      master_ = master;
      models_ = new ArrayList();
      // set up input label
      JLabel label = new JLabel("input: "); //$NON-NLS-1$
      //		label.setFont(Font.getFont("sans"));
      //label.setBackground(new Color(170, 170, 200));
      // set up input textfield
      input_ = new JTextField(20);
      //		input_.setFont(new Font("SansSerif", Font.BOLD, 14));
      //		input_.setForeground(new Color(35, 35, 55));
      //		input_.setText("");
      input_.addCaretListener(this);
      // set up accept sign
      acceptSign_ = new AcceptSign();
      // set up panel
      this.add(label);
      this.add(input_);
      this.add(acceptSign_);
      this.setBackground(Color.white);
   }

   /**
    * Add a target to marked according to input-string. Register as listener
    * with the model.
    * 
    * @param model
    *           The mode for the fsm to be marked.
    */
   public void addEditTarget (FSMModelInterface model) {
      if (!models_.contains(model)) {
         model.addFSMModelListener(this);
         models_.add(model);
      }
   }

   /**
    * Remove a target. This target won't be marked anymore. Deregister as
    * listener with the model.
    * 
    * @param model
    *           The mode for the fsm to be marked.
    */
   public void removeEditTarget (FSMModelInterface model) {
      if (models_.contains(model)) {
         model.removeFSMModelListener(this);
         models_.remove(model);
      }
   }

   /**
    * Remove all edit targets. InputTracker won't mark any targets after that.
    */
   public void removeAllEditTargets () {
      for (int i = 0; i < models_.size(); i++) {
         ((FSMModelInterface) models_.get(i)).removeFSMModelListener(this);
      }
      models_.clear();
   }

   /**
    * InputTracker may handels Events originating from any fsm model this
    * InputTracker is registered to.
    */
   public void notify (FSMEvent e) {}

   /**
    * Handles caret-event. Mark all edit-targets.
    */
   public void caretUpdate (CaretEvent e) {
      Set activeStates;
      Set finalStates;
      State current;
      Iterator iterator;
      Hashtable marks;
      boolean accept = true;
      if (!master_.isActive()) return;
      for (int i = 0; i < models_.size(); i++) {
         ((FSMModelInterface) models_.get(i)).notifyFSMModelListeners(FSMEvent
               .createUnMarkAllEvent());
         // mark states reached after reading input till current caret position
         activeStates = statesAfter((FSMModelInterface) models_.get(i), input_
               .getText().substring(0, input_.getCaretPosition()));
         if (null != activeStates) {
            // mark the states
            iterator = activeStates.iterator();
            while (iterator.hasNext()) {
               current = (State) iterator.next();
               marks = new Hashtable();
               marks.put(current, ComponentRenderer.MARK_RED);
               ((FSMModelInterface) models_.get(i))
                     .notifyFSMModelListeners(FSMEvent.createMarkEvent(marks));
            }
            // mark following transition
            iterator = activeStates.iterator();
            if (input_.getText().length() > input_.getCaretPosition()) {
               Character ch = new Character(input_.getText().charAt(
                     input_.getCaretPosition()));
               while (null != ch && iterator.hasNext()) {
                  current = (State) iterator.next();
                  if (null != current.getNextState(ch)) {
                     marks = new Hashtable();
                     Object[] trans = { current, current.getNextState(ch)};
                     marks.put(trans, ComponentRenderer.MARK_RED);
                     ((FSMModelInterface) models_.get(i))
                           .notifyFSMModelListeners(FSMEvent
                                 .createMarkEvent(marks));
                  }
               }
            }
         }
         // signal acceptance of whole word in textfield.
         finalStates = statesAfter((FSMModelInterface) models_.get(i), input_
               .getText());
         try {
            if (null == finalStates
                  || !((BasicFA) models_.get(i))
                        .isOneOfTheseAccepting(finalStates)) {
               accept = false;
            }
         } catch (FAException e1) {
            Debug.showException(e1);
         }
      }
      acceptSign_.setAccept(accept);
      repaint();
   }

   /**
    * Helper method to determine active states in a fsm after processing an
    * input.
    */
   public Set statesAfter (FSMModelInterface model, String inputString) {
      if (null == model) return null;
      State currState = (State) model.getStartState();
      if (null == currState) { return null; }
      Set currSet = new HashSet();
      currSet.add(currState);
      Character currChar;
      Set newSet;
      for (int i = 0; i < inputString.length(); i++) {
         currChar = new Character(inputString.charAt(i));
         newSet = State.getNextStateSet(currSet, currChar);
         currSet = newSet;
      }
      return currSet;
   }
}