/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.minfa;

import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

/**
 * Save operations done on TableModel and provide undo/redo mechanism
 * 
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class Operations implements ChangePropagator {

   private List listeners_;

   private Stack undoStack_;

   private Stack redoStack_;

   private boolean undoPossible_;

   private boolean redoPossible_;

   /**
    * Construct Operations.
    * 
    * @param tableModel
    *           TableModel operations are saved for
    */
   protected Operations () {
      undoStack_ = new Stack();
      redoStack_ = new Stack();
      listeners_ = new ArrayList();
      undoPossible_ = false;
      redoPossible_ = false;
   }

   /**
    * ChangePropagator method to add a ChangeListener
    */
   public void addChangeListener (ChangeListener listener) {
      if (listener != null && !listeners_.contains(listener)) {
         listeners_.add(listener);
      }
   }

   /**
    * ChangePropagator method to remove a ChangeListener
    */
   public void removeChangeListener (ChangeListener listener) {
      if (listener != null && listeners_.contains(listener)) {
         listeners_.remove(listener);
      }
   }

   // helper method to adjust flags and ev. notify ChangeListeners
   private void adjustPossibilities (boolean undoPossible, boolean redoPossible) {
      if ((undoPossible_ != undoPossible) || (redoPossible_ != redoPossible)) {
         undoPossible_ = undoPossible;
         redoPossible_ = redoPossible;
         for (int i = 0; i < listeners_.size(); i++) {
            ((ChangeListener) listeners_.get(i)).change(this, new Pair(
                  new Boolean(undoPossible), new Boolean(redoPossible)));
         }
      }
   }

   /**
    * Add an operation (undo and redo part) to operations that can be undone and
    * redone.
    */
   protected void addOperations (Pair operationPair) {
      undoStack_.push(operationPair);
      redoStack_.clear();
      adjustPossibilities(true, false);
   }

   /**
    * Flag if undo possible.
    */
   public boolean undoPossible () {
      return (!undoStack_.isEmpty());
   }

   /**
    * Flag if redo possible.
    */
   public boolean redoPossible () {
      return (!redoStack_.isEmpty());
   }

   /**
    * Undo last operation.
    */
   public void undo () {
      if (undoPossible()) {
         Pair operationPair = (Pair) undoStack_.pop();
         CellOperation operation = (CellOperation) operationPair.getFirst();
         do_(operation);
         redoStack_.push(operationPair);
      }
      adjustPossibilities(undoPossible(), redoPossible());
   }

   /**
    * Redo last operation.
    */
   public void redo () {
      if (redoPossible()) {
         Pair operationPair = (Pair) redoStack_.pop();
         CellOperation operation = (CellOperation) operationPair.getSecond();
         do_(operation);
         undoStack_.push(operationPair);
      }
      adjustPossibilities(undoPossible(), redoPossible());
   }

   // helper method to perform an operation
   private void do_ (CellOperation operation) {
      switch (operation.getType()) {
      case CellOperation.CELL_SET:
         operation.getCell().set((Character) operation.getArgument(), false);
         break;
      case CellOperation.CELL_CLEAR:
         operation.getCell().clear(false);
         break;
      case CellOperation.CELL_EQUIV_SET:
         operation.getCell().setEquiv(false);
         break;
      case CellOperation.CELL_EQUIV_CLEAR:
         operation.getCell().clearEquiv(false);
         break;
      case CellOperation.CELL_CLEAR_ALL:
         operation.getCell().clear(false);
         operation.getCell().clearEquiv(false);
         break;
      default:
         break;
      }
   }

   /**
    * Clear undo/redo stacks.
    */
   protected void clear () {
      undoStack_.clear();
      redoStack_.clear();
      adjustPossibilities(false, false);
   }
}