/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.minfa;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.ifa.State;

/**
 * Class that models a single cell of a TableModel.
 * 
 * @see TableModel
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class TableCell {

   private int rowIndex_;

   private int colIndex_;

   private Character content_;

   private TableModel tableModel_;

   private TableCell refCell_;

   private boolean equiv_;

   // set of cells that reference this cell
   protected Set refBy;

   // flag to indicate cyclic reference chain
   private boolean refCycle_;

   /**
    * Construct a TableCell.
    * 
    * @param rowIndex
    *           Index of the row this cell is in.
    * @param colIndex
    *           Index of the column this cell is in.
    * @param tableModel
    *           The TableModel this cell belongs to.
    * @see TableModel
    */
   public TableCell (int rowIndex, int colIndex, TableModel tableModel) {
      rowIndex_ = rowIndex;
      colIndex_ = colIndex;
      tableModel_ = tableModel;
      refBy = new HashSet();
      refCell_ = null;
      equiv_ = false;
      refCycle_ = false;
   }

   /**
    * Set the content of the cell. Keep consistency of cell references.
    * 
    * @param content
    *           Content to be set.
    */
   public void set (Character content) {
      set(content, true);
   }

   protected void set (Character content, boolean undo) {
      if (content == null) {
         clear(undo);
      } else {
         if (undo) {
            // save undo information
            CellOperation undoOp = getUndoOp_();
            CellOperation redoOp = new CellOperation(CellOperation.CELL_SET,
                  this, content);
            tableModel_.getOperations().addOperations(new Pair(undoOp, redoOp));
         }
         content_ = content;
         // clear equivalence if set
         clearEquiv(false);
         // update reference
         updateRef_();
         // reset this cell and all cells that reference this cell
         reset(new HashSet());
         // create event
         tableModel_.createEvent(this, TableEvent.CELL_SET, this);
      }
   }

   /**
    * Clears the content of the cell.
    */
   public void clear () {
      clear(true);
   }

   protected void clear (boolean undo) {
      if (undo) {
         // save undo information
         CellOperation undoOp = getUndoOp_();
         CellOperation redoOp = new CellOperation(CellOperation.CELL_CLEAR,
               this, null);
         tableModel_.getOperations().addOperations(new Pair(undoOp, redoOp));
      }
      content_ = null;
      clearRef_();
      tableModel_.createEvent(this, TableEvent.CELL_CLEAR, this);
      reset(new HashSet());
   }

   /**
    * Set equivalence flag. Mark corresponding states equivalent. This method
    * creates an event that is propagated to all listeners that are registered
    * with the model.
    */
   public void setEquiv () {
      setEquiv(true);
   }

   protected void setEquiv (boolean undo) {
      if (!equiv_) {
         tableModel_.equivCount++;
      }
      if (undo) {
         // save undo information
         CellOperation undoOp = getUndoOp_();
         CellOperation redoOp = new CellOperation(CellOperation.CELL_EQUIV_SET,
               this, null);
         tableModel_.getOperations().addOperations(new Pair(undoOp, redoOp));
      }
      content_ = null;
      equiv_ = true;
      tableModel_.createEvent(this, TableEvent.CELL_EQUIV_SET, this);
   }

   /**
    * Clear equivalence flag. Mark corresponding states equivalent. This method
    * creates an event that is propagated to all listeners that are registered
    * with the model.
    */
   public void clearEquiv () {
      clearEquiv(true);
   }

   public void clearEquiv (boolean undo) {
      if (equiv_) {
         tableModel_.equivCount--;
      }
      if (undo) {
         // save undo information
         CellOperation undoOp = getUndoOp_();
         CellOperation redoOp = new CellOperation(
               CellOperation.CELL_EQUIV_CLEAR, this, null);
         tableModel_.getOperations().addOperations(new Pair(undoOp, redoOp));
      }
      equiv_ = false;
      tableModel_.createEvent(this, TableEvent.CELL_EQUIV_CLEAR, this);
   }

   // helper method to determine undo-operation
   private CellOperation getUndoOp_ () {
      CellOperation undoOp = null;
      if (isEquiv()) {
         undoOp = new CellOperation(CellOperation.CELL_EQUIV_SET, this, null);
      } else if (!isEmpty()) {
         undoOp = new CellOperation(CellOperation.CELL_SET, this, getContent());
      } else {
         undoOp = new CellOperation(CellOperation.CELL_CLEAR_ALL, this, null);
      }
      return undoOp;
   }

   /**
    * Reset content of this cell and of all cells, that are on a chain of
    * references leading to this cell. Avoid setting of cells twice (cycles) by
    * always passing the set of cells that are already visited.
    */
   public void reset (HashSet visitedCells) {
      // check if flagged cycle is still valid
      if (refCycle_) {
         TableCell curCell = refCell_;
         while ((curCell != null) && (curCell != this)) {
            curCell = curCell.getRef();
         }
         refCycle_ = (curCell == this);
      }
      tableModel_.createEvent(this, TableEvent.CELL_RESET, this);
      visitedCells.add(this);
      for (Iterator iter = refBy.iterator(); iter.hasNext();) {
         TableCell cell = (TableCell) iter.next();
         if (!visitedCells.contains(cell)) {
            cell.reset(visitedCells);
         }
      }
   }

   /**
    * Compute cell that is referenced by this cell given an input.
    */
   public TableCell computeRef (Character ch) {
      State rowState = getRowState();
      State colState = getColState();
      State nextRowState = rowState.getNextState(ch);
      State nextColState = colState.getNextState(ch);
      if ((nextRowState != null) && (nextColState != null)) {
         rowState = nextRowState;
         colState = nextColState;
      }
      return tableModel_.getStatePairCell(rowState, colState);
   }

   // helper method to set referenced cell according to cell content
   private void updateRef_ () {
      TableCell refCell = computeRef(content_);
      if (refCell != refCell_) {
         if (refCell == this) {
            clearRef_();
         } else {
            setRef_(refCell);
         }
      }
      // check for cycle in references
      TableCell curCell = refCell_;
      while ((curCell != null) && (!curCell.hasRefCycle()) && (!refCycle_)) {
         refCycle_ = (curCell == this);
         curCell = curCell.getRef();
      }
   }

   // helper method to set referenced cell and update refBy-Sets
   private void setRef_ (TableCell cell) {
      if (refCell_ != null) {
         refCell_.refBy.remove(this);
      }
      refCell_ = cell;
      if (refCell_ != null) {
         refCell_.refBy.add(this);
      }
   }

   // helper method to clear referenced cell and update refBy-Sets
   private void clearRef_ () {
      if (refCell_ != null) {
         refCell_.refBy.remove(this);
      }
      refCell_ = null;
      refCycle_ = false;
   }

   /**
    * Getter for the equivalence flag.
    */
   public boolean isEquiv () {
      return equiv_;
   }

   /**
    * Checks if the content of the cell is empty.
    */
   public boolean isEmpty () {
      return (content_ == null);
   }

   /**
    * Getter for the index of the row this cell is in.
    */
   public int getRowIndex () {
      return rowIndex_;
   }

   /**
    * Getter for the index of the column this cell is in.
    */
   public int getColIndex () {
      return colIndex_;
   }

   /**
    * Getter for the content of the cell.
    * 
    * @return Character that forms the content of the cell.
    */
   public Character getContent () {
      return content_;
   }

   /**
    * Getter for the referenced cell.
    * 
    * @return Cell that is reference by this cell.
    */
   public TableCell getRef () {
      return refCell_;
   }

   /**
    * Getter for the last cell in the reference chain. Stop if cyclic reference
    * detected.
    */
   public TableCell getRefRecursive () {
      if ((refCell_ == null) || (refCycle_)) {
         return this;
      }
	return refCell_.getRefRecursive();
   }

   /**
    * Getter for the word built from the content of this cell and all referenced
    * cells (if there is no ref-cycle) as Character array.
    */
   public Character[] getWordChars () {
      List chars = new ArrayList();
      String word = getWordString_(this);
      for (int index = 0; index < word.length(); index++) {
         chars.add(new Character(word.charAt(index)));
      }
      Character[] array = new Character[chars.size()];
      return (Character[]) chars.toArray(array);
   }

   /**
    * Getter for the word built from the content of this cell and all referenced
    * cells (if there is no ref-cycle) as String.
    */
   public String getWordString () {
      return getWordString_(this);
   }

   // helper method to get String of content of cells in a reference chain
   private String getWordString_ (TableCell cell) {
      StringBuffer word = new StringBuffer();
      TableCell cur = cell;
      while ((cur != null) && (!cur.hasRefCycle())) {
         if (cur.getContent() != null) {
            word.append(cur.getContent());
         }
         cur = cur.getRef();
         if ((cur != null) && (cur.containsEpsilon())) {
            cur = null;
         }
      }
      if (cur != null) { // we found a reference cycle
         TableCell cycleStart = cur;
         word.append(cur.getContent() + "("); //$NON-NLS-1$
         while (true) {
            cur = cur.getRef();
            if (cur == cycleStart) {
               word.append(cur.getContent() + ")*"); //$NON-NLS-1$
               break;
            }
            word.append(cur.getContent());
         }
      }
      return word.toString();
   }

   /**
    * Getter for the length of the word the characters in the cells in the
    * reference chain starting from this cell form.
    * 
    * @return length of word or -1 if there is a cycle in the reference chain.
    */
   public int getWordLength () {
      int length = 0;
      TableCell cur = this;
      while ((cur.getRef() != null) && (!cur.hasRefCycle())) {
         length++;
         cur = cur.getRef();
      }
      if (cur.hasRefCycle()) {
         return -1;
      }
	return length;
   }

   /**
    * Getter for flag that indicates cyclic chain of references.
    */
   public boolean hasRefCycle () {
      return refCycle_;
   }

   /**
    * Check if cell contains the Character EPSILON.
    */
   public boolean containsEpsilon () {
      return ((content_ != null) && (content_.equals(Alphabet.EPSILON)));
   }

   /**
    * Check if cell should contain epsilon
    */
   public boolean isEpsilon () {
      return ((getRowState().isAccepting() && !getColState().isAccepting()) || (!getRowState()
            .isAccepting() && getColState().isAccepting()));
   }

   /**
    * Getter for the set of cells that reference this cell.
    */
   public Set getRefBy () {
      return refBy;
   }

   /**
    * Getter for the state corresponding to the row this cell is in.
    */
   public State getRowState () {
      return tableModel_.getRowState(this);
   }

   /**
    * Getter for the state corresponding to the column this cell is in.
    */
   public State getColState () {
      return tableModel_.getColState(this);
   }

   /**
    * Conversion of the cell-content to a String.
    */
   public String toString () {
      if (content_ != null) {
         /*
          * if (refCell_ == null) { return content_.toString(); } else { return
          * (content_.toString() + refCell_.toString()); }
          */
         return content_.toString();
      }
	return ""; //$NON-NLS-1$
   }
}