/*
 * Exorciser - www.educeth.ch/exorciser  Copyright 2000-2004 Swiss Federal
 * Institue of Technology, Zrich Switzerland
 */
package ch.ethz.exorciser.rl.minfa;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ch.ethz.exorciser.Messages;
import ch.ethz.exorciser.Notifications;
import ch.ethz.exorciser.Shared;
import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.fsmgui.FSMStateInterface;

/**
 * Class listening to a TableModel and checking its correctness. Give feedback
 * and hints to the user. Checks are performed comparing TableModel to
 * previously computed solution.
 * 
 * @see ch.ethz.ifa.mdfa.table.listener.TableModelListener
 * @see ch.ethz.ifa.mdfa.table.model.TableModel
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class TableChecker implements TableModelListener {

   private TableModel tableModel_;

   private TableModel solution_;

   private Notifications notifications;

   /**
    * Create a TableChecker.
    * 
    * @param tableModel
    *           Reference to TableModel this checker has to check
    * @param annotated
    *           Reference to a Annotated Object to log feedback
    * @see ch.ethz.ifa.mdfa.table.model.TableModel
    * @see ch.ethz.ifa.task.LogPanel
    */
   public TableChecker (TableModel tableModel, Notifications n) {
      this.notifications = n;
      tableModel_ = tableModel;
      tableModel_.addTableModelListener(this);
   }

   /**
    * Method to clear the AnnotationList via Annotated maintained by a
    * FeedbackLogger.
    * 
    * @see ch.ethz.ifa.mdfa.table.feedback.FeedbackLogger
    */
   public void clearLog () {
      tableModel_.createEvent(this, TableEvent.TABLE_DEACTIVATE);
      notifications.removeAll();
   }

   /**
    * Method that handles TableEvents.
    * 
    * @see ch.ethz.ifa.mdfa.table.event.TableEvent
    */
   public void notify (TableEvent event) {
      switch (event.getType()) {
      case TableEvent.TABLE_SET:
         TableModel tmp = new TableModel(tableModel_.getFA());
         solveTable(tmp);
         solution_ = tmp;
         clearLog();
         break;
      case TableEvent.TABLE_CLEAR:
         clearLog();
         break;
      case TableEvent.CELL_SET:
         clearLog();
         TableCell cell = (TableCell) event.getArgument();
         if (!cell.containsEpsilon() && cell.hasRefCycle()) {
            Pair cellPair = new Pair(cell, solution_.getCell(
                  cell.getRowIndex(), cell.getColIndex() - 1));
            notifications.addErrorNotification(Messages
                  .getString("MinFA.feedback.title.cycle_detected"), //$NON-NLS-1$
                  Messages.getString("MinFA.feedback.cycle_detected") //$NON-NLS-1$
                        .replaceFirst("\\$WORD", cell.getWordString()) //$NON-NLS-1$
                        .replaceFirst("\\$ROW", //$NON-NLS-1$
                              cell.getRowState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$COL", //$NON-NLS-1$
                              cell.getColState().getLabel().toString())); //$NON-NLS-1$
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  cellPair));
         }
         break;
      case TableEvent.CELL_CLEAR:
      case TableEvent.CELL_EQUIV_CLEAR:
      case TableEvent.CELL_EQUIV_SET:
         clearLog();
         break;
      default:
         break;
      }
   }

   /**
    * Method to check if something is wrong with the table and give a hint if
    * so. Also give a hint of how to proceed, if everything is fine so far. Log
    * feedback via a FeedbackLogger.
    * 
    * @see ch.ethz.ifa.mdfa.table.feedback.FeedbackLogger
    */
   public void checkTable () {
      TableCell[] cells = tableModel_.getCells();
      TableCell cell = null;
      TableCell solution = null;
      Pair cellPair = null;
      List epsilonWrong = new ArrayList();
      List epsilonMissing = new ArrayList();
      List equivWrong = new ArrayList();
      List equivMissing = new ArrayList();
      List wordWrong = new ArrayList();
      List wordMissing = new ArrayList();
      List wordNotMinimal = new ArrayList();
      clearLog();
      if (!tableModel_.isEmpty()) {
         for (int index = 0; index < cells.length; index++) {
            cell = cells[index];
            solution = solution_.getCell(cell.getRowIndex(),
                  cell.getColIndex() - 1);
            cellPair = new Pair(cell, solution);
            // case A: cell contains epsilon
            if (cell.containsEpsilon()) {
               // case A1: solution doesn't contain epsilon
               if (!solution.containsEpsilon()) {
                  epsilonWrong.add(cellPair);
               }
            }
            // case B: equivalence flag set for cell
            else if (cell.isEquiv()) {
               // case B1: equivalence flag not set for solution
               if (!solution.isEquiv()) {
                  equivWrong.add(cellPair);
               }
            }
            // case C: cell contains some character (all but epsilon)
            else if (!cell.isEmpty()) {
               // case C1: solution is flagged equivalent
               if (solution.isEquiv()) {
                  wordWrong.add(cellPair);
               }
               // case C2: solution contains epsilon
               else if (solution.containsEpsilon()) {
                  // case C2a: cell distinguishes states, but word is not
                  // minimal in
                  // length
                  if (cell.getRef() != null && // cond added by vt [2004.01.03]
                        cell.getRefRecursive().isEpsilon()) {
                     wordNotMinimal.add(0, cellPair);
                  }
                  // case C2b: cell does not distinguish states
                  else {
                     wordWrong.add(cellPair);
                  }
               }
               // case C3: solution contains some character (all but
               // epsilon)
               else if (!solution.isEmpty()) {
                  // case C3a: cell distinguishes states
                  if (cell.getRefRecursive().isEpsilon()) {
                     // case C3aI: word in cell is not minimal in length
                     if (cell.getWordLength() > solution.getWordLength()) {
                        wordNotMinimal.add(cellPair);
                     }
                  }
                  // case C3b: cell does not distinguish states
                  else {
                     wordWrong.add(cellPair);
                  }
               }
            }
            // case D: cell is empty
            else {
               // case D1: solution is flagged equivalent
               if (solution.isEquiv()) {
                  equivMissing.add(cellPair);
               }
               // case D2: solution contains epsilon
               else if (solution.containsEpsilon()) {
                  epsilonMissing.add(cellPair);
               }
               // case D3: solution contains some character (all but
               // epsilon)
               else if (!solution.isEmpty()) {
                  // case D3a: one character is missing (referenced cells
                  // are already
                  // filled correctly)
                  if ((cell.computeRef(solution.getContent()).getContent() != null)
                        && (cell.computeRef(solution.getContent()).getContent()
                              .equals(solution.getRef().getContent()))) {
                     wordMissing.add(cellPair);
                  }
               }
            }
         }
         // wrong equivalence
         if (!equivWrong.isEmpty()) {
            TableCell c = (TableCell) ((Pair) equivWrong.get(0)).getFirst();
            TableCell s = (TableCell) ((Pair) equivWrong.get(0)).getSecond();
            notifications.addErrorNotification(Messages
                  .getString("MinFA.feedback.title.missplaced_equivalence"), //$NON-NLS-1$
                  Messages.getString("MinFA.feedback.missplaced_equivalence") //$NON-NLS-1$
                        .replaceFirst("\\$A", //$NON-NLS-1$
                              c.getRowState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$B", //$NON-NLS-1$
                              c.getColState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$WORD", s.getWordString().toString())); //$NON-NLS-1$
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  equivWrong.get(0)));
         }
         // wrong epsilon
         else if (!epsilonWrong.isEmpty()) {
            TableCell c = (TableCell) ((Pair) epsilonWrong.get(0)).getFirst();
            notifications
                  .addErrorNotification(Messages
                        .getString("MinFA.feedback.title.missplaced_epsilon"), //$NON-NLS-1$
                        Messages.getString("MinFA.feedback.missplaced_epsilon") //$NON-NLS-1$
                              .replaceFirst("\\$EPSILON", //$NON-NLS-1$
                                    Alphabet.EPSILON.toString()) //$NON-NLS-1$
                              .replaceFirst("\\$A", //$NON-NLS-1$
                                    c.getRowState().getLabel().toString()) //$NON-NLS-1$
                              .replaceFirst("\\$B", //$NON-NLS-1$
                                    c.getColState().getLabel().toString())); //$NON-NLS-1$
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  epsilonWrong.get(0)));
         }
         // wrong word
         else if (!wordWrong.isEmpty()) {
            TableCell c = (TableCell) ((Pair) wordWrong.get(0)).getFirst();
            TableCell s = (TableCell) ((Pair) wordWrong.get(0)).getSecond();
            String hint = (s.isEquiv() ? Messages
                  .getString("MinFA.feeback.hint.equivalent") //$NON-NLS-1$
                  : Messages.getString("MinFA.feedback.hint.witness") //$NON-NLS-1$
                        .replaceFirst("\\$WORD", s.getWordString())); //$NON-NLS-1$ //$NON-NLS-2$
            notifications.addErrorNotification(Messages
                  .getString("MinFA.feedback.title.wrong_word"), //$NON-NLS-1$
                  Messages.getString("MinFA.feedback.wrong_word") //$NON-NLS-1$
                        .replaceFirst("\\$A", //$NON-NLS-1$
                              c.getRowState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$B", //$NON-NLS-1$
                              c.getColState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$WORD", c.getWordString().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$HINT", hint) //$NON-NLS-1$
                  );
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  wordWrong.get(0)));
         }
         // not minimal word
         else if (!wordNotMinimal.isEmpty()) {
            TableCell c = (TableCell) ((Pair) wordNotMinimal.get(0)).getFirst();
            TableCell s = (TableCell) ((Pair) wordNotMinimal.get(0))
                  .getSecond();
            notifications.addWarningNotification(Messages
                  .getString("MinFA.feedback.title.not_minimal_word"), //$NON-NLS-1$
                  Messages.getString("MinFA.feedback.not_minimal_word") //$NON-NLS-1$
                        .replaceFirst("\\$WORD", s.getWordString()) //$NON-NLS-1$
                        .replaceFirst("\\$A", //$NON-NLS-1$
                              c.getRowState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$B", //$NON-NLS-1$
                              c.getColState().getLabel().toString())); //$NON-NLS-1$);
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  wordNotMinimal.get(0)));
         }
         // missing epsilon
         else if (!epsilonMissing.isEmpty()) {
            TableCell c = (TableCell) ((Pair) epsilonMissing.get(0)).getFirst();
            String a_accepting = (c.getRowState().isAccepting() ? Messages
                  .getString("MinFA.feedback.accepting") : Messages //$NON-NLS-1$
                  .getString("MinFA.feedback.non_accepting")); //$NON-NLS-1$ //$NON-NLS-2$
            String b_accepting = (c.getColState().isAccepting() ? Messages
                  .getString("MinFA.feedback.accepting") : Messages //$NON-NLS-1$
                  .getString("MinFA.feedback.non_accepting")); //$NON-NLS-1$ //$NON-NLS-2$
            notifications
                  .addWarningNotification(Messages
                        .getString("MinFA.feedback.title.missing_epsilon"), //$NON-NLS-1$
                        Messages.getString("MinFA.feedback.missing_epsilon") //$NON-NLS-1$
                              .replaceFirst("\\$EPSILON", //$NON-NLS-1$
                                    Alphabet.EPSILON.toString()) //$NON-NLS-1$
                              .replaceFirst("\\$A_ACCEPTING", a_accepting) //$NON-NLS-1$
                              .replaceFirst("\\$A", //$NON-NLS-1$
                                    c.getRowState().getLabel().toString()) //$NON-NLS-1$
                              .replaceFirst("\\$B_ACCEPTING", b_accepting) //$NON-NLS-1$
                              .replaceFirst("\\$B", //$NON-NLS-1$
                                    c.getColState().getLabel().toString()) //$NON-NLS-1$);
                  );
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  epsilonMissing.get(0)));
         }
         // missing word
         else if (!wordMissing.isEmpty()) {
            int i = 1;
            int min = 0;
            while (i < wordMissing.size()) {
               TableCell si = (TableCell) ((Pair) wordMissing.get(i))
                     .getSecond();
               TableCell smin = (TableCell) ((Pair) wordMissing.get(min))
                     .getSecond();
               if (si.getWordLength() < smin.getWordLength()) {
                  min = i;
               }
               i++;
            }
            TableCell c = (TableCell) ((Pair) wordMissing.get(min)).getFirst();
            TableCell s = (TableCell) ((Pair) wordMissing.get(min)).getSecond();
            notifications.addWarningNotification(Messages
                  .getString("MinFA.feedback.title.missing_word"), //$NON-NLS-1$
                  Messages.getString("MinFA.feedback.missing_word") //$NON-NLS-1$
                        .replaceFirst("\\$WORD", s.getWordString()) //$NON-NLS-1$
                        .replaceFirst("\\$A", //$NON-NLS-1$
                              c.getRowState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$B", //$NON-NLS-1$
                              c.getColState().getLabel().toString())); //$NON-NLS-1$
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  wordMissing.get(min)));
         }
         // missing equivalence
         else if (!equivMissing.isEmpty()) {
            TableCell c = (TableCell) ((Pair) equivMissing.get(0)).getFirst();
            notifications.addWarningNotification(Messages
                  .getString("MinFA.feedback.title.missing_equivalence"), //$NON-NLS-1$
                  Messages.getString("MinFA.feedback.missing_equivalence") //$NON-NLS-1$
                        .replaceFirst("\\$A", //$NON-NLS-1$
                              c.getRowState().getLabel().toString()) //$NON-NLS-1$
                        .replaceFirst("\\$B", //$NON-NLS-1$
                              c.getColState().getLabel().toString())); //$NON-NLS-1$
            tableModel_.createEvent(new TableEvent(this, TableEvent.CELL_NOTOK,
                  equivMissing.get(0)));
         }
         // everything is fine
         else {
            notifications.addCorrectNotification(Shared.CORRECT, Messages
                  .getString("MinFA.feedback.well_done")); //$NON-NLS-1$
         }
      }
      // table is empty
      else {
         notifications.addWarningNotification(Messages
               .getString("MinFA.feedback.empty_table"), ""); //$NON-NLS-1$ //$NON-NLS-2$
      }
   }

   /**
    * Method to check if something is wrong with the table and give a hint if
    * so. Also give a hint of how to proceed, if everything is fine so far. Log
    * feedback via a FeedbackLogger.
    * 
    * @see ch.ethz.ifa.mdfa.table.feedback.FeedbackLogger
    */
   public void refineTable () {
      TableCell[] cells = tableModel_.getCells();
      TableCell cell = null;
      TableCell solution = null;
      Pair cellPair = null;
      List epsilonWrong = new ArrayList();
      List epsilonMissing = new ArrayList();
      List equivWrong = new ArrayList();
      List equivMissing = new ArrayList();
      List wordWrong = new ArrayList();
      List wordMissing = new ArrayList();
      List wordNotMinimal = new ArrayList();
      clearLog();
      for (int index = 0; index < cells.length; index++) {
         cell = cells[index];
         solution = solution_.getCell(cell.getRowIndex(),
               cell.getColIndex() - 1);
         cellPair = new Pair(cell, solution);
         // case A: cell contains epsilon
         if (cell.containsEpsilon()) {
            // case A1: solution doesn't contain epsilon
            if (!solution.containsEpsilon()) {
               epsilonWrong.add(cellPair);
            }
         }
         // case B: equivalence flag set for cell
         else if (cell.isEquiv()) {
            // case B1: equivalence flag not set for solution
            if (!solution.isEquiv()) {
               equivWrong.add(cellPair);
            }
         }
         // case C: cell contains some character (all but epsilon)
         else if (!cell.isEmpty()) {
            // case C1: solution is flagged equivalent
            if (solution.isEquiv()) {
               wordWrong.add(cellPair);
            }
            // case C2: solution contains epsilon
            else if (solution.containsEpsilon()) {
               // case C2a: cell distinguishes states, but word is not
               // minimal in length
               if (cell.getRefRecursive().isEpsilon()) {
                  wordNotMinimal.add(0, cellPair);
               }
               // case C2b: cell does not distinguish states
               else {
                  wordWrong.add(cellPair);
               }
            }
            // case C3: solution contains some character (all but epsilon)
            else if (!solution.isEmpty()) {
               // case C3a: cell distinguishes states
               if (cell.getRefRecursive().isEpsilon()) {
                  // case C3aI: word in cell is not minimal in length
                  if (cell.getWordLength() > solution.getWordLength()) {
                     wordNotMinimal.add(cellPair);
                  }
               }
               // case C3b: cell does not distinguish states
               else {
                  wordWrong.add(cellPair);
               }
            }
         }
         // case D: cell is empty
         else {
            // case D1: solution is flagged equivalent
            if (solution.isEquiv()) {
               equivMissing.add(cellPair);
            }
            // case D2: solution contains epsilon
            else if (solution.containsEpsilon()) {
               epsilonMissing.add(cellPair);
            }
            // case D3: solution contains some character (all but epsilon)
            else if (!solution.isEmpty()) {
               // case D3a: one character is missing (referenced cells are
               // already filled
               // correctly)
               if ((cell.computeRef(solution.getContent()).getContent() != null)
                     && (cell.computeRef(solution.getContent()).getContent()
                           .equals(solution.getRef().getContent()))) {
                  wordMissing.add(cellPair);
               }
            }
         }
      }
      // wrong equivalence
      if (!equivWrong.isEmpty()) {
         TableCell cel = (TableCell) ((Pair) equivWrong.get(0)).getFirst();
         cel.clear();
         cel.clearEquiv();
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.wrong_entry_removed")); //$NON-NLS-1$ //$NON-NLS-2$
         tableModel_
               .createEvent(this, TableEvent.CELL_NOTOK, equivWrong.get(0));
         return;
      }
      // wrong epsilon
      else if (!epsilonWrong.isEmpty()) {
         TableCell cel = (TableCell) ((Pair) epsilonWrong.get(0)).getFirst();
         cel.clear();
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.wrong_entry_removed")); //$NON-NLS-1$ //$NON-NLS-2$
         tableModel_.createEvent(this, TableEvent.CELL_NOTOK, epsilonWrong
               .get(0));
         return;
      }
      // wrong word
      else if (!wordWrong.isEmpty()) {
         TableCell cel = (TableCell) ((Pair) wordWrong.get(0)).getFirst();
         cel.clear();
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.wrong_entry_removed")); //$NON-NLS-1$ //$NON-NLS-2$
         tableModel_.createEvent(this, TableEvent.CELL_NOTOK, wordWrong.get(0));
         return;
      }
      // not minimal word
      else if (!wordNotMinimal.isEmpty()) {
         TableCell cel = (TableCell) ((Pair) wordNotMinimal.get(0)).getFirst();
         cel.clear();
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.not_minimal_entry_removed")); //$NON-NLS-1$ //$NON-NLS-2$
         tableModel_.createEvent(this, TableEvent.CELL_NOTOK, wordNotMinimal
               .get(0));
         return;
      }
      // missing epsilon
      else if (!epsilonMissing.isEmpty()) {
         TableCell cel = (TableCell) ((Pair) epsilonMissing.get(0)).getFirst();
         TableCell sol = (TableCell) ((Pair) epsilonMissing.get(0)).getSecond();
         Character c = sol.getContent();
         cel.set(c);
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.symbol_inserted").replaceFirst( //$NON-NLS-1$
                     "\\$SYMBOL", c.toString())); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
         tableModel_.createEvent(this, TableEvent.CELL_NOTOK, epsilonMissing
               .get(0));
         return;
      }
      // missing word
      else if (!wordMissing.isEmpty()) {
         int i = 1;
         int min = 0;
         while (i < wordMissing.size()) {
            TableCell si = (TableCell) ((Pair) wordMissing.get(i)).getSecond();
            TableCell smin = (TableCell) ((Pair) wordMissing.get(min))
                  .getSecond();
            if (si.getWordLength() < smin.getWordLength()) {
               min = i;
            }
            i++;
         }
         TableCell cel = (TableCell) ((Pair) wordMissing.get(min)).getFirst();
         TableCell sol = (TableCell) ((Pair) wordMissing.get(min)).getSecond();
         Character c = sol.getContent();
         cel.set(c);
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.symbol_inserted").replaceFirst( //$NON-NLS-1$
                     "\\$SYMBOL", c.toString())); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
         tableModel_.createEvent(this, TableEvent.CELL_NOTOK, wordMissing
               .get(min));
         return;
      }
      // missing equivalence
      else if (!equivMissing.isEmpty()) {
         TableCell cel = (TableCell) ((Pair) equivMissing.get(0)).getFirst();
         cel.setEquiv();
         tableModel_.createEvent(this, TableEvent.CELL_NOTOK, equivMissing
               .get(0));
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.equivalence_inserted")); //$NON-NLS-1$ //$NON-NLS-2$
         return;
      }
      // everything is fine
      else {
         notifications.addNotification(Messages
               .getString("MinFA.feedback.title.next_step"), Messages //$NON-NLS-1$
               .getString("MinFA.feedback.nothing_to_do")); //$NON-NLS-1$ //$NON-NLS-2$
      }
   }

   /**
    * Call this method to solve the table-model this checker is registered to.
    */
   public void solveTable () {
      solveTable(tableModel_);
   }

   /**
    * Method to compute the solution and fill all the cells of a table-model at
    * once.
    */
   private void solveTable (TableModel model) {
      Iterator alphIter;
      FSMStateInterface[] nextState = new FSMStateInterface[2];
      TableCell cell, nextCell;
      int wordLength, index;
      Character input;
      // first of all clear the table
      model.clear();
      TableCell[] cells = model.getCells();
      // check statepairs that can be distinguished with word of length 0
      for (index = 0; index < cells.length; index++) {
         cell = cells[index];
         if ((model.getRowState(cell).isAccepting() && !model.getColState(cell)
               .isAccepting())
               || (!model.getRowState(cell).isAccepting() && model.getColState(
                     cell).isAccepting())) {
            cell.set(Alphabet.EPSILON);
         }
      }
      // look at words with length 1 and greater
      for (wordLength = 1; wordLength <= model.getNumStates(); wordLength++) {
         // iterate over all cells and check if states can be distinguished
         for (index = 0; index < cells.length; index++) {
            cell = cells[index];
            if (cell.isEmpty()) {
               Character foo = null;
               int min = Integer.MAX_VALUE;
               for (alphIter = model.getAlphabet().iterator(); alphIter
                     .hasNext();) {
                  input = (Character) alphIter.next();
                  nextState[0] = (model.getRowState(cell)
                        .getNextState(input));
                  nextState[1] = (model.getColState(cell)
                        .getNextState(input));
                  if (nextState[0] != nextState[1]) {
                     nextCell = model.getStatePairCell(nextState[0],
                           nextState[1]);
                     if (!nextCell.isEmpty()) {
                        if (foo == null || nextCell.isEpsilon()
                              || min > nextCell.getWordChars().length) {
                           foo = input;
                           min = nextCell.getWordChars().length;
                        }
                     }
                  }
               }
               if (foo != null) {
                  cell.set(foo);
               }
            }
         }
      }
      // empty cell mean that corresponding states are equivalent
      for (index = 0; index < cells.length; index++) {
         cell = cells[index];
         if (cell.isEmpty()) {
            cell.setEquiv();
         }
      }
   }
}