/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.minfa;

import java.util.EventObject;
import java.util.Hashtable;

/**
 * Class representing and identifying events that originate from a TableModel.
 * 
 * @see ch.ethz.ifa.mdfa.table.model.TableModel
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class TableEvent extends EventObject {

   private static Hashtable eventString_ = new Hashtable();

   /**
    * New table according to some FA set in TableModel.
    * 
    * @see ch.ethz.ifa.mdfa.table.model.TableModel
    */
   public static final int TABLE_SET = 100;
   static {
      eventString_.put(new Integer(TABLE_SET), "TABLE_SET"); //$NON-NLS-1$
   }

   /**
    * Content of table was cleared.
    */
   public static final int TABLE_CLEAR = 101;
   static {
      eventString_.put(new Integer(TABLE_CLEAR), "TABLE_CLEAR"); //$NON-NLS-1$
   }

   /**
    * Deactivate all cells of the table.
    */
   public static final int TABLE_DEACTIVATE = 102;
   static {
      eventString_.put(new Integer(TABLE_DEACTIVATE), "TABLE_DEACTIVATE"); //$NON-NLS-1$
   }

   /**
    * Content of cell set or referenced cell changed.
    */
   public static final int CELL_SET = 200;
   static {
      eventString_.put(new Integer(CELL_SET), "CELL_SET"); //$NON-NLS-1$
   }

   /**
    * Content of cell was cleared.
    */
   public static final int CELL_CLEAR = 201;
   static {
      eventString_.put(new Integer(CELL_CLEAR), "CELL_CLEAR"); //$NON-NLS-1$
   }

   /**
    * Content of cell was reset.
    */
   public static final int CELL_RESET = 202;
   static {
      eventString_.put(new Integer(CELL_RESET), "CELL_RESET"); //$NON-NLS-1$
   }

   /**
    * States corresponding to cell are considered equivalent. Set equivalence
    * flag.
    */
   public static final int CELL_EQUIV_SET = 220;
   static {
      eventString_.put(new Integer(CELL_EQUIV_SET), "CELL_EQUIV_SET"); //$NON-NLS-1$
   }

   /**
    * Clear equivalence flag.
    */
   public static final int CELL_EQUIV_CLEAR = 221;
   static {
      eventString_.put(new Integer(CELL_EQUIV_CLEAR), "CELL_EQUIV_CLEAR"); //$NON-NLS-1$
   }

   /**
    * Cell activated. e.g. mouse over cell.
    */
   public static final int CELL_ACTIVATE = 230;
   static {
      eventString_.put(new Integer(CELL_ACTIVATE), "CELL_ACTIVATE"); //$NON-NLS-1$
   }

   /**
    * Cell deactivated.
    */
   public static final int CELL_DEACTIVATE = 231;
   static {
      eventString_.put(new Integer(CELL_DEACTIVATE), "CELL_DEACTIVATE"); //$NON-NLS-1$
   }

   /**
    * Cell content/equivalence flag ok.
    */
   public static final int CELL_OK = 240;
   static {
      eventString_.put(new Integer(CELL_OK), "CELL_OK"); //$NON-NLS-1$
   }

   /**
    * Cell content/equivalence flag not ok.
    */
   public static final int CELL_NOTOK = 241;
   static {
      eventString_.put(new Integer(CELL_NOTOK), "CELL_NOTOK"); //$NON-NLS-1$
   }

   /**
    * A character is beeing dragged.
    */
   public static final int CHAR_DRAG = 300;
   static {
      eventString_.put(new Integer(CHAR_DRAG), "CHAR_DRAG"); //$NON-NLS-1$
   }

   /**
    * A character was dropped.
    */
   public static final int CHAR_DROP = 301;
   static {
      eventString_.put(new Integer(CHAR_DROP), "CHAR_DROP"); //$NON-NLS-1$
   }

   /**
    * A character is selecte. Either started to drag and drop it or clicked to
    * select.
    */
   public static final int CHAR_SELECT = 401;
   static {
      eventString_.put(new Integer(CHAR_SELECT), "CHAR_SELECT"); //$NON-NLS-1$
   }

   /**
    * Deselect any previously selected character.
    */
   public static final int CHAR_DESELECT_ALL = 402;
   static {
      eventString_.put(new Integer(CHAR_DESELECT_ALL), "CHAR_DESELECT_ALL"); //$NON-NLS-1$
   }

   private Object creator_;

   private int type_;

   private Object argument_;

   /**
    * Create a new TableEvent of some type without providing an event-argument.
    * 
    * @param source
    *           The TableModel that generated this event.
    * @param type
    *           The type of the event.
    */
   public TableEvent (Object source, int type) {
      this(source, null, type, null);
   }

   /**
    * Create a new TableEvent of some type.
    * 
    * @param source
    *           The TableModel that generated this event.
    * @param type
    *           The type of the event.
    * @param argument
    *           Argument to the event.
    */
   public TableEvent (Object source, int type, Object argument) {
      this(source, null, type, argument);
   }

   /**
    * Create a new TableEvent of some type without providing an event-argument.
    * 
    * @param source
    *           The TableModel that generated this event.
    * @param creator
    *           Creator of the event.
    * @param type
    *           The type of the event.
    */
   public TableEvent (Object source, Object creator, int type) {
      this(source, creator, type, null);
   }

   /**
    * Create a new TableEvent of some type providing an event-argument.
    * 
    * @param source
    *           The TableModel this event belongs to.
    * @param creator
    *           The creator of the event.
    * @param type
    *           The type of the event.
    * @param eventArgument
    *           Argument to the event. e.g. a tablecell.
    */
   public TableEvent (Object source, Object creator, int type, Object argument) {
      super(source);
      creator_ = creator;
      type_ = type;
      argument_ = argument;
   }

   /**
    * Getter for the creator of the event.
    */
   public Object getCreator () {
      return creator_;
   }

   /**
    * Getter for the type of the event.
    * 
    * @return Type of the event.
    */
   public int getType () {
      return type_;
   }

   /**
    * Getter for the argument to the event.
    * 
    * @return event-argument.
    */
   public Object getArgument () {
      return argument_;
   }

   /**
    * String with info about event, including event-type and -argument.
    */
   public String toString () {
      return new String(eventString_.get(new Integer(type_)) + ":\t" //$NON-NLS-1$
            + argument_); //$NON-NLS-1$
   }
}