/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.minfa;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import ch.ethz.exorciser.Debug;
import ch.ethz.exorciser.Messages;
import ch.ethz.exorciser.fsmgui.ComponentRenderer;
import ch.ethz.exorciser.fsmgui.FSMEvent;
import ch.ethz.exorciser.ifa.FA;
import ch.ethz.exorciser.ifa.State;

/**
 * Provides FAs corresponding to a TableModel (original FA and minimized FA) and
 * updates them according to TableModel changes.
 * 
 * @see ch.ethz.ifa.mdfa.table.model.TableModel
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class TableFAs implements TableModelListener, ChangePropagator {

   // ChangeListeners
   private List listeners_;

   // the fas
   private FA origFA_;

   private FA minFA_;

   private Character selectedChar_;

   private State[] origFAStates_;

   private TableModel tableModel_;

   // table containing mapping between orig-state and min-state
   private Hashtable stateTable_;

   /**
    * Construct a TableFAUpdater. Register with TableModel and create two FA
    * models. A model for the original FA and a model for the minimized FA.
    */
   public TableFAs (TableModel tableModel) {
      listeners_ = new ArrayList();
      setModel(tableModel);
      setFAs(tableModel_.getFA());
   }

   /**
    * Setter for TableModel. Deregisters with old model and registers with new
    * model.
    */
   public void setModel (TableModel tableModel) {
      tableModel_ = tableModel;
      // first deregister with old model
      if (tableModel_ != null) {
         tableModel_.removeTableModelListener(this);
      }
      // register with new model
      tableModel_.addTableModelListener(this);
   }

   /**
    * Set the fas. Constructs a minimized fa by cloning the original fa. Notify
    * all ChangeListeners and provide them with the actual fas.
    */
   public void setFAs (FA origFA) {
      if (origFA == null) { return; }
      origFA_ = origFA;
      minFA_ = (FA) origFA_.clone();
      origFAStates_ = tableModel_.getStates();
      // set names of tables
      origFA_.setName(Messages.getString("MinFA.label.original_fa")); //$NON-NLS-1$
      minFA_.setName(Messages.getString("MinFA.label.reduced_fa")); //$NON-NLS-1$
      // build stateTable, assume all labels are unique
      stateTable_ = new Hashtable();
      for (int index = 0; index < origFAStates_.length; index++) {
         stateTable_.put(origFAStates_[index], minFA_
               .getState((String) (origFAStates_[index].getLabel())));
      }
      notifyChangeListeners(new FA[] { origFA_, minFA_});
   }

   /**
    * ChangePropagator method to add ChangeListener.
    * 
    * @see ch.ethz.ifa.mdfa.util.ChangePropagator
    * @see ch.ethz.ifa.mdfa.util.ChangeListener
    */
   public void addChangeListener (ChangeListener listener) {
      if (!listeners_.contains(listener)) {
         listeners_.add(listener);
      }
   }

   /**
    * ChangePropagator method to remove ChangeListener.
    * 
    * @see ch.ethz.ifa.mdfa.util.ChangePropagator
    * @see ch.ethz.ifa.mdfa.util.ChangeListener
    */
   public void removeChangeListener (ChangeListener listener) {
      if (listeners_.contains(listener)) {
         listeners_.remove(listener);
      }
   }

   /**
    * Notify all ChangeListeners of a change.
    */
   public void notifyChangeListeners (Object arg) {
      for (int index = 0; index < listeners_.size(); index++) {
         ((ChangeListener) (listeners_.get(index))).change(this, arg);
      }
   }

   /**
    * Getter for the original FA model. The FA is not meant to be changed!
    */
   public FA getOrigFA () {
      return origFA_;
   }

   /**
    * Getter for the minimized FA model. The FA is not meant to be changed!
    */
   public FA getMinFA () {
      return minFA_;
   }

   /**
    * TableEvent Handler: Mark states and transitions in fas. Construct new
    * minimized fa and notify ChangeListeners. Notify ChangeListener of need to
    * display both fas.
    */
   public void notify (TableEvent event) {
      TableCell cell;
      switch (event.getType()) {
      case TableEvent.TABLE_SET:
         setFAs(tableModel_.getFA());
         notifyChangeListeners(new Boolean(false));
         break;
      case TableEvent.TABLE_CLEAR:
         // unmark FAs
         origFA_.notifyFSMModelListeners(FSMEvent.createUnMarkAllEvent());
         minFA_.notifyFSMModelListeners(FSMEvent.createUnMarkAllEvent());
         notifyChangeListeners(new Boolean(false));
         break;
      case TableEvent.TABLE_DEACTIVATE:
         origFA_.notifyFSMModelListeners(FSMEvent.createUnMarkAllEvent());
         minFA_.notifyFSMModelListeners(FSMEvent.createUnMarkAllEvent());
      case TableEvent.CELL_ACTIVATE:
         cell = (TableCell) (event.getArgument());
         createMarks_(cell, ComponentRenderer.MARK_BLUE);
         break;
      case TableEvent.CELL_NOTOK:
         Character tmp = selectedChar_;
         cell = (TableCell) ((Pair) event.getArgument()).getSecond();
         selectedChar_ = cell.getContent();
         createMarks_(cell, ComponentRenderer.MARK_RED);
         selectedChar_ = tmp;
         break;
      case TableEvent.CELL_EQUIV_SET:
         cell = (TableCell) (event.getArgument());
         setEquiv_(cell);
         notifyChangeListeners(new Boolean(true));
         break;
      case TableEvent.CELL_EQUIV_CLEAR:
         cell = (TableCell) (event.getArgument());
         clearEquiv_(cell);
         if (tableModel_.getEquivCount() == 0) {
            notifyChangeListeners(new Boolean(false));
         }
         break;
      case TableEvent.CHAR_SELECT:
         selectedChar_ = (Character) event.getArgument();
         break;
      case TableEvent.CHAR_DESELECT_ALL:
         selectedChar_ = null;
         break;
      default:
         break;
      }
   }

   // create the marks for a cell
   private void createMarks_ (TableCell cell, Object markColor) {
      if (cell != null) {
         Hashtable marksOrigFA = new Hashtable();
         Hashtable marksMinFA = new Hashtable();
         // create marks for states
         createStateMarks_(marksOrigFA, cell, markColor, null);
         createStateMarks_(marksMinFA, cell, markColor, stateTable_);
         // create marks for transitions
         createTransMarks_(marksOrigFA, cell, markColor, null);
         createTransMarks_(marksMinFA, cell, markColor, stateTable_);
         // mark FAs
         origFA_.notifyFSMModelListeners(FSMEvent.createMarkEvent(marksOrigFA));
         minFA_.notifyFSMModelListeners(FSMEvent.createMarkEvent(marksMinFA));
      }
   }

   // create the marks for the states
   private void createStateMarks_ (Hashtable marks, TableCell cell,
         Object markColor, Hashtable stateMapping) {
      if (stateMapping == null) {
         marks.put(cell.getRowState(), markColor);
         marks.put(cell.getColState(), markColor);
      } else {
         marks.put(stateMapping.get(cell.getRowState()), markColor);
         marks.put(stateMapping.get(cell.getColState()), markColor);
      }
   }

   // create the marks for the transitions
   private void createTransMarks_ (Hashtable marks, TableCell cell,
         Object markColor, Hashtable stateMapping) {
      TableCell curCell = cell;
      // create transition on draggedChar
      if (selectedChar_ != null) {
         addTransMarks_(marks, curCell, selectedChar_,
               ComponentRenderer.MARK_GREEN, stateMapping);
         curCell = curCell.computeRef(selectedChar_);
      }
      // create transition on cell content
      else if (curCell.getContent() != null) {
         addTransMarks_(marks, curCell, curCell.getContent(), markColor,
               stateMapping);
         curCell = curCell.getRef();
      }
      // create all possible transitions
      else {
         for (Iterator alphIter = origFA_.getAlphabet().iterator(); alphIter
               .hasNext();) {
            addTransMarks_(marks, curCell, (Character) alphIter.next(),
                  ComponentRenderer.MARK_GREEN, stateMapping);
         }
      }
      // create transitions of referenced cells
      while ((curCell != null) && (curCell != cell)
            && (curCell.getRef() != null) && (!curCell.hasRefCycle())) {
         addTransMarks_(marks, curCell, curCell.getContent(), markColor,
               stateMapping);
         curCell = curCell.getRef();
      }
   }

   // helper method for adding marks for transitions
   private void addTransMarks_ (Hashtable marks, TableCell cell, Character ch,
         Object markColor, Hashtable stateMapping) {
      if (ch == null) { return; }
      State rowState = cell.getRowState();
      State colState = cell.getColState();
      State nextRowState = rowState.getNextState(ch);
      State nextColState = colState.getNextState(ch);
      if ((nextRowState != null) && (nextColState != null)) {
         if (stateMapping == null) {
            marks.put(new Object[] { rowState, nextRowState}, markColor);
            marks.put(new Object[] { colState, nextColState}, markColor);
         } else {
            marks.put(new Object[] { stateMapping.get(rowState),
                  stateMapping.get(nextRowState)}, markColor);
            marks.put(new Object[] { stateMapping.get(colState),
                  stateMapping.get(nextColState)}, markColor);
         }
      }
   }

   // construct minimized fa according to equivalence
   private void setEquiv_ (TableCell cell) {
      State origElimState = cell.getColState();
      State origKeepState = cell.getRowState();
      State elimState = (State) stateTable_.get(origElimState);
      State keepState = (State) stateTable_.get(origKeepState);
      if (elimState == keepState) { return; }
      // make sure we don't eliminate a start state
      if (elimState.isStartState()) {
         origElimState = cell.getRowState();
         origKeepState = cell.getColState();
         elimState = (State) stateTable_.get(origElimState);
         keepState = (State) stateTable_.get(origKeepState);
      }
      // let all transitions to elimState point to keepState
      moveTransitionsTo_(elimState, keepState, minFA_);
      // keepState should contain all transitions originating from elimState
      moveTransitionsFrom_(elimState, keepState, minFA_);
      // delete state
      try {
         minFA_.deleteState(elimState);
      } catch (Exception ex) {
         Debug.showException(ex);
      }
      // set acceptance of state
      keepState
            .setAccepting(keepState.isAccepting() || elimState.isAccepting());
      // adjust label
      keepState.setLabel((String) keepState.getLabel()
            + (String) elimState.getLabel());
      // put state in stateTable for proper marking
      for (Iterator iter = stateTable_.keySet().iterator(); iter.hasNext();) {
         Object key = iter.next();
         if (stateTable_.get(key) == elimState) {
            stateTable_.put(key, keepState);
         }
      }
   }

   // clear an equivalence and reconstruct minimized fa
   private void clearEquiv_ (TableCell cell) {
      State origElimState = cell.getColState();
      State origKeepState = cell.getRowState();
      State elimState = (State) stateTable_.get(origElimState);
      State keepState = (State) stateTable_.get(origKeepState);
      if (elimState != keepState) { return; }
      setFAs(origFA_);
      TableCell[] cells = tableModel_.getCells();
      for (int index = 0; index < cells.length; index++) {
         if (cells[index].isEquiv()) {
            setEquiv_(cells[index]);
         }
      }
      notifyChangeListeners(new FA[] { origFA_, minFA_});
   }

   // helper method for adding equivalence
   private void moveTransitionsTo_ (State oldState, State newState, FA fa) {
      for (Iterator stateIter = fa.getAllStates().iterator(); stateIter
            .hasNext();) {
         State curState = (State) stateIter.next();
         for (Iterator alphIter = fa.getAlphabet().iterator(); alphIter
               .hasNext();) {
            Character curChar = (Character) alphIter.next();
            if (curState.getNextStateSet(curChar).contains(oldState)) {
               curState.removeTransition(curChar, oldState);
               // move selftransitions properly
               if (curState != oldState) {
                  curState.addTransition(curChar, newState);
               } else {
                  newState.addTransition(curChar, newState);
               }
            }
         }
      }
   }

   // helper method for adding equivalence
   private void moveTransitionsFrom_ (State oldState, State newState, FA fa) {
      for (Iterator alphIter = fa.getAlphabet().iterator(); alphIter.hasNext();) {
         Character curChar = (Character) alphIter.next();
         for (Iterator stateIter = oldState.getNextStateSet(curChar).iterator(); stateIter
               .hasNext();) {
            State curState = (State) stateIter.next();
            oldState.removeTransition(curChar, curState);
            // move selftransitions properly
            if (curState != oldState) {
               newState.addTransition(curChar, curState);
            } else {
               newState.addTransition(curChar, newState);
            }
         }
      }
   }
}