/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.minfa;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Hashtable;
import java.util.List;

import ch.ethz.exorciser.fsmgui.Alphabet;
import ch.ethz.exorciser.fsmgui.FSMStateInterface;
import ch.ethz.exorciser.ifa.FA;
import ch.ethz.exorciser.ifa.State;

/**
 * This class holds the data-model of a table used in the fsm-minimization
 * algorithm described in the lecture notes of prof. Nievergelt.
 * 
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class TableModel {

   private FA fa_;

   private List faStates_;

   private List listeners_;

   private List rows_;

   private List cols_;

   private List cells_;

   private List rowNames_;

   private List colNames_;

   private String name_;

   private Alphabet alphabet_;

   private int numStates_;

   private Hashtable stateIndex_;

   protected int equivCount;

   private Operations operations_;

   /**
    * Construct an empty TableModel.
    */
   public TableModel () {
      this(null);
   }

   /**
    * Construct a TableModel. The table is constructed according to the fsm that
    * has to be mimimized.
    * 
    * @param fsmModel
    *           Model of the fsm that has to be minimized.
    */
   public TableModel (FA fa) {
      equivCount = 0;
      listeners_ = new ArrayList();
      operations_ = new Operations();
      setFA(fa);
   }

   /**
    * Set the model of a fsm. This method builds the table that can be used to
    * minimize the fsm.
    * 
    * @param fsmModel
    *           Model of the fsm that has to be minimized.
    */
   public void setFA (FA fa) {
      if (fa == null) { return; }
      fa_ = fa;
      name_ = fa.getName();
      alphabet_ = fa.getAlphabet();
      // fetch the states
      faStates_ = new ArrayList(fa_.getAllStates());
      // sort the states for nicer output
      //Collections.sort(faStates_, new StatesComparator());
      Collections.sort(faStates_, new Comparator() {

         public int compare (Object o1, Object o2) {
            return ((String) (((FSMStateInterface) o1).getLabel()))
                  .compareTo((String) ((FSMStateInterface) o2).getLabel());
         }

         public boolean equals (Object obj) {
            return false;
         }
      });
      // make sure the first state in the collection is the start-state
      //		if (!((State) (faStates_.get(0))).isStartState()) {
      //			fa_.reLabelAllStates("s");
      //			faStates_ = new ArrayList(fa_.getAllStates());
      //			Collections.sort(faStates_, new StatesComparator());
      //		}
      // init state index hashtable
      stateIndex_ = new Hashtable();
      for (int i = 0; i < faStates_.size(); i++) {
         stateIndex_.put(faStates_.get(i), new Integer(i));
      }
      // fill rows and cols
      numStates_ = faStates_.size();
      cells_ = new ArrayList(getNumCells());
      rows_ = new ArrayList(getNumRows());
      cols_ = new ArrayList(getNumCols());
      rowNames_ = new ArrayList(getNumRows());
      colNames_ = new ArrayList(getNumCols());
      for (int i = 1; i < numStates_; i++) {
         rowNames_.add(new String((String) (((State) faStates_.get(i - 1))
               .getLabel())));
         colNames_.add(new String((String) (((State) faStates_.get(i))
               .getLabel())));
         List rowList = new ArrayList(numStates_ - i);
         List colList = new ArrayList(i);
         rows_.add(rowList);
         cols_.add(colList);
         for (int j = 0; j < i; j++) {
            TableCell tableCell = new TableCell(j, i, this);
            ((ArrayList) rows_.get(j)).add(tableCell);
            colList.add(tableCell);
         }
      }
      // fill cell-vector
      for (int i = 0; i < getNumRows(); i++) {
         cells_.addAll((ArrayList) rows_.get(i));
      }
      // create event and notify the listeners that new table was created
      createEvent(this, TableEvent.TABLE_SET);
      // init undo/redo stacks
      operations_.clear();
   }

   /**
    * Add a listener that wishes to be notified of model changes.
    * 
    * @param listener
    *           TableModelListener that wants to register.
    * @see ch.ethz.ifa.mdfa.table.listener.TableModelListener
    */
   public void addTableModelListener (TableModelListener listener) {
      if (listener != null && !listeners_.contains(listener)) {
         listeners_.add(listener);
      }
   }

   /**
    * Remove a listener. The listener won't be notified of model changes in the
    * future.
    * 
    * @param listener
    *           TableModelListener that wants to deregister.
    * @see ch.ethz.ifa.mdfa.table.listener.TableModelListener
    */
   public void removeTableModelListener (TableModelListener listener) {
      if (listener != null && listeners_.contains(listener)) {
         listeners_.remove(listener);
      }
   }

   /**
    * Create a TableEvent and notify all registered TableModelListeners of that
    * event.
    * 
    * @param creator
    *           Creator of the event.
    * @param eventType
    *           Type of the event.
    * @see ch.ethz.ifa.mdfa.table.listener.TableModelListener
    * @see ch.ethz.ifa.mdfa.table.event.TableEvent
    */
   public void createEvent (Object creator, int eventType) {
      createEvent(creator, eventType, null);
   }

   /**
    * Create a TableEvent and notify all registered TableModelListeners of that
    * event.
    * 
    * @param creator
    *           Creator of the event.
    * @param eventType
    *           Type of the event.
    * @param eventArgument
    *           Argument Object to the event.
    * @see ch.ethz.ifa.mdfa.table.listener.TableModelListener
    * @see ch.ethz.ifa.mdfa.table.event.TableEvent
    */
   public void createEvent (Object creator, int eventType, Object eventArgument) {
      notifyTableModelListeners_(new TableEvent(this, creator, eventType,
            eventArgument));
   }

   /**
    * Create a TableEvent and notify all registered TableModelListeners of that
    * event.
    * 
    * @param event
    *           TableEvent to be fired.
    */
   public void createEvent (TableEvent event) {
      notifyTableModelListeners_(event);
   }

   /**
    * Notifies all registered TableModelListeners of an event that occurred.
    * 
    * @param event
    *           event that occurred.
    * @see ch.ethz.ifa.mdfa.table.listener.TableModelListener
    * @see ch.ethz.ifa.mdfa.table.event.TableEvent
    */
   private void notifyTableModelListeners_ (TableEvent event) {
      for (int i = 0; i < listeners_.size(); i++) {
         ((TableModelListener) listeners_.get(i)).notify(event);
      }
   }

   /**
    * Clear the model. Means clear content and equivalence flag of all cells.
    */
   public void clear () {
      TableCell cell;
      for (int i = 0; i < cells_.size(); i++) {
         cell = (TableCell) cells_.get(i);
         cell.clear(false);
         cell.clearEquiv(false);
      }
      operations_.clear();
      createEvent(this, TableEvent.TABLE_CLEAR);
   }

   /**
    * Getter for the FA this table is constructed for. The FA is not meant to be
    * changed!
    */
   // changes in the fa could be detected by implementing
   // FSMModelListener interface and registering with the fa
   public FA getFA () {
      return fa_;
   }

   /**
    * Getter for the sorted states of the fa this table is constructed for. The
    * states are not meant to be changed!
    */
   public State[] getStates () {
      State[] array = new State[faStates_.size()];
      return (State[]) faStates_.toArray(array);
   }

   /**
    * Getter for the name of the table.
    */
   public String getName () {
      return name_;
   }

   /**
    * Getter for the alphabet used in the fsm that belongs to this table.
    */
   public Alphabet getAlphabet () {
      return alphabet_;
   }

   /**
    * Getter for the number of tablerows.
    */
   public int getNumRows () {
      return numStates_ - 1;
   }

   /**
    * Getter for the number of tablecolumns.
    */
   public int getNumCols () {
      return numStates_ - 1;
   }

   /**
    * Getter for the number of tablecells.
    */
   public int getNumCells () {
      return (numStates_ * (numStates_ - 1)) / 2;
   }

   /**
    * Getter for the number of states the fsm of this model contains.
    */
   public int getNumStates () {
      return numStates_;
   }

   /**
    * Getter for the names of the rows.
    * 
    * @return Array containing the names of the rows.
    */
   public String[] getRowNames () {
      String[] array = new String[rowNames_.size()];
      return (String[]) rowNames_.toArray(array);
   }

   /**
    * Getter for the names of the columns.
    * 
    * @return Array containing the names of the columns.
    */
   public String[] getColNames () {
      String[] array = new String[colNames_.size()];
      return (String[]) colNames_.toArray(array);
   }

   /**
    * Getter for a name of a row
    * 
    * @param index
    *           Index of the row.
    */
   public String getRowName (int index) {
      return (String) rowNames_.get(index);
   }

   /**
    * Getter for a name of a column.
    * 
    * @param index
    *           Index of the column.
    */
   public String getColName (int index) {
      return (String) colNames_.get(index);
   }

   /**
    * Getter for a row of TableCells.
    * 
    * @param rowIndex
    *           Index of the row.
    * @return Array containing cells of the selected row.
    * @see ch.ethz.ifa.mdfa.table.model.TableCell
    */
   public TableCell[] getRow (int rowIndex) {
      TableCell[] array = new TableCell[((List) rows_.get(rowIndex)).size()];
      return (TableCell[]) (((List) rows_.get(rowIndex)).toArray(array));
   }

   /**
    * Getter for a column of TableCells.
    * 
    * @param colIndex
    *           Index of the column.
    * @return Array containing cells of the selected column.
    * @see ch.ethz.ifa.mdfa.table.model.TableCell
    */
   public TableCell[] getCol (int colIndex) {
      TableCell[] array = new TableCell[((List) cols_.get(colIndex)).size()];
      return (TableCell[]) (((List) cols_.get(colIndex)).toArray(array));
   }

   /**
    * Getter for all TableCells.
    * 
    * @return Array containing all cells of the table.
    * @see ch.ethz.ifa.mdfa.table.model.TableCell
    */
   public TableCell[] getCells () {
      TableCell[] array = new TableCell[cells_.size()];
      return (TableCell[]) (cells_.toArray(array));
   }

   /**
    * Getter for a specific TableCells.
    * 
    * @param rowIndex
    *           Index of row.
    * @param colIndex
    *           Index of column.
    * @return Selected TableCell.
    * @see ch.ethz.ifa.mdfa.table.model.TableCell
    */
   public TableCell getCell (int rowIndex, int colIndex) {
      return (TableCell) (((List) cols_.get(colIndex)).get(rowIndex));
   }

   /**
    * Getter for Operations Object.
    */
   public Operations getOperations () {
      return operations_;
   }

   /**
    * Getter for the index of a state. Every state has an unique index. Start
    * state has index 0.
    */
   public int getStateIndex (FSMStateInterface state) {
      return ((Integer) (stateIndex_.get(state))).intValue();
   }

   /**
    * Get a TableCell that corresponds to a pair of states.
    * 
    * @param state1
    *           A state of the fsm.
    * @param state2
    *           Another state of the fsm.
    * @return The TableCell that corresponds to the pair of states. null if the
    *         given state-pair doesn't correspond to a TableCell.
    * @see ch.ethz.ifa.mdfa.table.model.TableCell
    */
   public TableCell getStatePairCell (FSMStateInterface state1,
         FSMStateInterface state2) {
      // int index1 = faStates_.indexOf(state1);
      // int index2 = faStates_.indexOf(state2);
      int index1 = ((Integer) (stateIndex_.get(state1))).intValue();
      int index2 = ((Integer) (stateIndex_.get(state2))).intValue();
      if (index1 < index2) {
         return getCell(index1, index2 - 1);
      } else if (index1 > index2) {
         return getCell(index2, index1 - 1);
      } else {
         return null;
      }
   }

   /**
    * Getter for the count of equivalence flags set in cells.
    */
   public int getEquivCount () {
      return equivCount;
   }

   /**
    * Check if all cells are empty.
    */
   public boolean isEmpty () {
      int i;
      for (i = 0; i < cells_.size(); i++) {
         if (!(((TableCell) cells_.get(i)).isEmpty())
               || (((TableCell) cells_.get(i)).isEquiv())) {
            break;
         }
      }
      return (i >= cells_.size());
   }

   /**
    * Getter for the state that corresponds to the row a cell is in.
    */
   public State getRowState (TableCell cell) {
      return (State) (faStates_.get(cell.getRowIndex()));
   }

   /**
    * Getter for the state that corresponds to the column a cell is in.
    */
   public State getColState (TableCell cell) {
      return (State) (faStates_.get(cell.getColIndex()));
   }

   /**
    * Conversion of the TableModel-data to a String.
    */
   public String toString () {
      StringBuffer buffer = new StringBuffer();
      for (int colIndex = 0; colIndex < numStates_ - 1; colIndex++) {
         buffer.append("\t" + colNames_.get(colIndex)); //$NON-NLS-1$
      }
      buffer.append("\n"); //$NON-NLS-1$
      for (int rowIndex = 0; rowIndex < numStates_ - 1; rowIndex++) {
         for (int colIndex = 0; colIndex < numStates_; colIndex++) {
            if (colIndex == 0) {
               buffer.append(rowNames_.get(rowIndex));
            } else {
               if (colIndex <= rowIndex) {
                  buffer.append("\t"); //$NON-NLS-1$
               } else {
                  buffer.append("\t" //$NON-NLS-1$
                        + ((TableCell) (((List) rows_.get(rowIndex))
                              .get(colIndex - rowIndex - 1))).toString()); //$NON-NLS-1$
               }
            }
         }
         buffer.append("\n"); //$NON-NLS-1$
      }
      return buffer.toString();
   }
}