/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.minfa;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Frame;

import javax.swing.JPanel;
import javax.swing.JSplitPane;

import ch.ethz.exorciser.fsmgui.FSMView;
import ch.ethz.exorciser.ifa.FA;

/**
 * Class showing up to two fas. TwinFAView controls the activity of an
 * InputTracker and is informed of changes in fas-models from a
 * ChangePropagator.
 * 
 * @see ch.ethz.ifa.mdfa.itrack.InputTracker
 * @see ch.ethz.ifa.mdfa.util.ChangePropagator
 * @author Alexander Imfeld
 * @author Vincent Tscherter
 * @author Reto Lamprecht
 */
public class TwinFAView
	extends JPanel
	implements ActivityController, ChangeListener {

	private TwinFAMVCFactory origFAMVC_;
	private TwinFAMVCFactory minFAMVC_;
	private TwinFAController origFAController_;
	private TwinFAController minFAController_;
	private JSplitPane splitPane_;
	private InputTracker inputTracker_;
	private boolean inputTrackerActive_;
	private boolean showBoth_;

	/**
	 * Create a TwinFAView.
	 * @param propagator ChangePropagator that informs this TwinFAView of changes in fa-models.
	 */
	public TwinFAView(ChangePropagator propagator) {
		this(null, null, propagator);
	}

	/**
	 * Create a TwinFAView.
	 * @param origFA fa model for the first (original) fa.
	 * @param minFA fa model for the second (minimized) fa.
	 * @param propagator ChangePropagator that informs this TwinFAView of changes in fa-models.
	 */
	public TwinFAView(FA origFA, FA minFA, ChangePropagator propagator) {
		// create splitPane
		splitPane_ = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, true);
		// create inputTracker
		inputTracker_ = new InputTracker(this);
		// deactivate inputTracker
		inputTrackerActive_ = false;
		// create factories
		origFAMVC_ = new TwinFAMVCFactory();
		minFAMVC_ = new TwinFAMVCFactory();
		// get controllers
		origFAController_ = (TwinFAController) origFAMVC_.getController();
		minFAController_ = (TwinFAController) minFAMVC_.getController();
		// set the fa models
		setFAs(origFA, minFA);
		// register with propagator
		propagator.addChangeListener(this);
		// create the popup-menus
		Frame f = new Frame();
		origFAController_.createDefaultPopup(f, this);
		minFAController_.createDefaultPopup(f, this);
		// set up panel
		setLayout(new BorderLayout());
		// show FAs
		showBoth_ = true;
		showFAs(false);
		// set input tracker
		setActivity(false);
	}

	/**
	 * Setter for the two fa models this view displays.
	 */
	public void setFAs(FA origFA, FA minFA) {
		if ((origFA != null) && (minFA != null)) {
			inputTracker_.removeAllEditTargets();
			origFAMVC_.setModel(origFA);
			minFAMVC_.setModel(minFA);
			inputTracker_.addEditTarget(origFAMVC_.getModel());
			inputTracker_.addEditTarget(minFAMVC_.getModel());
		}
	}

	/**
	 * Setter for the two fa models this view displays.
	 */
	public void setFAs(FA[] fas) {
		if (fas.length == 2) {
			setFAs(fas[0], fas[1]);
		}
	}

	/**
	 * Update the panel.
	 * @param both Specify if both or just one of the fas has to be displayed.
	 */
	public void showFAs(boolean both) {
		if (showBoth_ && !both) {
			remove(splitPane_);
			add(origFAMVC_.getView(), BorderLayout.CENTER);
			showBoth_ = both;
			fit_();
		} else if (!showBoth_ && both) {
			remove(origFAMVC_.getView());
			add(splitPane_, BorderLayout.CENTER);
			showBoth_ = both;
			validate();
			// set up split pane
			splitPane_.setDividerLocation(0.5);
			splitPane_.setDividerSize(6);
			splitPane_.setForeground(Color.white);
			splitPane_.setBackground(Color.white);
			splitPane_.setLeftComponent(origFAMVC_.getView());
			splitPane_.setRightComponent(minFAMVC_.getView());
			fit_();
		}
	}

	/**
	 * Let the fas fit into their environment.
	 */
	private void fit_() {
		validate();
		// fit fas
		 ((FSMView) origFAMVC_.getView()).fit();
		((FSMView) minFAMVC_.getView()).fit();
	}

	/**
	 * Handle changes provided by ChangePropagator.
	 * Update view or reset fas.
	 * @see ch.ethz.ifa.mdfa.util.ChangePropagator
	 */
	public void change(ChangePropagator propagator, Object arg) {
		if (arg instanceof FA[]) {
			setFAs((FA[]) arg);
		} else if (arg instanceof Boolean) {
			showFAs(((Boolean) arg).booleanValue());
		}
	}

	/**
	 * Method to control activity of InputTracker.
	 * @see ch.ethz.ifa.mdfa.util.ActivityController
	 */
	public boolean isActive() {
		return inputTrackerActive_;
	}

	/**
	 * Set activity of InputTracker.
	 */
	public void setActivity(boolean active) {
		if (!active && inputTrackerActive_) {
			this.remove(inputTracker_);
			inputTrackerActive_ = active;
			origFAController_.getPopup().setTrackInput(inputTrackerActive_);
			minFAController_.getPopup().setTrackInput(inputTrackerActive_);
			fit_();
		} else if (active && !inputTrackerActive_) {
			this.add(inputTracker_, BorderLayout.SOUTH);
			inputTrackerActive_ = active;
			origFAController_.getPopup().setTrackInput(inputTrackerActive_);
			minFAController_.getPopup().setTrackInput(inputTrackerActive_);
			fit_();
		}
	}
}