/*
 * Exorciser - Automated generation and grading of exercises in an introductory
 * course on the Theory of Computation,
 * 
 * Copyright (C) 2000-2004 Vincent Tscherter et al., Swiss Federal Institue of
 * Technology, Zrich Switzerland
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 */
package ch.ethz.exorciser.rl.re;

import java.util.EmptyStackException;
import java.util.Stack;

import ch.ethz.exorciser.fsmgui.Alphabet;

/**
 * This class contains methods to convert REs with alphabet {'a', 'b', EPSILON,
 * '.', '|', '*', ''} to RegularExpression. Semesterarbeit 2002: Nicolas
 * Lefebvre, Matthias Dreier.
 * 
 * 2004-08-16: extentions added for prefix mode: intersection '&' and set minus
 * '-'
 * 
 * @author Nicolas Lefebvre
 * @author Matthias Dreier
 * @author Vincent Tscherter
 */
public abstract class ABRegExpParser {

    static boolean verbose = true;

    static final int PREFIX = 0;

    static final int INFIX = 1;

    static final int SUFFIX = 2;

    static final int EBNF = 3;

    static final int UNKNOWN = 4;

    /**
     * Method createREfromString. Method to create a RegularExpression object
     * from a given RE string in prefix-, infix-, EBNF- or suffix-format with
     * alphabet {'a', 'b', EPSILON, '.', '|', '*', ''}.
     * 
     * @param re
     * @return RegularExpression
     */
    public static RegularExpression createRE(String re)
            throws REParserException {
        try {
            return createPrefixFormRE(re);
        } catch (REParserException e) {
        	// empty block
        }
        try {
            return createBackusNaurFormRE(re);
        } catch (REParserException e) { // empty block
        }
        try {
            return createSuffixFormRE(re);
        } catch (REParserException e) {// empty block
        }
        return createInfixFormRE(re);
    }

    private static boolean isEpsilon(char ch) {
        return ch == Alphabet.EPSILON.charValue() || ch == '' // Alt-Gr-8
                || ch == 'e' || ch == ''; // Alt-Gr-e
    }

    private static boolean isEmpty(char ch) {
        return ch == '' || ch == '' || ch == 'o' || ch == 'O';
    }

    private static int pos;

    private static RegularExpression InfixFactor(String re)
            throws REParserException {
        if (pos >= re.length()) {
            throw new REParserException(re, pos);
        }
        RegularExpression expression;
        char ch = re.charAt(pos);
        if (ch == 'a') {
            expression = Symbol.createASymbol();
        } else if (ch == 'b') {
            expression = Symbol.createBSymbol();
        } else if (isEpsilon(ch)) { // alternativen ...
            expression = new Epsilon();
        } else if (isEmpty(ch)) { // alternativen ...
            expression = new Empty();
        } else if (ch == '(') {
            pos++;
            expression = InfixExpression(re);
            if (pos < re.length() && re.charAt(pos) == ')') {
                //pos++;
            } else {
                throw new REParserException(re, pos);
            }
        } else {
            throw new REParserException(re, pos);
        }
        pos++;
        // check kleenestar
        while (pos < re.length() && re.charAt(pos) == '*') {
            pos++;
            expression = KleeneStar.kleen(expression);
        }
        return expression;
    }

    private static RegularExpression InfixTerm(String re)
            throws REParserException {
        RegularExpression factor = InfixFactor(re);
        if (pos < re.length() && re.charAt(pos) != '|' && re.charAt(pos) != ')') {
            factor = Concatenation.concatenation(factor, InfixTerm(re));
        }
        return factor;
    }

    private static RegularExpression InfixExpression(String re)
            throws REParserException {
        RegularExpression term = InfixTerm(re);
        if (pos < re.length()) {
            if (re.charAt(pos) == '|') {
                pos++;
                term = Union.union(term, InfixExpression(re));
            }
        }
        return term;
    }

    public static RegularExpression createInfixFormRE(String re)
            throws REParserException {
        pos = 0;
        RegularExpression expression = InfixExpression(re);
        if (pos == re.length()) {
            return expression;
        }
		throw new REParserException(re, pos);
    }

    public static RegularExpression createPrefixFormRE(String re)
            throws REParserException {
        Stack s = new Stack();
        int i = re.length() - 1;
        //	try {
        try {
            if (i == -1) {
                throw new REParserException(re, i);
            }
            while (i >= 0 && re.charAt(i) != '_') {
                char cur = re.charAt(i);
                if (cur == 'a') {
                    s.push(Symbol.createASymbol());
                } else if (cur == 'b') {
                    s.push(Symbol.createBSymbol());
                } else if (isEpsilon(cur)) {
                    s.push(new Epsilon());
                } else if (cur == '*') {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = KleeneStar.kleen(re1);
                    s.push(re2);
                } else if (cur == '') {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = Complement.complement(re1);
                    s.push(re2);
                } else if (cur == '&') {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = (RegularExpression) s.pop();
                    RegularExpression re3 = Intersection.intersect(re1, re2);
                    s.push(re3);
                } else if (cur == '|') {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = (RegularExpression) s.pop();
                    RegularExpression re3 = Union.union(re1, re2);
                    s.push(re3);
                } else if (cur == '-') {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = (RegularExpression) s.pop();
                    RegularExpression re3 = Minus.minus(re1, re2);
                    s.push(re3);
                } else if (cur == '.') {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = (RegularExpression) s.pop();
                    RegularExpression re3 = Concatenation.concatenation(re1,
                            re2);
                    s.push(re3);
                } else if (isEmpty(cur)) {
                    s.push(new Empty());
                } else {
                    throw new REParserException(re, i);
                }
                i--;
            }
            if (!s.empty()) {
                RegularExpression result = (RegularExpression) s.pop();
                if (s.empty()) {
                    return result;
                }
				throw new REParserException(re, re.length());
            }
			throw new REParserException(re, re.length());
        } catch (EmptyStackException e) {
            throw new REParserException(re, i);
        }
    } //createREfromPrefix

    /**
     * Method createREfromSuffix Method to create a RegularExpression object
     * from a given suffix RE string.
     * 
     * @param re
     *            String representation of a suffix RE with alphabet {'a', 'b',
     *            EPSILON, '.', '|', '*', '' }.
     * @return RegularExpression Returns a new RegularExpression object created
     *         from the given RE string.
     */
    public static RegularExpression createSuffixFormRE(String re)
            throws REParserException {
        Stack s = new Stack();
        int i = 0;
        //		try {
        if (re.length() == 0) {
            throw new REParserException(re, i);
        }
        while (i < re.length()) {
            char cur = re.charAt(i);
            if (cur == 'a') {
                s.push(Symbol.createASymbol());
            } else if (cur == 'b') {
                s.push(Symbol.createBSymbol());
            } else if (isEpsilon(cur)) {
                s.push(new Epsilon());
            } else if (cur == '*') {
                try {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = KleeneStar.kleen(re1);
                    s.push(re2);
                } catch (EmptyStackException e) {
                    throw new REParserException(re, i);
                }
            } else if (cur == '|') {
                try {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = (RegularExpression) s.pop();
                    RegularExpression re3 = Union.union(re2, re1);
                    s.push(re3);
                } catch (EmptyStackException e) {
                    throw new REParserException(re, i);
                }
            } else if (cur == '.') {
                try {
                    RegularExpression re1 = (RegularExpression) s.pop();
                    RegularExpression re2 = (RegularExpression) s.pop();
                    RegularExpression re3 = Concatenation.concatenation(re2,
                            re1);
                    s.push(re3);
                } catch (EmptyStackException e) {
                    throw new REParserException(re, i);
                }
            } else if (isEmpty(cur)) {
                s.push(new Empty());
            } else {
                throw new REParserException(re, i);
            }
            i++;
        }
        if (!s.empty()) {
            try {
                RegularExpression result = (RegularExpression) s.pop();
                if (s.empty()) {
                    return result;
                }
				throw new REParserException(re, re.length());
            } catch (EmptyStackException e) {
                throw new REParserException(re, re.length());
            }
        }
        return null;
    } //createREfromSuffix

    /**
     * Method createREfromEBNF Method to create a RegularExpression object from
     * a given EBNF RE string.
     * 
     * @param re
     *            String representation of a EBNF RE with alphabet {'a', 'b',
     *            EPSILON, '.', '|', '*', '' }.
     * @return RegularExpression Returns a new RegularExpression object created
     *         from the given RE string.
     */
    public static RegularExpression createBackusNaurFormRE(String re)
            throws REParserException {
        RegularExpression result = null;
        char cur;
        int i = 0;
        //		try {
        while (result == null) {
            if (i >= re.length()) {
                throw new REParserException(re, i);
            }
            cur = re.charAt(i);
            if (cur == 'a') {
                result = Symbol.createASymbol();
                i++;
            } else if (cur == 'b') {
                result = Symbol.createBSymbol();
                i++;
            } else if (isEpsilon(cur)) {
                result = new Epsilon();
                i++;
            } else if (cur == '(') {
                int oldI = ++i;
                for (int j = 1; j > 0; i++) {
                    if (i == re.length()) {
                        throw new REParserException(re, i);
                    }
                    if (re.charAt(i) == '(') {
                        j++;
                    } else if (re.charAt(i) == ')') {
                        j--;
                    }
                }
                i--;
                String tmp = re.substring(oldI, i);
                if (tmp.length() > 0) {
                    result = createBackusNaurFormRE(tmp);
                } else {
                    i++;
                }
            } else if (cur == '{') {
                int oldI = ++i;
                for (int j = 1; j > 0; i++) {
                    if (i == re.length()) {
                        throw new REParserException(re, i);
                    }
                    if (re.charAt(i) == '{') {
                        j++;
                    } else if (re.charAt(i) == '}') {
                        j--;
                    }
                }
                i--;
                String tmp = re.substring(oldI, i);
                if (tmp.length() > 0) {
                    result = createBackusNaurFormRE(tmp);
                    result = KleeneStar.kleen(result);
                } else {
                    i++;
                }
            } else if (isEmpty(cur)) {
                result = new Empty();
                i++;
            } else if (cur != '}' && cur != '(') {
                throw new REParserException(re, i);
            }
        }
        for (; i < re.length(); i++) {
            cur = re.charAt(i);
            if (cur == 'a') {
                RegularExpression a = Symbol.createASymbol();
                result = Concatenation.concatenation(result, a);
            } else if (cur == 'b') {
                RegularExpression b = Symbol.createBSymbol();
                result = Concatenation.concatenation(result, b);
            } else if (isEpsilon(cur)) {
                RegularExpression e = new Epsilon();
                result = Concatenation.concatenation(result, e);
            } else if (cur == '|') {
                result = Union.union(result, createBackusNaurFormRE(re
                        .substring(i + 1)));
                i = re.length();
            } else if (cur == '(') {
                i++;
                int oldI = i;
                for (int j = 1; j > 0; i++) {
                    if (i == re.length()) {
                        throw new REParserException(re, i);
                    }
                    if (re.charAt(i) == '(')
                        j++;
                    else if (re.charAt(i) == ')')
                        j--;
                }
                i--;
                String tmp = re.substring(oldI, i);
                if (tmp.length() > 0) {
                    result = Concatenation.concatenation(result,
                            createBackusNaurFormRE(tmp));
                }
            } else if (cur == '{') {
                i++;
                int oldI = i;
                for (int j = 1; j > 0; i++) {
                    if (i == re.length()) {
                        throw new REParserException(re, i);
                    }
                    if (re.charAt(i) == '{')
                        j++;
                    else if (re.charAt(i) == '}')
                        j--;
                }
                i--;
                String tmp = re.substring(oldI, i);
                if (tmp.length() > 0) {
                    result = Concatenation.concatenation(result, KleeneStar
                            .kleen(createBackusNaurFormRE(tmp)));
                }
            } else if (cur != '}' && cur != ')') {
                throw new REParserException(re, i);
            }
        }
        return result;
    } //createREfromEBNF
}